%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC/ibmrs-SPEC.SL
% Title:          Some special things for IBM RS6000 PSL compiler
% Author:         Winfried Neun , ZIB Berlin
% Created:        November 9, 1990
% Status:         Experimental
% Package:        Compiler
% (c) Copyright 1990, Konrad-Zuse Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load fasl-decls))

% a special Linke to save the NOP instruction after final BA

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (cond ((FlagP FunctionName 'ForeignFunction)
              `((*ForeignLink ,FunctionName ,FunctionType ,NumberOfArguments)
            (*Exit 0)))
        ((get functionname 'exitopencode)
              `((*DeAlloc ,DeAllocCount)
                (*JCall ,FunctionName)))
        ((get functionname 'opencode)
              `((*Call ,FunctionName)
                (*exit ,DeAllocCount)))
        ((InternallyCallableP functionname)
              `((*dealloc ,Dealloccount)
		(B (InternalEntry ,functionname))))
	(t    `((*dealloc ,Dealloccount)
                (*JCall ,functionname)))))

% change branch to exitlabel into *exit when framesize ne 0 too

(fluid '(exitt&))

(de &jumpopt (u)
    (when (equal (cadar u) exitt&) (rplaca u '(*exit (*framesize)))))

% turn off lapopt in faslend

(fluid '(*lapopt))

(setq *lapopt t)

(when (funboundp 'old-faslend) (copyd 'old-faslend 'faslend))

(de faslend ()
    (prog (*lapopt)
       (return (old-faslend))))

% open codes for cons ncons xcons

(defcmacro *SPEC-JumpIF)

(put 'cons 'opencode
      '((st (reg 1) (indexed (reg r20) 0))
	(rlinm (reg 1) (reg 20) 0 4 31)
	(oriu (reg 1) (reg 1) 16#4800 )   % result
	(st (reg 2) (indexed (reg r20) 4))
	(*wplus2 (reg r20) 8)
        (*jumpwlessp (labelgen templabel) (reg r20) (reg r21))
	(*alloc 1)
        (*move (reg 1) (frame 1))
        (*link !%reclaim expr 0)
        (*move (frame 1) (reg 1))
	(*dealloc 1)

        (labelref templabel)))

(put 'ncons 'opencode
      '((st (reg 1) (indexed (reg r20) 0))
        (rlinm (reg 1) (reg 20) 0 4 31)
        (oriu (reg 1) (reg 1) 16#4800 )   % result
        (st (reg nil) (indexed (reg r20) 4))
        (*wplus2 (reg r20) 8)
        (*jumpwlessp (labelgen templabel) (reg r20) (reg r21))
        (*alloc 1)
        (*move (reg 1) (frame 1))
        (*link !%reclaim expr 0)
        (*move (frame 1) (reg 1))
        (*dealloc 1)

        (labelref templabel))
)

(put 'xcons 'opencode
      '((st (reg 1) (indexed (reg r20) 4))
        (rlinm (reg 1) (reg 20) 0 4 31)
        (oriu (reg 1) (reg 1) 16#4800 )   % result
        (st (reg 2) (indexed (reg r20) 0))
        (*wplus2 (reg r20) 8)
        (*jumpwlessp (labelgen templabel) (reg r20) (reg r21))
        (*alloc 1)
        (*move (reg 1) (frame 1))
        (*link !%reclaim expr 0)
        (*move (frame 1) (reg 1))
        (*dealloc 1)

        (labelref templabel))
)

(put 'bittable 'opencode
%PPC%       '((sriq (reg 3) (reg 2) 2) % offset
       '((rlinm (reg 3) (reg 2) 30 2 31) % offset
	 (*mkitem (reg 1) string-tag)
         (lbzx (reg 5) (reg 1) (reg 3))
         (rlinm (reg 4) (reg 2) 1 29 30) % lower two bits <<1
         (ai (reg 4) (reg 4) -6)        % - fraction
         (neg (reg 4) (reg 4))  %   fraction
         (sr (reg 1) (reg 5) (reg 4))
         (andil!. (reg 1) (reg 1) 3)))

(put 'putbittable 'opencode
%PPC%  '((sriq (reg t3) (reg 2) 2) % offset
       '((rlinm (reg t3) (reg 2) 30 2 31) % offset
         (*mkitem (reg 1) string-tag)
         (lbzx (reg 5) (reg 1) (reg t3))
         (rlinm (reg 4) (reg 2) 1 29 30) % lower two bits <<1
         (ai (reg 4) (reg 4) -6)        % - fraction
         (neg (reg 4) (reg 4))  %   fraction
         (*move 3 (reg t2))
         (sl (reg t2) (reg t2) (reg 4))
         (andc (reg 5) (reg 5) (reg t2))
         (andil!. (reg 3) (reg 3) 3)        % value & 3
         (sl (reg t1) (reg 3) (reg 4)) % <<fraction
         (or (reg 5) (reg t1) (reg 5))  % or
         (stbx (reg 5) (reg 1) (reg t3))
         (*move (reg 3) (reg 1))))


% This code is intended for filling the 'holes' in the frame.
% e.g. in the following function
%
%  (de hugo (x) (cond ((null x) (call-another))
%                     (t (prog (a b c) ....))
%
% a frame of length 4 will be allocated, but it will not be preset (by nil)
% completely at entering the function but the frame elements for a b and c
% will be set (to nil) if the prog is entered.
% So if the (null x) case is taken we have 3 unset frame elements on the stack
% which contain random information and this is dangerous for gc.
% to resolve that problem, the following function will work on the input
% to pass-1-lap and insert nil settings where needed.

(fluid '(&fillframeholes &frame-numbers))

(de &fillframeholes (code)
   (prog (allo cmactoextern externfound ll lll nn linearcode restcode)
      (when (not &fillframeholes) (return code))
      (setq cmactoextern
                  '(*link *linke *call *jcall *cons *xcons *fastmkfltn))
             % trivial tests whether we have to work at all
      (unless (setq allo (findcmacro code '*alloc)) (return code))
      (setq ll cmactoextern)
      (while (and (not externfound) ll)
            (setq externfound (findcmacro code (car ll)))
            (setq ll (cdr ll)))
      (when (not externfound) (return code))

            % so we have to work, starting at the *alloc

      (setq nn (cadar allo))  %the frmae size
      (setq linearcode t)
      (setq restcode (cdr allo))

      (setq &frame-numbers nil)

      (while (and linearcode restcode)
         (setq linearcode (lookatinstruction (car restcode)))
         (setq restcode (cdr restcode)))

      (when linearcode (return code))

      (setq lll (setq ll (cons nil nil)))

             % search frame that have not been set, prepare code,
             % put into place after *alloc

      (for (from i 1 nn 1) (do
            (unless (memq i &frame-numbers)
                  (rplaca ll `(*move (reg nil) (frame ,i)))
                  (push 'nil ll))))

      (when (eq lll ll) (return code))
      (rplacd lll (cdr allo))

      (rplacd allo (cdr ll))
      (return code)))

(de lookatinstruction (instr)
      (cond ((atom instr) t)
            ((and (eq (car instr) '*move)
                  (eqcar (caddr instr) 'frame))
                       (push (cadr (caddr instr)) &frame-numbers) 17)
            ((memq (car instr) '(*link *call *linke *jcall *jumpintype
                 *jumpeq *jump  *JumpType *JumpNotType *JumpNotInType
                 *jumpnoteq *jumpwlessp *jumpwgreaterp *jumpwleq *jumpwgeq
                 *jumpon)) nil)
            (t t)))


(de findcmacro (list pat)
     (cond ((atom list ) nil)
           ((eqcar (car list) pat) list)
           (t        (findcmacro (cdr list) pat ))))

(setq &fillframeholes t)

(de codefiletrailer ()
  (prog (s)
        (systemfaslfixup)
        (binarywrite codeout* (idifference (isub1 nextidnumber*) 
					   first-local-id-number))
        % Number of local IDs
        (foreach x in (car orderedidlist*) do
                 (progn (remprop x fasl-idnumber-property*)
                        (setq x (faslid2string x))
                        (setq s (strlen x))
			(binarywrite codeout* s)
                        (binarywriteblock codeout* (strbase x)
					  (strpack s))))
        (binarywrite codeout* % S is size in words
                     (setq s
                      (iquotient
                       (iplus2 currentoffset*
                        (isub1 addressingunitsperitem))
		       addressingunitsperitem)))
        (binarywrite codeout* initoffset*)
        (binarywriteblock codeout* codebase* s)
        (binarywrite codeout*
                     (setq s
                      (iquotient
                       (iplus2 bittableoffset*
                        (isub1 bittable-entries-per-word))
                       bittable-entries-per-word)))
        (binarywriteblock codeout* bittablebase* s)
	(deallocatefaslspaces)))
