%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         GENINSTR
% Description:  CONVEX Generate instruction set
% Author:       H. Melenk
% Created:
% Modified:
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin, for modifications
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Note form of DefOpcode, similar to DefAnyreg or DefCmacro
%      A trailing OpFailure will be appended unless (T is used)
%      Currently TESTs require arguments and explicit AND
 
% (DefOpCode OP (ARGS)
%   ( (tests) (actions))
%   ( (tests) (actions))
% ....
% )
 
% those instructions having multiple suffixes are generated only in
% forms which are contained in the following list:
(setq relevant
  '(add!.w add!.l add!.d
    cvtw!.l cvth!.w  cvtl!.d cvtd!.l
    cvts!.l cvtw!.s cvts!.d cvtd!.s
    div!.w div!.l div!.d
    eq!.w  eq!.l
    jmpa!.t jmpa!.f
    jmps!.t jmps!.f
    br bra!.t bra!.f brs!.t brs!.f
    ld!.w ld!.l ld!.dl ld!.lu   ld!.b ld!.h
    ldea!.w ldea!.lu
  % ld@!.w ld@!.l ld@!.l ld@!.du
    le!.w  le!.l le!.d
    lt!.w  lt!.l lt!.d
    mov!.w mov!.l
    mul!.w mul!.l mul!.d
    pop!.w
    psh!.w
    st!.w st!.l st!.d  st!.b st!.h
    sub!.w sub!.l sub!.d
    neg!.w
    ))
 
(load strings)
(DE SORT (LST FN)
   (PROG (TREE)
      (COND ((OR (NULL LST) (NULL (CDR LST))) (RETURN LST)))
      (SETQ TREE (LIST (CAR LST) NIL))
      (WHILE (PAIRP (SETQ LST (CDR LST))) (TREEADD (CAR LST) TREE FN))
      (RETURN (TREE2LIST TREE NIL))))
(DE TREE2LIST (TREE LST)
   (COND
      ((NULL TREE) LST)
      (T (TREE2LIST
            (CADR TREE)
            (CONS (CAR TREE) (TREE2LIST (CDDR TREE) LST)))) ))
(DE TREEADD (ITEM NODE FN)
   (COND
      ((APPLY FN (list ITEM (CAR NODE)))
         (COND
            ((CADR NODE) (TREEADD ITEM (CADR NODE) FN))
            (T (RPLACA (CDR NODE) (LIST ITEM NIL)))) )
      ((CDDR NODE) (TREEADD ITEM (CDDR NODE) FN))
      (T (RPLACD (CDR NODE) (LIST ITEM NIL)))) )
 
% instructions are generated by a patters:
%
% each instruction ID has a slot INSTRCASES where the different
% possible cases are collected. When all instructions are complete,
% the final defOpCode calls are generated from these slots
 
(fluid '(allInstrs!* allInstrPatterns!* formalParameters!* instr*))
(fluid '(Op-k Op-j-k Op-i-j-k Op-load Op-store Op-immediate-000-k
         Op-immediate-001-k Op-branch-short Op-branch-long))
 
(setq formalParameters!* '(par1 par2 par3))
 
(ds newInstruction (i) (when (not (memq i allInstrs!*)) (push i allInstrs!*)))
 
(df instr (l)
     (prog (name operands format suffix gname group code n)
       (setq instr* l)
       (setq name (pop l))
       (setq operands (pop l))
       (setq format (pop l))
       (when (not(numberp (car l))) (go multiple))
    % simple instruction
       (newInstruction name)
       (setq code (car l))
       (push (mktest format code operands) (get name 'INSTRCASES))
       (setq gname name)
       (setq code (car l))
 %     (when (or (null (get gname 'argno))
 %               (lessp (get gname 'argno) (length operands)))
 %           (put gname 'argno (length operands)))
       (push (list name (car l) format operands) allInstrPatterns!*)
       (set format (cons (mkdisass code name operands) (eval format)))
       (return nil)
    % generated instruction
  multiple
       (when (null l) (return nil))
       (setq group (pop l))
       (setq code (pop group))
       (setq n 0)
    group
       (when (null group) (go multiple))
       (setq suffix (pop group))
       (when (eq suffix '-)(setq n (plus n 1)) (go group))
       (setq gname (intern (bldmsg "%w.%w" name suffix)))
           % unerwuenschte Befehle ausblenden
       (when (not (memq gname relevant)) (setq n(add1 n)) (go group))
       (push (mktest format (plus code n) operands) (get gname 'INSTRCASES))
 %     (when (or (null (get gname 'argno))
 %               (lessp (get gname 'argno) (length operands)))
 %           (put gname 'argno (length operands)))
       (newInstruction gname)
       (push (list gname (plus code n) format operands) allInstrPatterns!*)
       (set format (cons (mkdisass (plus code n)
                           gname operands) (eval format)))
       (setq n (plus n 1))
       (go group)))
 
 
(de mktest(format code operands)
   (prog (params lhs rhs type val)
      (setq params formalParameters!*)
  loop (when (null operands) (go ready))
      (setq val  (operandval  (car operands)))
      (setq type (operandtype (pop operands)))
      (if   (not (eqcar type 'UNQUOTE))
            (push (append type (list (car params))) lhs)
            (progn
               (setq params (cons(cadr type) (cdr params))) % artific. param
               (setq type '(QUOTE))))
      (if   (not (eqcar type 'EQUAL))
            (setq rhs(append rhs (list (append val  (list (pop params))))))
            (pop params))
      (go loop)
  ready
     (if (null lhs)(setq lhs '(T)))
     (return (if (cdr lhs)
                `((and .,(reversip lhs))(,format ,code ., rhs))
                `(,(car lhs)(,format ,code ., rhs))))
    ))
 
(de operandtype (op)
    (cond ((eqcar op 'QUOTE) (list 'EQUAL op))
          ((eqcar op 'UNQUOTE) op)
          ((memq op '(Ai Aj Ak)) '(AREGP))
          ((memq op '(Si Sj Sk)) '(Sregp))
          ((memq op '(Vi Vj Vk)) '(Vregp))
          ((eq op '!#N) '(STDIMMEDIATEP))
          ((eq op '!#!n) '(SMALLIMMEDIATEP))
          ((eq op '<effa>) '(EFFAP))
          ((eq op 'shortlabel) '(SHORTLABELP))
          (t (prin2t "unknown operand type during instruction generation:")
             (prin2t op)
             (prin2t instr*)
             (error  nil))))
 
(de operandval(op)
   % according to type of operand select valuefunction
    (cond ((eqcar op 'QUOTE) (list 'EQUAL op))
          ((eqcar op 'UNQUOTE) (list 'QUOTE (cadr op)))
          ((memq op '(Ai Aj Ak)) '(AREGP))
          ((memq op '(Si Sj Sk)) '(Sregp))
          ((memq op '(Vi Vj Vk)) '(Vregp))
          ((eq op '!#N) '(STDIMMEDIATEVAL))
          ((eq op '!#!n) '(SMALLIMMEDIATEP))
          ((eq op '<effa>) '(EFFAVAL))
          ((eq op 'shortlabel) '(shortlabelval))
          (t (prin2t "unknown operand type during instruction generation:")
             (prin2t op)
             (prin2t instr*)
             (error  nil))))
 
 
 
(de parameterlist (n) % (parameterlist1 n formalParameters!*))
                     '(par1 par2 par3))
 
(de parameterlist1(n l)
   (if (eqn n 0) nil (cons (car l)(parameterlist1(difference n 1)(cdr l)))))
 
% clear all instructions
(de clearInstructions ()
    (setq allInstrPatterns!* nil)
    (mapc allInstrs!* (function(lambda(u)(remprop u 'INSTRCASES))))
    (setq allInstrs!* nil))
 
% make a disassemble-record
(de mkdisass(code gname operands)
    `(,code (NAME . ,gname) (PAT . ,(dissasemblepat operands))))
 
(de dissasemblepat(o)  % special handling for quotes
   (cond ((atom o) o)
         ((eqcar (car o) 'QUOTE) (cons (cadr (car o)) (cdr o)))
         ((eqcar (car o) 'UNQUOTE) (cdr o))
         (T (cons (car o) (dissasemblepat (cdr o))))))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% 1st application: generation of defOpcode file
%
 
(de collectInstructions (file)
   (when file (setq file (open file 'OUTPUT)) (wrs file))
   (setq allInstrs!* (sort allInstrs!* (function string-lessp)))
   (mapc allInstrs!* (function makeOneInstruction))
   (when file (wrs nil) (close file)))
 
(de makeOneInstruction (u)
  (setq u
   `(DefOpcode ,u ,(parameterlist (get u 'ARGNO))
                 .,(get u 'INSTRCASES)))
  (eval (list 'pp u))
  (terpri))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% 2nd application: generation of a reference table
%
 
(fluid '(pattern))
(de oneparcel (x) (when (null pattern) (setq pattern x)))
(de convexputparcel (x) (when (null pattern) (setq pattern x)))
 
(de instructionPattern(pat)
       (prog (fct code)
          (setq pattern nil)
          (setq fct (caddr pat))
          (setq code (cadr pat))
          (when (not(numberp code))
              (prin2 pat) (prin2 " has code ") (prin2t code))
          (cond((eq fct 'Op-i-j-k)(Op-i-j-k code 0 0 0))
               ((eq fct 'Op-j-k)(Op-j-k code 0 0))
               ((eq fct 'Op-k)(Op-k code  0))
               ((eq fct 'Op-immediate-000-k)(Op-immediate-000-k code 0 0))
               ((eq fct 'Op-immediate-001-k)(Op-immediate-001-k code 0 0))
               ((eq fct 'Op-load)(op-load code '(0 . 0) 0))
               ((eq fct 'Op-store)(op-store code 0 '(0 . 0)))
               ((eq fct 'Op-branch-short)(Op-branch-short code 0))
               ((eq fct 'Op-branch-long)(Op-branch-long code '(0 . 0)))
               (t (print pat)(prin2t "without legal code")(error 99)))
         (return (cons pattern pat)) ))
 
(de instructionDisplay(id)
       (setq outputbase!* 16)
       (prin2 (car id))
       (setq outputbase!* 10)
       (tab 10)
       (prin2 (cadr id))
       (tab 20)
       (prin2 (caddr (cddr id)))
       (tab 40)
       (prin2 (cadddr id))
       (terpri))
 
(de displayInstructions(file)
 (prog (l)
   (setq *GenerateCode T)
   (setq l (mapcar allInstrPatterns!* (function instructionPattern)))
   (when file (setq file (open file 'OUTPUT)) (wrs file))
   (prin2t "********** ascending code")
   (setq l (sort l (function (lambda(u v) (lessp (car u) (car v))))))
   (mapc l (function instructionDisplay))
   (prin2t "********** ascending name")
   (setq l (sort l (function (lambda(u v)
                 (mylessp (cons (cadr u)(caddr (cddr u)))
                          (cons (cadr v)(caddr (cddr v))))))))
   (mapc l (function instructionDisplay))
   (when file (wrs nil) (close file))))
 
(de mylessp (u v)
   (cond ((null u)(null v))
         ((equal (car u)(car v)) (mylessp (cdr u)(cdr v)))
         ((or (not (idp (car u)))(not(idp(car  v)))) t)
         (t (string-lessp (car u)(car v)))))
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%     3rd application: generate tables for disassemble
%
 
(de disassembletables(file)
   (when file (setq file (open file 'OUTPUT)) (wrs file))
   (eval `(pp (setq Op-k ',Op-k)))
   (eval `(pp (setq Op-j-k ',Op-j-k)))
   (eval `(pp (setq Op-i-j-k ',Op-i-j-k)))
   (eval `(pp (setq Op-immediate-000-k ',Op-immediate-000-k)))
   (eval `(pp (setq Op-immediate-001-k ',Op-immediate-001-k)))
   (eval `(pp (setq Op-load ',Op-load)))
   (eval `(pp (setq Op-store ',Op-store)))
   (eval `(pp (setq Op-branch-long ',Op-branch-long)))
   (eval `(pp (setq Op-branch-short ',Op-branch-short)))
   (when file (wrs nil) (close file))
   nil)
 
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% the instructions in the sequence defined by the architecture
%% manual
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(clearInstructions)
 
 
%%%%%
%%%%%
%%%%% address Register Instruction Set
%%%%%
%%%%%
(instr ld (<effa> Ak) Op-load (2#001010000 b - h - w))
(instr st (Ak <effa>) Op-store (2#001011000 b - h - w))
(instr ld@ (<effa> Ak) Op-load (2#001010001 b - h - w))
(instr st@ (Ak <effa>) Op-store (2#001011001 b - h - w))
(instr ld(!#N Ak) Op-immediate-000-k  (2#000100010 h w))
(instr ld (!#!n Ak) Op-j-k (2#0100010010 h w))
(instr ldea(!#N Ak) Op-immediate-000-k  (2#000100010 h w))
(instr ldea(!#!n Ak) Op-j-k (2#0100010010 h w))
(instr add(Aj Ak) Op-j-k (2#0101100000 h w))
(instr sub (Aj Ak) Op-j-k (2#0101101000 h w))
(instr mul (Aj Ak) Op-j-k (2#0101110000 h w))
(instr div (Aj Ak) Op-j-k (2#0101111000 h w))
(instr neg (Aj Ak) Op-j-k (2#0101011010 h w))
(instr add (!#N Ak) Op-immediate-000-k  (2#000101000 h w))
(instr add (!#!n Ak) Op-j-k (2#0101100010 h w))
(instr sub (!#N Ak) Op-immediate-000-k  (2#000101010 h w))
(instr sub (!#!n Ak) Op-j-k (2#0101101010 h w))
(instr mul (!#N Ak) Op-immediate-000-k  (2#000101100 h w))
(instr mul (!#!n Ak) Op-j-k (2#0101110010 h w))
(instr div (!#N Ak) Op-immediate-000-k  (2#000101110 h w))
(instr div (!#!n Ak) Op-j-k (2#0101111010 h w))
(instr add (Sj Ak) Op-j-k (2#0101000000 w))
(instr and (Aj Ak) Op-j-k 2#0101001000)
(instr or (Aj Ak) Op-j-k 2#0101001001)
(instr xor (Aj Ak) Op-j-k 2#0101001010)
(instr not (Aj Ak) Op-j-k 2#0101001011)
(instr and (!#N Ak) Op-immediate-000-k  2#000100100)
(instr or (!#N Ak) Op-immediate-000-k  2#000100101)
(instr xor(!#N Ak) Op-immediate-000-k  2#000100110)
(instr shf(Aj Ak) Op-j-k 2#0101000001)
(instr shf(!#N Ak) Op-immediate-000-k  2#000100111)
(instr shf(!#!n Ak) Op-j-k 2#0100010001)
(instr ldea (<effa> Ak) Op-load 2#000010010)
(instr ldea@ (<effa> Ak) Op-load 2#000010011)
% pshea
(instr mov('PC Ak) Op-k 2#0111110001010)
(instr mov(Aj Ak) Op-j-k 2#0101000010)
(instr psh!.w (Ak) Op-k 2#0111110100000)
(instr pop!.w (Ak) Op-k 2#0111110100010)
% tas
% ldpa
(instr mov('PSW Ak) Op-k 2#0111110001000)
(instr mov(Ak 'PSW) Op-k 2#0111110001001)
(instr le (Aj Ak) Op-j-k (2#0100110000 h w))
(instr lt (Aj Ak) Op-j-k (2#0100111000 h w))
(instr eq (Aj Ak) Op-j-k (2#0100011000 h w))
(instr leu (Aj Ak) Op-j-k (2#0100100000 h w))
(instr ltu (Aj Ak) Op-j-k (2#0100101000 h w))
(instr le (!#N Ak) Op-immediate-000-k (2#000111100 h w))
(instr lt (!#N Ak) Op-immediate-000-k (2#000111110 h w))
(instr eq (!#N Ak) Op-immediate-000-k (2#000110110 h w))
(instr le (!#!n Ak) Op-j-k (2#0100110010 h w))
(instr lt (!#!n Ak) Op-j-k (2#0100111010 h w))
(instr eq (!#!n Ak) Op-j-k (2#0100011010 h w))
(instr leu (!#N Ak) Op-immediate-000-k (2#000111000 h w))
(instr ltu (!#N Ak) Op-immediate-000-k (2#000111010 h w))
(instr leu (!#!n Ak) Op-j-k (2#0100100010 h w))
(instr ltu (!#!n Ak) Op-j-k (2#0100101010 h w))
(instr cvtw (Aj Ak) Op-j-k (2#0100000000 b h))
(instr cvtb!.w (Aj Ak) Op-j-k 2#0100000010)
(instr cvth!.w (Aj Ak) Op-j-k 2#0100000011)
 
%%%%%
%%%%%
%%%%% scalar Register Instruction Set
%%%%%
%%%%%
(instr ld (<effa> Sk) Op-load  (2#001100000 b - h - w - l)
                               (2#001100100 s - d))
(instr st (Sk <effa>) Op-store (2#001101000 b - h - w - l)
                               (2#001101100 s - d))
(instr ld@(<effa> Sk) Op-load  (2#001100001 b - h - w - l)
                               (2#001100101 s - d))
(instr st@(Sk <effa>) Op-store (2#001101001 b - h - w - l)
                               (2#001101101 s - d))
(instr ld(!#N Sk) Op-immediate-001-k (2#000100000 d u l w)
                                     (2#000100001 du - dl)
                                     (2#000100001 lu - ll))
(instr ldea(!#N Sk) Op-immediate-001-k (2#000100000 d u l w)
                                     (2#000100001 du - dl)
                                     (2#000100001 lu - ll))
(instr add (sj sk) Op-j-k (2#0101100100 b h w l) (2#0101010100 s d))
(instr sub (sj sk) Op-j-k (2#0101101100 b h w l) (2#0101010110 s d))
(instr mul (sj sk) Op-j-k (2#0101110100 b h w l) (2#0101011100 s d))
(instr div (sj sk) Op-j-k (2#0101111100 b h w l) (2#0101011110 s d))
(instr neg (sj sk) Op-j-k (2#0110111100 b h w l) (2#0110010110 s d))
(instr and (sj sk) Op-j-k 2#0101001100)
(instr or (sj sk) Op-j-k 2#0101001101)
(instr xor (sj sk) Op-j-k 2#0101001110)
(instr not (sj sk) Op-j-k 2#0101001111)
(instr add (!#N Sk) Op-immediate-001-k (2#000101000 h w) (2#000110000 s))
(instr sub (!#N Sk) Op-immediate-001-k (2#000101010 h w) (2#000110001 s))
(instr mul (!#N Sk) Op-immediate-001-k (2#000101100 h w) (2#000110010 s))
(instr div (!#N Sk) Op-immediate-001-k (2#000101110 h w) (2#000110011 s))
(instr and (!#N Sk) Op-immediate-001-k 2#000100100)
(instr or  (!#N Sk) Op-immediate-001-k 2#000100101)
(instr xor (!#N Sk) Op-immediate-001-k 2#000100110)
(instr shf (!#N Sk) Op-immediate-001-k 2#00100111)
(instr psh (Sk) Op-k (2#0111110100100 w l))
(instr pop (Sk) Op-k (2#0111110100110 w l))
(instr le  (sj sk) Op-j-k   (2#0100110100 b h w l)   (2#0101010000 s d))
(instr lt  (sj sk) Op-j-k   (2#0100111100 b h w l)   (2#0101010010 s d))
(instr eq  (sj sk) Op-j-k   (2#0100011100 b h w l)   (2#0101011000 s d))
(instr leu  (sj sk) Op-j-k  (2#0100100100 b h w l))
(instr ltu  (sj sk) Op-j-k  (2#0100101100 b h w l))
(instr le (!#N Sk) Op-immediate-001-k (2#000111100 h w) (2#000110100 s))
(instr lt (!#N Sk) Op-immediate-001-k (2#000111110 h w) (2#000110101 s))
(instr eq (!#N Sk) Op-immediate-001-k (2#000110110 h w))
(instr leu(!#N Sk) Op-immediate-001-k (2#000111000 h w))
(instr ltu(!#N Sk) Op-immediate-001-k (2#000111010 h w))
(instr cvtw!.s (Sj Sk) Op-j-k 2#0100001000)
(instr cvts!.w (Sj Sk) Op-j-k 2#0100001001)
(instr cvtd!.s (Sj Sk) Op-j-k 2#0100001010)
(instr cvts!.d (Sj Sk) Op-j-k 2#0100001011)
(instr cvtw!.b (Sj Sk) Op-j-k 2#0100000100)
(instr cvtw!.h (Sj Sk) Op-j-k 2#0100000101)
(instr cvtb!.w (Sj Sk) Op-j-k 2#0100000110)
(instr cvth!.w (Sj Sk) Op-j-k 2#0100000111)
(instr cvts!.l (Sj Sk) Op-j-k 2#0100001100)
(instr cvtd!.l (Sj Sk) Op-j-k 2#0100001101)
(instr cvtl!.s (Sj Sk) Op-j-k 2#0100001110)
(instr cvtl!.d (Sj Sk) Op-j-k 2#0100001111)
(instr cvtl!.w (Sj Sk) Op-j-k 2#0100010100)
(instr cvtw!.l (Sj Sk) Op-j-k 2#0100010101)
(instr shf (Sj Sk) Op-j-k 2#0101000101)
(instr tzc (Sj Sk) Op-j-k 2#0100010111)
(instr plc!.t (Sj Sk) Op-j-k 2#0100010110)
(instr mov (Sj Ak) Op-j-k 2#0101000011)
(instr mov (Aj Sk) Op-j-k 2#0101000111)
(instr mov(Sj Sk) Op-j-k (2#0101000100 w - l))
(instr mov(Sj Sk) Op-j-k (2#0101000100 s - d))
 
%%%%%
%%%%%
%%%%% Program Control Instruction Set
%%%%%
%%%%%
(instr nop ((unquote 0)) Op-branch-short 2#01110000)
   % now the true br instructions (sequence here important)
(instr br (shortlabel) Op-branch-short 2#01110001)
(instr bri(shortlabel) Op-branch-short (2#01110010 f t))
(instr bra(shortlabel) Op-branch-short (2#01110100 f t))
(instr brs(shortlabel) Op-branch-short (2#01110110 f t))
    %
(instr bkpt ((unquote 0)) Op-k 2#0111110101010)
(instr exit ((unquote 0)) Op-branch-long 2#000000000)
(instr jmp (<effa>) Op-branch-long 2#000000010)
(instr jmp! !@ (<effa>) Op-branch-long 2#000000011)
(instr jmpi (<effa>) Op-branch-long (2#000000100 f - t))
(instr jmpa (<effa>) Op-branch-long (2#000001000 f - t))
(instr jmps (<effa>) Op-branch-long (2#000001100 f - t))
%(instr call (<effa>) Op-branch-long 2#001000010)
(instr rtn ((unquote 0)) Op-k 2#0111110010010)
(instr callq (<effa>) Op-branch-long 2#001000100)
(instr callq! !@ (<effa>) Op-branch-long 2#001000101)
(instr calls (<effa>) Op-branch-long 2#001000010)
(instr calls! !@ (<effa>) Op-branch-long 2#001000011)
(instr rtnq ((unquote 0)) Op-k 2#0111110010000)
(instr rtn  ((unquote 0)) Op-k 2#0111110010010)
% sys
 
%%%%%
%%%%%
%%%%% scalar/vector Instruction Set
%%%%%
%%%%%
(instr ld (<effa> Vk) Op-load (2#001110000 b - h - w - l)(2#001110010 s - d))
(instr ld@(<effa> Vk) Op-load (2#001110001 b - h - w - l)(2#001110011 s - d))
(instr st (Vk <effa>) Op-store(2#011110000 b - h - w - l)(2#011110010 s - d))
(instr st@(Vk <effa>) Op-store(2#011110001 b - h - w - l)(2#011110011 s - d))
(instr ldvi (Vj Vk) Op-j-k(2#0111100000 b h w l)(2#0111100010 s d))
(instr stvi (Vj Vk) Op-j-k (2#0111101000 b h w l)(2#0111101010 s d))
(instr stvi (Sk Vj) Op-j-k (2#0111101100 b h w l)(2#0111101110 s d))
(instr ste (Sk Aj !#N) Op-store(2#001001000 b - h - w - l)(2#001001100 s - d))
(instr add (Vi Vj Vk) Op-i-j-k(2#1100000 b h w l)(2#1011000 s d))
(instr sub (Vi Vj Vk) Op-i-j-k(2#1101000 b h w l)(2#1011010 s d))
(instr mul (Vi Vj Vk) Op-i-j-k(2#1110000 b h w l)(2#1001000 s d))
(instr div (Vi Vj Vk) Op-i-j-k(2#1111000 b h w l)(2#1001010 s d))
(instr neg (Vj Vk) Op-j-k (2#0110111000 b h w l) (2#0110010010 s d))
(instr add (Vi Sj Vk) Op-i-j-k(2#1100100 b h w l)(2#1011100 s d))
(instr sub (Vi Sj Vk) Op-i-j-k(2#1101100 b h w l)(2#1011110 s d))
(instr mul (Vi Sj Vk) Op-i-j-k(2#1110100 b h w l)(2#1001100 s d))
(instr div (Vi Sj Vk) Op-i-j-k(2#1111100 b h w l)(2#1001110 s d))
(instr and (Vi Vj Vk) Op-i-j-k 2#1010000)
(instr or  (Vi Vj Vk) Op-i-j-k 2#1010001)
(instr xor (Vi Vj Vk) Op-i-j-k 2#1010010)
(instr not (Vj Vk) Op-j-k 2#10110001011)
(instr and (Vi Sj Vk) Op-i-j-k 2#1010100)
(instr or  (Vi Sj Vk) Op-i-j-k 2#1010101)
(instr xor (Vi Sj Vk) Op-i-j-k 2#1010110)
(instr shf (Sj Vk) Op-j-k 2#0110001100)
(instr mov (Si Sj Vk) Op-i-j-k 2#1000001)
(instr mov (Vi Sj Sk) Op-i-j-k 2#1000000)
 
%%%%%
%%%%%
%%%%% comparison/Mask/Merge/Compress
%%%%%
%%%%%
(instr le (Vj Vk) Op-j-k (2#0110101000 b h w l)(2#0110011000 s d))
(instr lt (Vj Vk) Op-j-k (2#0110110000 b h w l)(2#0110011010 s d))
(instr eq (Vj Vk) Op-j-k (2#0110100000 b h w l)(2#0110010000 s d))
(instr le (Sj Vk) Op-j-k (2#0110101100 b h w l)(2#0110011100 s d))
(instr lt (Sj Vk) Op-j-k (2#0110110100 b h w l)(2#0110011110 s d))
(instr eq (Sj Vk) Op-j-k (2#0110100100 b h w l)(2#0110010100 s d))
(instr cprs (Vj Vk) Op-j-k (2#0110001110 f t))
(instr merg!.t (Vi Vj Vk) Op-i-j-k 2#1000010)
(instr merg (Vi Sj Vk) Op-i-j-k (2#1000100 f - t))
(instr mask!.t (Vi Vj Vk) Op-i-j-k 2#1000011)
(instr mask  (Vi Sj Vk) Op-i-j-k (2#1000101 f - t))
 
%%%%%
%%%%%
%%%%%  vector reduction
%%%%%
%%%%%
(instr sum (Vk) Op-k (2#0111111000000 b h w l)(2#0111111010000 s d))
(instr prod(Vk) Op-k (2#0111111011000 b h w l)(2#0111111010010 s d))
(instr max (Vk) Op-k (2#0111111001000 b h w l)(2#0111111010100 s d))
(instr min (Vk) Op-k (2#0111111001100 b h w l)(2#0111111010110 s d))
(instr all (Vk) Op-k 2#0111111000100)
(instr any (Vk) Op-k 2#0111111000101)
(instr parity (Vk) Op-k 2#0111111000110)
(instr plc!.t (Vj Vk) Op-j-k 2#0110001101)
 
%%%%%
%%%%%
%%%%%  Vl,Vs and VM
%%%%%
%%%%%
(instr mov(Ak 'VL) Op-k 2#0111110110011)
(instr mov('VL Ak) Op-k 2#0111110110010)
(instr ld!.w (!#N 'VL) Op-immediate-000-k 2#000110000)
(instr mov ('VS Ak) Op-k 2#0111110110000)
(instr mov (Ak 'VS) Op-k 2#0111110110001)
(instr ld!.w  (!#N 'VS) Op-immediate-000-k 2#000110001)
(instr ld!.l (<effa> 'VLS) Op-load 2#000010100)
(instr st!.l ('VLS <effa>) Op-store 2#000011100)
(instr mov!.w (Sk 'Vl) Op-k 2#0111110110111)
(instr mov!.w (Sk 'VS) Op-k 2#0111110110101)
(instr ld!.x (<effa> 'VM) Op-load 2#000010110)
(instr st!.x ('VM <effa>) Op-store 2#000011110)
(instr plc ('VM Sk) Op-k (2#0111111011100 f t))
(instr mov (Sj 'VM Sk) Op-J-k 2#0110000101)
(instr mov (Sj Sk 'VM) Op-J-k 2#0110000100)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% now we generate the CXINSTR dataset
(pp nil)
(collectInstructions "convex-instr.sl")
(disassembletables "disasstb")
(displayInstructions NIL)
 
 
 
