%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File: VECTOR-INSTRUCTIONS
%
% Description:   Vector instruction interface for Cray X-MP:
%                Vector instructions, vector Cmacros
%
% Date:          19 Jun 1986
% Author:        H. Melenk, ZIB Berlin
%
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(*VectorInstructions*))   % list of all vector instrs
 
(loadtime (load compiler))
 
(compiletime (load big-faslend))
 
(compiletime
                      % defopcode with table update
(dm defVopcode (u)
   `(progn
       (setq *VectorInstructions*
              (cons (quote ,(cadr u)) *vectorinstructions*))
      ,(cons 'defopcode (cdr u)) ))
 
)                                           % compiletime
 
% associate with a number 0-7 a vector register and a vectorlength reg
 
(de vregname (x)  (nth '((reg v0) (reg v1) (reg v2) (reg v3)
                         (reg v4) (reg v5) (reg v6) (reg v7))
                         (wplus2 x 1) ))
(put 'vregname 'operandvaluefunction 'vregname)
 
(de vlthname (x)  (nth '((reg b10) (reg b11) (reg b12) (reg b13)
                         (reg b14) (reg b15) (reg b16) (reg b17))
                         (wplus2 x 1) ))
(put 'vlthname 'operandvaluefunction 'vlthname)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(put 'vector1pat 'pattern '(nil nil (any (mac a1))))
 
(put 'vector1valuepat 'pattern '(nil nil (any (mac a1 dest)
						('*DESTROY dest))))
 
(put 'vector2pat 'pattern '(nil nil (any (mac a1 a2))))
 
(put 'vector2valuepat 'pattern '(nil nil (any (mac a1 a2 dest)
						 ('*DESTROY dest))))
 
(put 'vector3pat 'pattern '(nil nil (any (mac a1 a2 a3))))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  definition of vector op codes, cmacros and
%  interface functions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  load, store, element handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%----------------------------------------------------------------------%
% (vstore (vreg 3) (loc (getmem x)) 4)
% store from vectorregister v3 to memory with offset 4
%----------------------------------------------------------------------%
 
(put 'vstore 'openfn '(vector3pat *vstore))
 
(de *vstore (a1 a2 a3)
    (if (eqcar a2 '$aplus2)
        (expand4operandcmacro a1 (cadr a2) (caddr a2) a3 '*vstore4op )
        (expand3operandcmacro a1 a2 a3 '*vstore )))
 
(de *vstore4op (a1 a2 a3 a4)
        (expand4operandcmacro a1 a2 a3 a4 '*vstore4op))
 
(defcmacro *vstore
    ((numberp anyp onep) (*move argtwo (reg a0))
                         (vstore (vregname argone) (reg A0)))
    ((numberp anyp anyp) (*move argtwo (reg a0))
                         (*move argthree (reg a6))
                         (vstore (vregname argone) (reg A6)))
    (                    (*cerror "weird vector instr")))
 
(defcmacro *vstore4op
    ((numberp aregp aregp onep) (*wplus3 (reg a0) argtwo argthree)
                                (vstore (vregname argone) (reg A0)))
    ((numberp aregp aregp anyp) (*wplus3 (reg a0) argtwo argthree)
                                (*move argfour (reg a6))
                                (vstore (vregname argone) (reg A6)))
    ((numberp aregp anyp anyp)  (*move argthree (Reg a7))
                                (*vstore4op argone argtwo (reg a7)
                                           argfour))
    ((numberp anyp  anyp anyp)  (*move argtwo (Reg a6))
                                (*vstore4op argone (reg a6) argthree
                                           argfour))
    (                    (*cerror "weird vector instr")))
 
 
 (DefVOpCode    Vstore  (v m)                     % ,A0,Am   Vj
  ((And (VregP v)(Aregp m))  (op-i-j-k 8#177 0 (VregP v) (AregP m)))
 )
 
%----------------------------------------------------------------------%
% (vscatter (vreg 3) (loc (getmem x)) (vreg 4))
% store from vectorregister v3 to memory with offset vreg 4
%----------------------------------------------------------------------%
 
(put 'vscatter 'openfn '(vector3pat *vscatter))
 
(de *vscatter (a1 a2 a3)
    (if (eqcar a2 '$aplus2)
        (expand4operandcmacro a1 (cadr a2) (caddr a2) a3 '*vscatter4op )
        (expand3operandcmacro a1 a2 a3 '*vscatter )))
 
(de *vscatter4op (a1 a2 a3 a4)
        (expand4operandcmacro a1 a2 a3 a4 '*vscatter4op))
 
(defcmacro *vscatter
    ((numberp anyp numberp) (*move argtwo (reg a0))
                    (vscatter (vregname argone) (vregname argthree)))
    (               (*cerror "weird vector instr")))
 
(defcmacro *vscatter4op
    ((numberp aregp aregp numberp) (*wplus3 (reg a0) argtwo argthree)
                       (vscatter (vregname argone) (vregname argfour)))
    ((numberp aregp anyp numberp) (*move argthree (Reg a7))
                                  (*vscatter4op argone argtwo (reg a7)
                                                            argfour))
    ((numberp anyp anyp numberp) (*move argtwo (Reg a6))
                                 (*vscatter4op argone (reg a6) argthree
                                                              argfour))
    (                    (*cerror "weird vector instr")))
 
 
 (DefVOpCode    Vscatter (v k)                    % ,A0,Vk   Vj
  ((And (VregP v)(Vregp k))  (op-i-j-k 8#177 1 (VregP v) (VregP k)))
 )
 
%----------------------------------------------------------------------%
% (vload (vreg 3) (loc (getmem x)) 4)
% store from vectorregister v3 to memory with offset 4
%----------------------------------------------------------------------%
 
(put 'vload  'openfn '(vector3pat *vload ))
 
(de *vload  (a1 a2 a3)
    (if (eqcar a2 '$aplus2)
        (expand4operandcmacro a1 (cadr a2) (caddr a2) a3 '*vload4op )
        (expand3operandcmacro a1 a2 a3 '*vload )))
 
(de *vload4op (a1 a2 a3 a4)
        (expand4operandcmacro a1 a2 a3 a4 '*vload4op))
 
(defcmacro *vload
    ((numberp anyp onep) (*move argtwo (reg a0))
                         (vload  (vregname argone) (reg A0)))
    ((numberp anyp anyp) (*move argtwo (reg a0))
                         (*move argthree (reg a6))
                         (vload (vregname argone) (reg A6)))
    (                    (*cerror "weird vector instr")))
 
(defcmacro *vload4op
    ((numberp aregp aregp onep) (*wplus3 (reg a0) argtwo argthree)
                                (vload  (vregname argone) (reg A0)))
    ((numberp aregp aregp anyp) (*wplus3 (reg a0) argtwo argthree)
                                (*move argfour (reg a6))
                                (vload (vregname argone) (reg A6)))
    ((numberp aregp anyp anyp)  (*move argthree (Reg a7))
                                (*vload4op argone argtwo (reg a7)
                                           argfour))
    ((numberp anyp  anyp anyp)  (*move argtwo (Reg a6))
                                (*vload4op argone (reg a6) argthree
                                           argfour))
    (                    (*cerror "weird vector instr")))
 
 
 (DefVopCode     Vload   (v k)                     %  Vj   ,A0,Ak
  ((And (VregP v)(Aregp k))  (op-i-j-k 8#176 (VregP v) 0 (AregP k)))
 )
 
%----------------------------------------------------------------------%
% (vgather (vreg 3) (loc (getmem x)) (vreg 4))
% store from vectorregister v3 to memory with offset vreg 4
%----------------------------------------------------------------------%
 
(put 'vgather 'openfn '(vector3pat *vgather ))
 
(de *vgather (a1 a2 a3)
    (if (eqcar a2 '$aplus2)
        (expand4operandcmacro a1 (cadr a2) (caddr a2) a3 '*vgather4op )
        (expand3operandcmacro a1 a2 a3 '*vgather )))
 
(de *vgather4op (a1 a2 a3 a4)
        (expand4operandcmacro a1 a2 a3 a4 '*vgather4op))
 
(defcmacro *vgather
    ((numberp anyp numberp) (*move argtwo (reg a0))
                       (vgather (vregname argone) (vregname argthree)))
    (                  (*cerror "weird vector instr")))
 
(defcmacro *vgather4op
    ((numberp aregp aregp numberp) (*wplus3 (reg a0) argtwo argthree)
                        (vgather (vregname argone) (vregname argfour)))
    ((numberp aregp anyp numberp)  (*move argthree (Reg a7))
                                   (*vgather4op argone argtwo (reg a7)
                                                 argfour))
    ((numberp anyp anyp numberp)  (*move argtwo (Reg a6))
                                  (*vgather4op argone (reg a6) argthree
                                                              argfour))
    (                    (*cerror "weird vector instr")))
 
 
 (DefVopCode     Vgather (v k)                     %  Vj   ,A0,Vk
  ((And (VregP v)(Vregp k)) (op-i-j-k 8#176 (VregP v) 1 (VregP k))))
 
%----------------------------------------------------------------------%
% (VsetElement (vreg 3) n x)
% set element n of v3 to value x
%----------------------------------------------------------------------%
 
(put 'VsetElement  'openfn '(vector3pat *vsetel))
 
(de *vsetel (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vsetel))
 
(defcmacro *vsetel
    ((numberp anyp sregp)
                         (*move argtwo (reg a6))
                         (VmoveS (vregname argone) (reg A6) argthree))
    ((numberp anyp anyp)
                         (*move argtwo (reg a6))
                         (*move argthree (reg s6))
                         (VmoveS (vregname argone) (reg A6)(reg s6)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VmoveS  (v a s)                   % Vi,Ak    Sj
  ((and (Vregp v) (Aregp a) (Sregp s))
              (op-i-j-k 8#077 (Vregp v)(SregP s)(aregp a)))
 )
 
%----------------------------------------------------------------------%
% (VgetElement (vreg 3) n)
% get element n of v3 (function value)
%----------------------------------------------------------------------%
 
(put 'VgetElement  'openfn '(vector2valuepat *vgetel))
 
(de *vgetel (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vgetel))
 
(defcmacro *vgetel
    ((numberp anyp sregp)
                         (*move argtwo (reg a6))
                         (SmoveV argthree (vregname argone) (reg A6) ))
    ((numberp anyp anyp)
                         (*move argtwo (reg a6))
                         (SmoveV (reg S6) (vregname argone) (reg A6))
                         (*move (reg s6) argthree))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     SmoveV  (s v a)                   % Si       Vj,Ak
  ((and (Vregp v) (Aregp a) (Sregp s))
              (op-i-j-k 8#076 (Sregp s)(VregP v)(aregp a)))
 )
 
%----------------------------------------------------------------------%
% (Vset (vreg 3) x)
% set all elements of v3 to value x
%----------------------------------------------------------------------%
 
(put 'Vset  'openfn '(vector2pat *vset))
 
(de *vset (a1 a2) (expand2operandcmacro a1 a2 '*vset))
 
(defcmacro *vset
    ((zerop anysregbuts7p)
                         (*move (wconst -1)(reg s7))
                         (*VsetVm (reg s7))
                         (VmergeSV (vregname argone) argtwo (reg V7)) )
    ((numberp anysregbuts7p)
                         (*move (wconst -1) (reg s7))
                         (*VsetVm (reg s7))
                         (VmergeSV (vregname argone) argtwo (reg V0)) )
    ((numberp anyp)
                         (*move argtwo (reg s6))
                         (*vset argone (reg s6)))
    (                    (*cerror "weird *vset")))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  INTEGER arithmetic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
%----------------------------------------------------------------------%
% (vpluss (vreg 3) (vreg 5) x)
% vector summ: v3 = v5 +  scalar x
%----------------------------------------------------------------------%
 
(put 'vpluss 'openfn '(vector3pat *vpluss))
 
(de *vpluss (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vpluss))
 
(defcmacro *vpluss
    ((numberp numberp sregp)
                (VaddS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (VaddS (vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VaddS (vregname argone)(reg s6)(vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VaddS (vregname argone)(reg s6)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VaddS   (vi s vk)                 % vi       Sj-Vk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#154 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (vplusV (vreg 3) (vreg 5) (vreg 7))
% vector summ: v3 = v5 + v7
%----------------------------------------------------------------------%
 
(put 'vplusv 'openfn '(vector3pat *vplusv))
 
(de *vplusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vplusv))
 
(defcmacro *vplusv
    ((numberp numberp numberp)
       (VaddV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VaddV   (vi vj vk)                 % vi       Vj+Vk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#155 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
 
%----------------------------------------------------------------------%
% (vsminusv (vreg 3) x (vreg 5))
% vector difference: v3 = scalar x - V5
%----------------------------------------------------------------------%
 
(put 'vSminusV 'openfn '(vector3pat *sminusv))
 
(de *sminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*sminusv))
 
(defcmacro *sminusv
    ((numberp sregp numberp)
                (VdiffS (vregname argone) argtwo(vregname argthree)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VdiffS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VdiffS  (vi s vk)                 % vi       Sj-Vk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#156 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
%----------------------------------------------------------------------%
% (minusv (vreg 3) (vreg 5))
% calculate : v3 =  - V5
%----------------------------------------------------------------------%
 
(put 'minusV 'openfn '(vector2pat *minusv))
 
(de *minusv (a1 a2) (expand2operandcmacro a1 a2  '*minusv))
 
(defcmacro *minusv
    ((numberp numberp)
                (VdiffS (vregname argone) (Reg S0)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VminusV (vreg 3) (vreg 5) (vreg 7))
% vector difference: v3 = v5 - v7
%----------------------------------------------------------------------%
 
(put 'vminusv 'openfn '(vector3pat *vminusv))
 
(de *vminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vminusv))
 
(defcmacro *vminusv
    ((numberp numberp numberp)
      (VminusV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VminusV  (vi vj vk)                 % vi       Vj+Vk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#157 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  REAL arithmetic (float)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
%----------------------------------------------------------------------%
% (vfpluss (vreg 3) (vreg 5) x)
% vector summ: v3 = v5 +  scalar x
%----------------------------------------------------------------------%
 
(put 'vfpluss 'openfn '(vector3pat *vfpluss))
 
(de *vfpluss (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vfpluss))
 
(defcmacro *vfpluss
    ((numberp numberp sregp)
                (vfaddS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (vfaddS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VFaddS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argwo (reg s6))
                (VFaddS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFaddS  (vi s vk)                 % vi       Sj+FVk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#170 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (VFplusV (vreg 3) (vreg 5) (vreg 7))
% vector summ: v3 = v5 + v7
%----------------------------------------------------------------------%
 
(put 'VFplusv 'openfn '(vector3pat *VFplusv))
 
(de *VFplusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFplusv))
 
(defcmacro *VFplusv
    ((numberp numberp numberp)
       (VFaddV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFaddV  (vi vj vk)                 % vi       Vj+FVk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#171 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
 
%----------------------------------------------------------------------%
% (VFsminusv (vreg 3) x (vreg 5))
% vector difference: v3 = scalar x - V5
%----------------------------------------------------------------------%
 
(put 'VFSminusV 'openfn '(vector3pat *VFsminusv))
 
(de *VFsminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFsminusv))
 
(defcmacro *VFsminusv
    ((numberp sregp numberp)
                (VFdiffS (vregname argone) argtwo(vregname argthree)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VFdiffS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFdiffS (vi s vk)                 % vi       Sj-FVk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#172 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
%----------------------------------------------------------------------%
% (Fminusv (vreg 3) (vreg 5))
% calculate : v3 =  - V5
%----------------------------------------------------------------------%
 
(put 'FminusV 'openfn '(vector2pat *Fminusv))
 
(de *Fminusv (a1 a2) (expand2operandcmacro a1 a2  '*Fminusv))
 
(defcmacro *Fminusv
    ((numberp numberp)
                (VFdiffS (vregname argone) (Reg S0)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
%----------------------------------------------------------------------%
% (VFminusV (vreg 3) (vreg 5) (vreg 7))
% vector difference: v3 = v5 - v7
%----------------------------------------------------------------------%
 
(put 'VFminusv 'openfn '(vector3pat *VFminusv))
 
(de *VFminusv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFminusv))
 
(defcmacro *VFminusv
    ((numberp numberp numberp)
      (VFminusV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFminusV (vi vj vk)                 % vi       Vj-FVk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#173 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (vftimess (vreg 3) (vreg 5) x)
% vector summ: v3 = v5 +  scalar x
%----------------------------------------------------------------------%
 
(put 'vftimess 'openfn '(vector3pat *vftimess))
 
(de *vftimess (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vftimess))
 
(defcmacro *vftimess
    ((numberp numberp sregp)
                (vftimesS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (vftimesS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VFtimesS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VFtimesS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFtimesS (vi s vk)                % vi       Sj*FVk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#160 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (VFtimesV (vreg 3) (vreg 5) (vreg 7))
% vector product: v3 = v5 * v7
%----------------------------------------------------------------------%
 
(put 'VFtimesv 'openfn '(vector3pat *VFtimesv))
 
(de *VFtimesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFtimesv))
 
(defcmacro *VFtimesv
    ((numberp numberp numberp)
       (VFtimesV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFtimesV (vi vj vk)                % vi       Vj*FVk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#161 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (vfhtimess (vreg 3) (vreg 5) x)
% vector half times: v3 = v5 *  scalar x
%----------------------------------------------------------------------%
 
(put 'vfhtimess 'openfn '(vector3pat *vfhtimess))
 
(de *vfhtimess (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vfhtimess))
 
(defcmacro *vfhtimess
    ((numberp numberp sregp)
               (vfhtimesS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
               (vfhtimesS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
               (*move argthree (reg s6))
               (VFhtimesS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
               (*move argtwo (reg s6))
               (VFhtimesS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFhtimesS (vi s vk)               % vi       Sj*HVk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#162 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (VFhtimesV (vreg 3) (vreg 5) (vreg 7))
% vector half product: v3 = v5 * v7
%----------------------------------------------------------------------%
 
(put 'VFhtimesv 'openfn '(vector3pat *VFhtimesv))
 
(de *VFhtimesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFhtimesv))
 
(defcmacro *VFhtimesv
    ((numberp numberp numberp)
       (VFhtimesV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFhtimesV (vi vj vk)               % vi       Vj*FVk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#163 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
%----------------------------------------------------------------------%
% (vrtimess (vreg 3) (vreg 5) x)
% vector prod rounded: v3 = v5 *  scalar x
%----------------------------------------------------------------------%
 
(put 'vrtimess 'openfn '(vector3pat *vrtimess))
 
(de *vrtimess (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vrtimess))
 
(defcmacro *vrtimess
    ((numberp numberp sregp)
                (vrtimesS (vregname argone) argthree(vregname argtwo)))
    ((numberp sregp numberp)
                (vrtimesS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (vrtimesS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (vrtimesS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     vrtimesS (vi s vk)                % vi       Sj*RVk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#164 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (vrtimesV (vreg 3) (vreg 5) (vreg 7))
% vector rounded product: v3 = v5 * v7
%----------------------------------------------------------------------%
 
(put 'vrtimesv 'openfn '(vector3pat *vrtimesv))
 
(de *vrtimesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vrtimesv))
 
(defcmacro *vrtimesv
    ((numberp numberp numberp)
       (vrtimesV (vregname argone) (vregname argthree)(vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     vrtimesV (vi vj vk)                % vi       Vj*RVk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#165 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
 
%----------------------------------------------------------------------%
% (VFrezappV (vreg 3) (vreg 5))
% vector reziprocal approximantion: v3 = rez (v5)
%----------------------------------------------------------------------%
 
(put 'VFrezappv 'openfn '(vector2pat *VFrezappv))
 
(de *VFrezappv (a1 a2) (expand2operandcmacro a1 a2 '*VFrezappv))
 
(defcmacro *VFrezappv
    ((numberp numberp numberp)
       (VFrezappV (vregname argone) (vregname argtwo)                   ))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFrezappV (vi vj)                  % vi     /HVj
  ((and (vregp vi)(vregP vj))
              (op-i-j-k 8#174 (vregp vi)(vregp vj)0))
 )
 
%----------------------------------------------------------------------%
% (VpopcV (vreg 3) (vreg 5))
% vector popuplation count : v3 = p (v5)
%----------------------------------------------------------------------%
 
(put 'VFpopcv 'openfn '(vector2pat *VFpopcv))
 
(de *VFpopcv (a1 a2) (expand2operandcmacro a1 a2 '*VFpopcv))
 
(defcmacro *VFpopcv
    ((numberp numberp)
       (VFpopcV (vregname argone) (vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFpopcV (vi vj)                    % vi     pVj
  ((and (vregp vi)(vregP vj))
              (op-i-j-k 8#174 (vregp vi)(vregp vj) 1))
 )
 
 
%----------------------------------------------------------------------%
% (VpopcpV (vreg 3) (vreg 5))
% vector popuplation count parity: v3 = q (v5)
%----------------------------------------------------------------------%
 
(put 'VFpopcpv 'openfn '(vector2pat *VFpopcpv))
 
(de *VFpopcpv (a1 a2) (expand2operandcmacro a1 a2 '*VFpopcpv))
 
(defcmacro *VFpopcpv
    ((numberp numberp)
       (VFpopcpV (vregname argone) (vregname argtwo)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VFpopcpV (vi vj)                   % vi     qVj
  ((and (vregp vi)(vregP vj))
              (op-i-j-k 8#174 (vregp vi)(vregp vj) 2))
 )
 
 
%----------------------------------------------------------------------%
% (vfrecits (vreg 3) x (vreg 5))
% vector reciprocal iteration: v3 = scalar x *I v5
%----------------------------------------------------------------------%
 
(put 'vfrecits 'openfn '(vector3pat *vfrecits))
 
(de *vfrecits (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vfrecits))
 
(defcmacro *vfrecits
    ((numberp numberp sregp)
                (vfrecitS (vregname argone) argthree (vregname argtwo)))
    ((numberp sregp numberp)
                (vfrecitS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VFrecitS(vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VFrecitS(vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr vfrecits" )))
 
 (DefVopCode     VFrecitS (vi s vk)                % vi       Sj*IVk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#166 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (VFrecitV (vreg 3) (vreg 5) (vreg 7))
% vector reciprocal iteration: v3 = v5 *I v7
%----------------------------------------------------------------------%
 
(put 'VFrecitv 'openfn '(vector3pat *VFrecitv))
 
(de *VFrecitv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VFrecitv))
 
(defcmacro *VFrecitv
    ((numberp numberp numberp)
       (VFrecitV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector instr VFrecitV")))
 
 (DefVopCode     VFrecitV (vi vj vk)                % vi       Vj*IVk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#167 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  LOGICAL operations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%----------------------------------------------------------------------%
% (VandS (vreg 3) x (vreg 7))
% vector and: v3 = x & v7
%----------------------------------------------------------------------%
 
(put 'VandS 'openfn '(vector3pat *VandS))
 
(de *VandS (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VandS))
 
(defcmacro *VandS
    ((numberp sregp numberp)
      (VandS (vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
      (*Move argtwo (reg S6))
      (VandS (vregname argone) (Reg S6) (vregname argthree)))
    (                    (*cerror "weird vector &")))
 
 (DefVopCode     VandS  (vi sj vk)                 % Vi       Sj&Vk
  ((and (vregp vi)(sregP sj)(Vregp vk))
              (op-i-j-k 8#140 (vregp vi)(sregp sj)(Vregp vk)))
 )
 
 
 
%----------------------------------------------------------------------%
% (VandV (vreg 3) (vreg 5) (vreg 7))
% vector and: v3 = v5 & v7
%----------------------------------------------------------------------%
 
(put 'vandv 'openfn '(vector3pat *vandv))
 
(de *vandv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vandv))
 
(defcmacro *vandv
    ((numberp numberp numberp)
      (VandV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector !")))
 
 (DefVopCode     VandV  (vi vj vk)                 % Vi       Vj!Vk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#141 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
%----------------------------------------------------------------------%
% (VorS (vreg 3) x (vreg 7))
% vector ohr: v3 = x ! v7
%----------------------------------------------------------------------%
 
(put 'VorS 'openfn '(vector3pat *VorS))
 
(de *VorS (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*VorS))
 
(defcmacro *VorS
    ((numberp sregp numberp)
      (VorS (vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
      (*Move argtwo (reg S6))
      (VorS (vregname argone) (Reg S6) (vregname argthree)))
    (                    (*cerror "weird vector &")))
 
 (DefVopCode     VorS   (vi sj vk)                 % Vi       Sj!Vk
  ((and (vregp vi)(sregP sj)(Vregp vk))
              (op-i-j-k 8#142 (vregp vi)(sregp sj)(Vregp vk)))
 )
 
%----------------------------------------------------------------------%
% (VmoveV (vreg 3)  (vreg 7))
% move V7 to V3
% special case of vector ohr
%----------------------------------------------------------------------%
 
(put 'VmoveV 'openfn '(vector2pat *VmoveV))
 
(de *VmoveV (a1 a3) (expand3operandcmacro a1 '(Reg S0) a3 '*VorS))
 
(defCmacro *VmoveV)
 
 
%----------------------------------------------------------------------%
% (VorV (vreg 3) (vreg 5) (vreg 7))
% vector or: v3 = v5 ! v7
%----------------------------------------------------------------------%
 
(put 'vorv 'openfn '(vector3pat *vorv))
 
(de *vorv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vorv))
 
(defcmacro *vorv
    ((numberp numberp numberp)
      (VorV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector &")))
 
 (DefVopCode     VorV   (vi vj vk)                 % Vi       Vj\Vk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#143 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (VxorV (vreg 3) (vreg 5) (vreg 7))
% vector xor: v3 = v5 \ v7
%----------------------------------------------------------------------%
 
(put 'vxorv 'openfn '(vector3pat *vxorv))
 
(de *vxorv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vxorv))
 
(defcmacro *vxorv
    ((numberp numberp numberp)
      (VxorV (vregname argone) (vregname argtwo)(vregname argthree)))
    (                    (*cerror "weird vector \")))
 
 (DefVopCode     VxorV  (vi vj vk)                 % Vi       Vj&Vk
  ((and (vregp vi)(vregP vj)(Vregp vk))
              (op-i-j-k 8#145 (vregp vi)(vregp vj)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% special form of VxorV:
% (Vclear (vreg 3))
% clear vector register
%----------------------------------------------------------------------%
 
(put 'vclear 'openfn '(vector1pat *vclear))
 
(de *vclear (a1) (expand3operandcmacro a1 a1 a1 '*vxorv))
 
(defcmacro *vclear)
 
%----------------------------------------------------------------------%
% (vlminuss (vreg 3) (vreg 5) x)
% vector logical difference: v3 = v5 minus scalar x
%----------------------------------------------------------------------%
 
(put 'vlminuss 'openfn '(vector3pat *vlminuss))
 
(de *vlminuss (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vlminuss))
 
(defcmacro *vlminuss
    ((numberp numberp sregp)
                (VldiffS (vregname argone) argthree (vregname argtwo)))
    ((numberp sregp numberp)
                (VldiffS (vregname argone) argtwo (vregname argthree)))
    ((numberp numberp anyp)
                (*move argthree (reg s6))
                (VldiffS (vregname argone)(reg s6)(vregname argtwo)))
    ((numberp anyp numberp)
                (*move argtwo (reg s6))
                (VldiffS (vregname argone)(reg s6)(vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
 (DefVopCode     VldiffS  (vi s vk)                % vi       Sj-Vk
  ((and (vregp vi)(SregP s)(Vregp vk))
              (op-i-j-k 8#144 (vregp vi)(Sregp s)(Vregp vk)))
 )
 
%----------------------------------------------------------------------%
% (vmergevv i j k)
% vector merge: vi = vj (if vm=1) or vk
%----------------------------------------------------------------------%
 
(put 'vmergevv 'openfn '(vector3pat *vmergevv))
 
(de *vmergevv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vmergevv))
 
(defcmacro *vmergevv
    ((numberp numberp numberp)
      (Vmergevv(vregname argone) (vregname argtwo) (vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     Vmergevv (vi vj vk)                 % Vi  Vj!Vk&VM
  ((and (vregp vi)(VregP vj)(Vregp vk))
              (op-i-j-k 8#147 (vregp vi)(Vregp vj)(Vregp vk)))
 )
 
 
%----------------------------------------------------------------------%
% (vmergesv i j k)
% vector merge: vi = sj (if vm=1) or vk
%----------------------------------------------------------------------%
 
(put 'vmergesv 'openfn '(vector3pat *vmergesv))
 
(de *vmergesv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vmergesv))
 
(defcmacro *vmergesv
    ((numberp sregp numberp)
         (Vmergesv(vregname argone) argtwo (vregname argthree)))
    ((numberp anyp numberp)
         (*move argtwo (reg s6))
         (Vmergesv(vregname argone) (Reg S6) (vregname argthree)))
    (                    (*cerror "weird vector instr")))
 
 
 (DefVopCode     VmergeSV(v s vv)                   % Vi   Sj!Vk&VM
  ((and (Vregp v) (Sregp s) (Vregp vv))
              (op-i-j-k 8#146 (Vregp v)(SregP s)(Vregp vv)))
 )
 
 
%----------------------------------------------------------------------%
% (VlshiftV (vreg 3) (vreg 5) 17)
% vector lshift: v3 = v5 < 17
%----------------------------------------------------------------------%
 
(put 'vlshiftv 'openfn '(vector3pat *vlshiftv))
 
(de *vlshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vlshiftv))
 
(defcmacro *vlshiftv
    ((numberp numberp onep)
      (VlshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VlshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VlshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector <")))
 
 (DefVopCode     VlshiftV (vi vj ak)               % Vi       Vj<ak
  ((and (vregp vi)(vregP vj)(Aregp ak))
              (op-i-j-k 8#150 (vregp vi)(vregp vj)(Aregp ak)))
 )
 
%----------------------------------------------------------------------%
% (VrshiftV (vreg 3) (vreg 5) 17)
% vector rshift: v3 = v5 > 17
%----------------------------------------------------------------------%
 
(put 'vrshiftv 'openfn '(vector3pat *vrshiftv))
 
(de *vrshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vrshiftv))
 
(defcmacro *vrshiftv
    ((numberp numberp onep)
      (VrshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VrshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VrshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector >")))
 
 (DefVopCode     VrshiftV (vi vj ak)               % Vi       Vj<Ak
  ((and (vregp vi)(vregP vj)(Aregp ak))
              (op-i-j-k 8#151 (vregp vi)(vregp vj)(Aregp ak)))
 )
 
 
 
%----------------------------------------------------------------------%
% (VdlshiftV (vreg 3) (vreg 5) 17)
% vector dlshift: v3 = v5 << 17
%----------------------------------------------------------------------%
 
(put 'vdlshiftv 'openfn '(vector3pat *vdlshiftv))
 
(de *vdlshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vdlshiftv))
 
(defcmacro *vdlshiftv
    ((numberp numberp onep)
      (VdlshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VdlshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VdlshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector <")))
 
 (DefVopCode     VdlshiftV (vi vj ak)              % Vi       Vj,Vj<Ak
  ((and (vregp vi)(vregP vj)(Aregp ak))
              (op-i-j-k 8#152 (vregp vi)(vregp vj)(Aregp ak)))
 )
 
%----------------------------------------------------------------------%
% (VdrshiftV (vreg 3) (vreg 5) 17)
% vector drshift: v3 = v5 >> 17
%----------------------------------------------------------------------%
 
(put 'vdrshiftv 'openfn '(vector3pat *vdrshiftv))
 
(de *vdrshiftv (a1 a2 a3) (expand3operandcmacro a1 a2 a3 '*vdrshiftv))
 
(defcmacro *vdrshiftv
    ((numberp numberp onep)
      (VdrshiftV (vregname argone) (vregname argtwo) (Reg A0)))
    ((numberp numberp aregp)
      (VdrshiftV (vregname argone) (vregname argtwo) argthree))
    ((numberp numberp anyp)
      (*move argthree (Reg A6))
      (VdrshiftV (vregname argone) (vregname argtwo)(Reg A6)))
    (                    (*cerror "weird vector >")))
 
 (DefVopCode     VdrshiftV (vi vj ak)              % Vi       Vj<Ak
  ((and (vregp vi)(vregP vj)(Aregp ak))
              (op-i-j-k 8#153 (vregp vi)(vregp vj)(Aregp ak)))
 )
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% administration of vectors
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%----------------------------------------------------------------------%
% (vsetvl x)
% set vector length register
%----------------------------------------------------------------------%
 
(put 'vsetvl  'openfn '(vector1pat *vsetvl))
 
(de *vsetvl  (a1) (expand1operandcmacro a1 '*vsetvl))
 
(defcmacro *vsetvl
    ((aregp)
                (VLmoveA (reg a0) argone)) % a0 is dummy to mark argone
                                           % as source of move(lapopt)
 
    ((anyp)
                (*move argone (reg a6))
                (VLmoveA (reg a0) (reg a6)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VLmoveA (a b)                     % VL       Ak
  ((and (anyp a) (AregP b)) (op-i-j-k 8#002 0 0 (AregP b)))
 )
 
 
%----------------------------------------------------------------------%
% (vsetvm x)
% set vector mask register
%----------------------------------------------------------------------%
 
(put 'vsetvm  'openfn '(vector1pat *vsetvm))
 
(de *vsetvm  (a1) (expand1operandcmacro a1 '*vsetvm))
 
(defcmacro *vsetvm
    ((Sregp)
                (vmmoveS (reg a0) argone))
    ((anyp)
                (*move argone (reg S6))
                (vmmoveS (reg a0) (reg S6)))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     vmmoveS (a s)                      % VM       Sj
  ((and (anyp a) (SregP s))  (op-i-j-k 8#003 0 (SregP s) 0))
 )
 
%----------------------------------------------------------------------%
% (vsetvlsave vreg x)
% set vector length register to x and save for specific vreg
%----------------------------------------------------------------------%
 
(put 'vsetvlsave  'openfn '(vector2pat *vsetvlsave))
 
(de *vsetvlsave  (a1 a2) (expand2operandcmacro a1 a2 '*vsetvlsave))
 
(defcmacro *vsetvlsave
    ((anyp aregp)
                (VLmoveA (reg a0) argtwo)
                (move (vlthname argone) argtwo)
                (*move argtwo (reg B21))  )    % sim vl
    (%(anyp anyp)
                (*move argtwo (reg a6))
                (VLmoveA (reg a0) (reg a6))
                (move (vlthname argone) (reg a6))
                (*move (reg a6)(reg b21))  )  )
%----------------------------------------------------------------------%
% savevl / restorevl
%----------------------------------------------------------------------%
(put 'Vsavevl 'opencode '((*move (reg b21)(reg a6))
                          (*move (reg a6)(reg b22))))
(put 'Vrestorevl 'opencode '(
                         (*move (reg b22)(reg a6))
                         (*move (reg a6)(reg b21))
                         (VLmoveA (reg a0) (reg a6)) ))
%----------------------------------------------------------------------%
% read vector mask register
%----------------------------------------------------------------------%
 
 (DefVopCode     SmoveVM (s)                       % Si        VM
  ((SregP s)  (op-i-j-k 8#073 (Sregp s) 0 0))
 )
 
%----------------------------------------------------------------------%
% (VtestZ (vreg 3) )
% test  zeroes in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestz  'openfn '(vector1valuepat *vtestz))
 
(de *vtestz  (a1 a2) (expand2operandcmacro a1  a2 '*vtestz))
 
(defcmacro *vtestz
    ((numberp sregp)
                (VMtestZ (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestZ (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VMtestZ (v)                       % VM        Vj,Z
  ((Vregp v)  (op-i-j-k 8#175 0 (Vregp v) 0))
 )
 
%----------------------------------------------------------------------%
% (VtestN (vreg 3) )
% test  NonZeroes in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestn  'openfn '(vector1valuepat *vtestn))
 
(de *vtestn  (a1 a2) (expand2operandcmacro a1  a2 '*vtestn))
 
(defcmacro *vtestn
    ((numberp sregp)
                (VMtestN (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestN (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     VMtestN (v)                       % VM        Vj,N
  ((Vregp v)  (op-i-j-k 8#175 0 (Vregp v) 1))
 )
 
 
%----------------------------------------------------------------------%
% (VtestPos (vreg 3) )
% test  positives in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestpos  'openfn '(vector1valuepat *vtestpos))
 
(de *vtestpos  (a1 a2) (expand2operandcmacro a1  a2 '*vtestpos))
 
(defcmacro *vtestpos
    ((numberp sregp)
                (VMtestPos (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestPos (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
     )
 
 (DefVopCode     VMtestPos (v)                       % VM      Vj,P
  ((Vregp v)  (op-i-j-k 8#175 0 (Vregp v) 2))
 )
 
 
%----------------------------------------------------------------------%
% (VtestNeg (vreg 3) )
% test  negatives in v3 (function value = mask)
%----------------------------------------------------------------------%
 
(put 'vtestneg  'openfn '(vector1valuepat *vtestneg))
 
(de *vtestneg  (a1 a2) (expand2operandcmacro a1  a2 '*vtestneg))
 
(defcmacro *vtestneg
    ((numberp sregp)
                (VMtestNeg (vregname argone))
                (SmoveVM argtwo))
    ((numberp anyp)
                (VMtestNeg (vregname argone))
                (SmoveVM (reg s6))
                (*move (reg s6) argtwo))
     )
 
 (DefVopCode     VMtestNeg (v)                       % VM      Vj,M
  ((Vregp v)  (op-i-j-k 8#175 0 (Vregp v) 3))
 )
 
%----------------------------------------------------------------------%
% (Vldzero x)
% calculate leading zeroes in value x
%----------------------------------------------------------------------%
 
(put 'vldzero 'openfn '(vector1valuepat *vldzero))
 
(de *vldzero (a1 a2) (expand2operandcmacro a1  a2 '*vldzero))
 
(defcmacro *vldzero
    ((sregp anyp)
                (LDzero (reg a6) argone)
                (*move (reg a6) argtwo) )
    ((anyp anyp)
                (*move argone (reg s6))
                (LDzero (reg a6) (reg s6))
                (*move (reg a6) argtwo))
    (                    (*cerror "weird vector instr")))
 
 (DefVopCode     LDzero (a s)                      % Ai        ZSj
  ((and (Sregp s) (Aregp a))
              (op-i-j-k 8#027 (AregP a)(Sregp s) 0))
 )
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  magic numbers for arithmetic
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(put 'magicSconst06 'opencode '((sconst06 (reg 1))))
(put 'magicSconst04 'opencode '((sconst04 (reg 1))))
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  definition of high level vector macros
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
% allow the writing of (vreg 3) for a vector register
 
(ds vreg (n) n)
 
% allow the writing (Vreg3) too
 
(ds Vreg0 () 0)
(ds Vreg1 () 1)
(ds Vreg2 () 2)
(ds Vreg3 () 3)
(ds Vreg4 () 4)
(ds Vreg5 () 5)
(ds Vreg6 () 6)
(ds Vreg7 () 7)
 
% VsearchVeqS:
%
% find first eq-occurence of a scalar in a vector register
% v object vector register
% s scalar value
% vaux1: auxiliary vector register
 
(ds VsearchVeqS (v s vaux1)
   (prog(u)
     (vSminusv vaux1 s v)       % vaux = s - v  arithmetic difference
     (setq u (vtestz vaux1))   % test vaux to zero
     (return
         (cond ((eq u 0) -1)   % not found: -1
               (t (vldzero u)) % found: calculate position number
))))
