%
% DATA-MACHINE.RED - Macros for fast access to data structures
%
% Author:      Eric Benson
%              Symbolic Computation Group
%              Computer Science Dept.
%              University of Utah
% Date:        5 April 1982
% Copyright (c) 1982 University of Utah
%

% Primitives handled by the compiler are BYTE, PUTBYTE, GETMEM, PUTMEM,
% MKITEM, FIELD, SIGNEDFIELD, PUTFIELD, HALFWORD, PUYTHALFWORD

(on syslisp)

(fluid '(wplus2optimizationfn*))

(off r2i)

% These definitions are for interpretive testing of Syslisp code.
% They may be dangerous in some cases.

(commentoutcode (progn 
(de byte (waddr byteoffset) 
    (byte waddr byteoffset)) 

(de putbyte (waddr byteoffset val) 
    (putbyte waddr byteoffset val)) 

(de halfword (waddr halfwordoffset) 
    (halfword waddr halfwordoffset)) 

(de puthalfword (waddr halfwordoffset val) 
    (puthalfword waddr halfwordoffset val)) 

(de getmem (addr) 
    (getmem addr)) 

(de putmem (addr val) 
    (putmem addr val)) 

(de mkitem (tagpart infpart) 
    (mkitem tagpart infpart)) 

(commentoutcode (progn                  % can't do FIELD w/ non constants
(de field (cell startingbit bitlength) 
    (field cell startingbit bitlength)) 

(de signedfield (cell startingbit bitlength) 
    (signedfield cell startingbit bitlength)) 

(de putfield (cell startingbit bitlength val) 
    (putfield cell startingbit bitlength val)) nil)) 


(de wplus2 (r1 r2) (wplus2 
    r1 r2)) 

(de wdifference (r1 r2) (wdifference 
    r1 r2)) 

(de wtimes2 (r1 r2) (wtimes2 
    r1 r2)) 

(de wquotient (r1 r2) (wquotient 
    r1 r2)) 

(de wremainder (r1 r2) (wremainder 
    r1 r2)) 

(de wminus (r1) (wminus 
    r1)) 

(de wshift (r1 r2) (wshift 
    r1 r2)) 

(de wand (r1 r2) (wand 
    r1 r2)) 

(de wor (r1 r2) (wor 
    r1 r2)) 

(de wxor (r1 r2) (wxor 
    r1 r2)) 

(de wnot (r1) (wnot 
    r1)) 

(de wlessp (r1 r2) (wlessp 
    r1 r2)) 

(de wgreaterp (r1 r2) (wgreaterp 
    r1 r2)) 

(de wleq (r1 r2) (wleq 
    r1 r2)) 

(de wgeq (r1 r2) (wgeq 
    r1 r2)) nil))


(on r2i)

(off syslisp)

% SysLisp array accessing primitives

(dm wgetv (u) 
    (list 'getmem (list wplus2optimizationfn* 
                        (cadr u) (list 'wtimes2 (caddr u) '(wconst 
                                           addressingunitsperitem)))))

(dm wputv (u) 
   (list 'putmem (list wplus2optimizationfn* 
                       (cadr u) 
                       (list 'wtimes2 (caddr u) '(wconst 
                                      addressingunitsperitem))) 
                 (cadddr u)))

% tags

(compiletime (progn 
(de declaretagrange (namelist startingvalue increment) 
(prog (result) 
    (setq result (list 'progn)) 
    (while namelist 
    (progn (setq result 
                             
                             
                  (cons (list 'put (mkquote (car namelist)) ''wconst startingvalue) result)) 
        (setq startingvalue (plus startingvalue increment)) 
        (setq namelist (cdr namelist)))) 
    (return (reversip result)))) 


(dm lowtags (u) 
    (declaretagrange (cdr u) 0 1)) 

(dm hightags (u) 
    (declaretagrange (cdr u) (difference (lsh 1 (get 'tagbitlength 'wconst)) 1) -1)) nil))


(lowtags posint fixn bign fltn str bytes halfwords wrds vect pair 
        evect)

(put 'code 'wconst 15)

(hightags negint id unbound btrtag forward 
         hvect hwrds hhalfwords hbytes 
         hbpsitems)                     % July 16, 1984 /JLM/ (to identify code&data blocks)

% Item constructor macros

(de makeitemconstructor (tagpart infpart) 
    (list 'mkitem tagpart infpart))

(dm mkbtr (u) 
    (makeitemconstructor '(wconst btrtag) (cadr u)))

(dm mkid (u) 
    (makeitemconstructor '(wconst id) (cadr u)))

(dm mkfixn (u) 
    (makeitemconstructor '(wconst fixn) (cadr u)))

(dm mkfltn (u) 
    (makeitemconstructor '(wconst fltn) (cadr u)))

(dm mkbign (u) 
    (makeitemconstructor '(wconst bign) (cadr u)))

(dm mkpair (u) 
    (makeitemconstructor '(wconst pair) (cadr u)))

(dm mkvec (u) 
    (makeitemconstructor '(wconst vect) (cadr u)))

(dm mkevect (u) 
    (makeitemconstructor '(wconst evect) (cadr u)))

(dm mkwrds (u) 
    (makeitemconstructor '(wconst wrds) (cadr u)))

(dm mkstr (u) 
    (makeitemconstructor '(wconst str) (cadr u)))

(dm mkbytes (u) 
    (makeitemconstructor '(wconst bytes) (cadr u)))

(dm mkhalfwords (u) 
    (makeitemconstructor '(wconst halfwords) (cadr u)))

(dm mkcode (u) 
    (makeitemconstructor '(wconst code) (cadr u)))

% Access to tag (type indicator) of Lisp item in ordinary code

(dm tag (u) 
    (list 'field (cadr u) '(wconst tagstartingbit) '(wconst tagbitlength)))


% Access to info field of item (pointer or immediate operand)

(dm inf (u) 
    (list 'field (cadr u) '(wconst infstartingbit) '(wconst infbitlength)))

(dm putinf (u) 
    (list 'putfield (cadr u) '(wconst infstartingbit) '(wconst 
                            infbitlength) (caddr u)))

(foreach x in '(idinf strinf vecinf evecinf pairinf wrdinf halfwordinf 
                codeinf fixinf fltinf biginf) do 
    (putd x 'macro (cdr (getd 'inf))))

(foreach x in '(putidinf putstrinf putvecinf putpairinf putwrdinf 
                putevecinf puthalfwordinf 
                putfixinf putfltinf putbiginf) do 
    (putd x 'macro (cdr (getd 'putinf))))

% IntInf is no longer needed, will be a macro no-op
% for the time being

(remprop 'intinf 'openfn)

(dm intinf (u) 
    (cadr u))

% Similarly for MkINT

(dm mkint (u) 
    (cadr u))

% # of words in a pair

(dm pairpack (u) 2)
    

% length (in characters, words, etc.) of a string, vector, or whatever,
% stored in the first word pointed to

(dm getlen (u) 
    (list 'signedfield (list 'getmem (cadr u)) '(wconst infstartingbit) '(wconst 
                                              infbitlength)))

(dm strbase (u)                         % point to chars of string
    (list wplus2optimizationfn* (cadr u) '(wconst addressingunitsperitem)))

% chars string length --> words string length

% Note that StrPack and HalfWordPack do not include the header word,
% VectPack and WrdPack do.

(dm strpack (u) 
    (list 'wquotient (list wplus2optimizationfn* (cadr u) 
                   (list wplus2optimizationfn* '(wconst charactersperword) 1)) '(wconst 
                                                 
                     charactersperword)))

% access to bytes of string; skip first word


% 7/30/84 - IMSSS - Changed to accomodate addressing arithmetic
%        All occurrences of WPlus2OptimizationFn!* here were Plus2 before

(dm strbyt (u) 
 (list 'byte 
       (list wplus2optimizationfn* (cadr u) '(wconst addressingunitsperitem)) 
       (caddr u)))

(dm putstrbyt (u) 
 (list 'putbyte 
       (list wplus2optimizationfn* (cadr u) '(wconst addressingunitsperitem)) 
       (caddr u) (cadddr u)))

% access to halfword entries; skip first word

(dm halfworditm (u) 
 (list 'halfword 
       (list wplus2optimizationfn* (cadr u) '(wconst addressingunitsperitem)) 
       (caddr u)))

(dm puthalfworditm (u) 
 (list 'puthalfword 
       (list wplus2optimizationfn* (cadr u) '(wconst addressingunitsperitem)) 
       (caddr u) (cadddr u)))

% halfword length --> words  length

(dm halfwordpack (u) 
    (list wplus2optimizationfn* (list 'wshift (cadr u) -1) 1))



% length (in Item size quantities) of Lisp vectors

% size of Lisp vector in words

(dm vectpack (u) 
    (list wplus2optimizationfn* (cadr u) 1))

% size of Lisp Evector in words

(dm evectpack (u) 
    (list wplus2optimizationfn* (cadr u) 1))

% access to elements of Lisp vector

(dm vecitm (u) 
    (list 'wgetv (cadr u) 
                 (list wplus2optimizationfn* (caddr u) 1)))

(dm putvecitm (u) 
    (list 'wputv (cadr u) 
                 (list wplus2optimizationfn* (caddr u) 1) 
                 (cadddr u)))

% access to elements of Lisp Evector

(dm evecitm (u) 
    (list 'wgetv (cadr u) 
                 (list wplus2optimizationfn* (caddr u) 1)))

(dm putevecitm (u) 
    (list 'wputv (cadr u) 
                 (list wplus2optimizationfn* (caddr u) 1) 
                 (cadddr u)))


% Wrd is like Vect, but not traced by the garbage collector

(dm wrdpack (u) 
    (list wplus2optimizationfn* (cadr u) 1))

(foreach x in '(strlen bytelen veclen eveclen wrdlen halfwordlen) do 
    (putd x 'macro (cdr (getd 'getlen))))

(putd 'wrditm 'macro (cdr (getd 'vecitm)))

(putd 'putwrditm 'macro (cdr (getd 'putvecitm)))

(dm fixval (u) 
    (list 'wgetv (cadr u) 1))

(dm putfixval (u) 
    (list 'wputv (cadr u) 1 (caddr u)))


(dm floatbase (u) 
    (list wplus2optimizationfn* (cadr u) '(wconst addressingunitsperitem)))

(dm floathighorder (u) 
    (list 'wgetv (cadr u) 1))

(dm floatloworder (u) 
    (list 'wgetv (cadr u) 2))


% New addition: A code pointer can have the number of arguments it expects
% stored in the word just before the entry
(dm !%code-number-of-arguments (u) 
    (list 'wgetv (cadr u) -1))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  The four basic cells for each symbol: Val, Nam, Fnc, Prp, corresponding  %
%  to variable value, symbol name (as string), function cell (jump to       %
%  compiled code or lambda linker) and property list (pairs for PUT, GET,   %
%  atoms for FLAG, FLAGP).  These are currently 4 separate arrays, but this %
%  representation may be changed to a contiguous 4 element record for each  %
%  symbol or something else and therefore should not be accessed as arrays. %
%                                                                           %
%  UPDATE -- Dec 9, 1983 -- JLM                                             %
%                                                                           %
%  The old definitions here assume that the array base addresses are        %
%  accessible as constants, which is incorrect for the IBM 370.             %
%                                                                           %
%  On the 370, each syslisp array has a variable in the data area which     %
%  (at runtime) holds the 24-bit base address of the array.  This address   %
%  is not determined until runtime startup, when memory is allocated.       %
%                                                                           %
%  Note that in the 370 version, WArray is a tag returned by !&PaNonLocal,  %
%  is flagged Var, and is defined as an anyreg synonymous with Wvar.        %
%  In particular, (!&VarP '(WArray Foo)) returns T.                         %
%                                                                           %
%  The macros here have thus been changed to reference 'xxx instead of      %
%  '(WConst xxx).  It works fine for both(??) IBM and TENEX(??) versions.   %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(dm symval (u) 
%   list('WGetV, '(WConst SymVal), cadr U);
    (list 'wgetv 'symval (cadr u)))


(dm putsymval (u) 
%   list('WPutV, '(WConst SymVal), cadr U, caddr U);
    (list 'wputv 'symval (cadr u) (caddr u)))


(dm lispvar (u)                         % Access value cell by name
%   list('(WConst SymVal), list('IDLoc, cadr U));
    (list 'symval (list 'idloc (cadr u))))

(dm putlispvar (u) 
    (list 'putsymval (list 'idloc (cadr u)) (caddr u)))
%   no change

(dm symnam (u) 
%   list('WGetV, '(WConst SymNam), cadr U);
    (list 'wgetv 'symnam (cadr u)))

(dm putsymnam (u) 
%   list('WPutV, '(WConst SymNam), cadr U, caddr U);
    (list 'wputv 'symnam (cadr u) (caddr u)))

% Retrieve the address stored in the function cell

% SymFnc and PutSymFnc are not defined portably

(dm symprp (u) 
%   list('WGetV, '(WConst SymPrp), cadr U);
    (list 'wgetv 'symprp (cadr u)))

(dm putsymprp (u) 
%   list('WPutV, '(WConst SymPrp), cadr U, caddr U);
    (list 'wputv 'symprp (cadr u) (caddr u)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Binding stack primitives

(dm bndstkid (u) 
    (list 'wgetv (cadr u) -1))

(dm putbndstkid (u) 
    (list 'wputv (cadr u) -1 (caddr u)))

(dm bndstkval (u) 
    (list 'getmem (cadr u)))

(dm putbndstkval (u) 
    (list 'putmem (cadr u) (caddr u)))

(dm adjustbndstkptr (u) 
    (list wplus2optimizationfn* (cadr u) 
                  (list 'wtimes2 (caddr u) 
                                 (list 'wtimes2 '(wconst 
                                        addressingunitsperitem) 2))))
                                         

% ObArray is a linearly allocated hash table containing ID numbers of entries
% maintained as a circular buffer.  It is referenced only via these macros
% because we may decide to change to some other representation.

(ds obarray (i) 
    (halfword hashtable i))

(ds putobarray (i x) 
    (setf (halfword hashtable i) x))

(put 'obarray 'assign-op 'putobarray)

(ds occupiedslot (u) 
    (greaterp (obarray u) 0))

(deflist '((getmem putmem) 
          (field putfield) 
          (byte putbyte) 
          (halfword puthalfword) 
          (tag puttag) 
          (inf putinf) 
          (idinf putidinf) 
          (strinf putstrinf) 
          (vecinf putvecinf) 
          (wrdinf putwrdinf) 
          (pairinf putpairinf) 
          (fixinf putfixinf) 
          (fixval putfixval) 
          (fltinf putfltinf) 
          (biginf putbiginf) 
          (strlen putstrlen) 
          (strbyt putstrbyt) 
          (veclen putveclen) 
          (vecitm putvecitm) 
          (wrdlen putwrdlen) 
          (wrditm putwrditm) 
          (symval putsymval) 
          (lispvar putlispvar) 
          (symnam putsymnam) 
          (symfnc putsymfnc) 
          (symprp putsymprp) 
          (bndstkid putbndstkid) 
          (bndstkval putbndstkval)) 'assign-op)

% of file
