%
% 370-ASMOUTLAP.RED - part of LAP to assembler translator
%

(fluid '(opcode* operands*              % opcode and operands, used
        operand*1 operand*2 operand*3   %  by instruction-formatting
        operand*4 operand*5 operand*6)) %  templates

(fluid '(lapcode* lapdata* entryname* codeout* expressioncount*))
(fluid '(locallabels*))                 % labels within one function

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%      AsmOutLap -- Used to print an entire function.                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de asmoutlap (u) 
 (prog (locallabels* entryname* oldout fn) 
  (setq lapcode* (pass1lap u))          % Expand cmacros, quoted expressions
  (setq oldout (wrs codeout*)) 
  (resolvebranchesandlabels lapcode*) 
           % sets:  CodeBlockSize!*     DataBlockSize!*
           %        CodeBlockOffsets!*  DataBlockOffsets!*
           % may optimize or lengthen code
           % may detect addressibility problems
  (codeblockheader) 
  (setq fn (cadr (assoc '*entry lapcode*))) 
  (cond ((or (cdddr lapdata*) 
     (not (stringp (cadr lapdata*))) 
     (not (stringp (caddr lapdata*)))) 
   (progn (printf "%f* ---(data block for %w)---%n" fn) 
       (foreach datum in lapdata* do 
           (asmoutinstruction datum)) nil)) (t 	% ignores labels
  (printf "%f* ---(no data block for %w)---%n" fn))) 
  (printf "%f* ---(code block for %w)---%n" fn) 
  (setq codeblockoffsets* nil) 
  (foreach instruction in lapcode* do 
     (cond (instruction 
      (asmoutinstruction instruction)))) 	% ignores labels, notes entries

  (codeblocktrailer) 
  (wrs oldout) 
  (return (list 'internalcodeptr entryname*))))	% NOT InternalEntry
                                  % this makes anonymous functions work
 

(de codeblockheader nil nil)
(de codeblocktrailer nil nil)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%      AsmOutInstruction -- Used to print one line of assembler code.      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de asmoutinstruction (x) 
 (prog (type fn format operands 
              opcode* operands* 
              operand*1 operand*2 operand*3 
              operand*4 operand*5 operand*6) 
   (cond ((stringp x) (return (printlabel (asmformatlabel x)))) 
   ((idp x) (return (printlabel (asmformatlabel (findlocallabel x))))) 
   ((atom x) (stderror (bldmsg "AsmOutInstruction: %w" x)))) 
   (setq opcode* (car x)) 
   (setq type (get opcode* 'opcodetype)) 
   (cond ((null type) 
   (progn (setq fn (get opcode* 'asmpseudoop)) 
       (cond ((null fn) 
        (stderror (bldmsg "AsmOutInstruction: no type or PseudoOp for: %w" 
                         x)))) 
       (return (apply fn (list x)))))) 
   (setq format (get type 'asminstructionformat)) 
   (cond ((null format) 
        (stderror (bldmsg "AsmOutInstruction: no format for: %w" x)))) 
   (setq operands* (cdr x)) 
   (setq operands operands*) 
   (cond ((and (numberp (car format)) (neq (car format) (length operands))) 
        (stderror (bldmsg "AsmOutInstruction: wrong number of args: %w" x)))) 
   %
   %  Set fluid variables for evaluation of template forms
   %
   (cond ((numberp (car format)) 
      (cond ((not (null operands)) 
      (progn (setq operand*1 (car operands)) 
         (cond ((not (null (setq operands (cdr operands)))) 
         (progn (setq operand*2 (car operands)) 
            (cond ((not (null (setq operands (cdr operands)))) 
            (progn (setq operand*3 (car operands)) 
               (cond ((not (null (setq operands (cdr operands)))) 
               (progn (setq operand*4 (fourth operands)) 
                  (cond ((not (null (setq operands (cdr operands)))) 
                  (progn (setq operand*5 (fifth operands)) 
                     (cond ((not (null (setq operands (cdr operands)))) 
                     (progn (setq operand*6 (sixth operands)) 
                        (cond ((not (null (setq operands (cdr operands)))) 
                        (stderror 
                            (bldmsg 
                               "AsmOutInstruction: more than 6 operands: %w" 
                               x)))))))))))))))))))))))) 
   %
   %  The field formats use primitive formatting routines defined below.
   %
   (printf "%t%w%t" 9 (car x) 18)       % Opcode
   (foreach fieldformat in (cdr format) do (prin2 (eval fieldformat)))))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       Format Labels                                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de printlabel (x) 
 (progn (cond ((lessp (posn) 9) (printf "%tDS       0H" 9))) 	% happens rarely
     % avoid printing nonsense labels
     (cond ((or (not (stringp x)) (not (equal (strbyt x 0) (char *)))) 
        (printf "%f%w " x)))))
 

(de asmformatlabel (x) 
 x)

(de locallabelp (x) 
 (or (idp x) (stringp x)))

(de findlabel (x) 
 (prog (y) 
  (return (cond ((setq y (atsoc x locallabels*)) (cdr y)) 
    ((setq y (get x 'asmsymbol)) y) (t 
%   else if ( Y := get( X, 'WConst )) then Y
    (findlocallabel x))))))
 

(de findlocallabel (x) 
 (prog (y) 
  (return 
    (cond ((setq y (atsoc x locallabels*)) (cdr y)) (t 
    (progn (setq locallabels* (cons (cons x (setq y (stringgensym))) locallabels*)) 
             y))))))
 

(de findgloballabel (x) 
 (or (get x 'asmsymbol) (errorprintf "***** Undefined symbol %r" x)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       Format Registers                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(fluid '(numericregisternames*))

(de asmformatregister (reg) 
 (prog (name) 
  (cond ((or (atom reg) (not (memq (car reg) '(reg implicitreg)))) 
     (stderror (bldmsg "%w is not a register" reg)))) 
  (setq reg (cadr reg)) 
  (cond ((numberp reg) 
     (cond ((and (geq reg 0) (leq reg 15)) (return reg)) (t 
     (stderror (bldmsg "%w is an illegal register" reg)))))) 
  (cond ((not (idp reg)) 
     (stderror (bldmsg "%w is an illegal register" reg)))) 
  (setq name (get reg 'registername)) 
  (cond ((not name) 
     (stderror (bldmsg "%w is an illegal register" reg)))) 
  (return name)))
 

% see 370-lap.red
 (de extrareg (n) 
  (case n ((0) "EXREG0") 
          ((6) "EXREG6") 
          ((7) "EXREG7") 
          ((8) "EXREG8") 
          ((9) "EXREG9") 
          ((10) "EXREG10") 
          ((11) "EXREG11") 
          ((12) "EXREG12") 
          ((13) "EXREG13") 
          ((14) "EXREG14") 
          ((15) "EXREG15") 
          (nil "EXREGERROR")))
  

(setq numericregisternames* '[0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15])

(deflist '((nilr nil0) (nil             % 0  (Reg NilR) is synonym for (Reg Nil)
           nil0)                        % 0
                                % 1-5 reserved for compiler
                                %
                                % 6-9 temp/call (two even-odd pairs)
           (t0 tmp6)                    % 6   temp (= LinkReg)
           (t1 tmp7)                    % 7   temp
           (t2 tmp8)                    % 8   temp (= NargReg)
           (t3 tmp9)                    % 9   temp (= CallReg)
           (tapply tmp6)                % Safe during function call
                                % i.e neither call9 nor narg8
           (linkreg link6)              % 6   link register             (= T0)
           (nargreg narg8)              % 8   number-of-args register   (= T2)
           (callreg call9)              % 9   call register             (= T3)
                                % 10-15 bases/control
           (symfncr fnc10)              % 10  base of symfnc array
           (symvalr val11)              % 11  base of symval array
           (datar dat12)                % 12  base of dsect
           (baser bas13)                % 13  base of active routine
           (retr ret14)                 % 14  return register (must be BaseR+1)
           (ctlstkr stk15)              % 15  active end of control stack
           (st stk15)                   % 15  active end of control stack
           (r15 r15))                   % 15  (used by some SVC's, etc.)
           'registername)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  Format Immediate-Bytes and Masks                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de asmformatimmediatebyte (operand) 
  (cond ((stringp operand) operand) 
        (t (asmformatnumber (operandimmediatebyteof operand) 
                            "immediate byte" 255))))

(de asmformatfourbitmask (operand) 
  (cond ((stringp operand) operand) 
        (t (asmformatnumber (operandfourbitmaskof operand) "mask" 15))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  Format Complex Operands                                 %
%                                                                          %
%       Disp(Base)  Disp(Index,Base)  Disp(Length,Base)  etc.              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de asmformatdispbase (operand) 
 (prog (disp formatteddisp breg 
              expressioncount*) 
  (setq expressioncount* -1)            % suppress top level parens
  (cond ((or (stringp operand) (numberp operand)) (return operand))) 
  (cond ((operandindexregisterof operand) 
     (stderror (bldmsg "Index register not allowed in %w" operand)))) 
  (setq disp (operanddisplacementof operand)) 
  (setq formatteddisp (asmformatnumber disp "displacement" 4095)) 
  (cond ((not (numberp formatteddisp)) 
     (setq formatteddisp (asmformatexpression disp)))) 
  (setq breg (operandbaseregisterof operand)) 
  (return (cond ((null breg) formatteddisp) 
         ((eqcar breg 'implicitreg) 
              (asmformatexpression disp)) (t 
         (bldmsg "%w(%w)" formatteddisp (asmformatregister breg)))))))
 


(de asmformatdispindexbase (operand) 
 (prog (disp formatteddisp xreg breg 
              expressioncount*) 
  (setq expressioncount* -1)            % suppress top level parens
  (cond ((or (stringp operand) (numberp operand)) (return operand))) 
  (setq disp (operanddisplacementof operand)) 
  (setq formatteddisp (asmformatnumber disp "displacement" 4095)) 
  (cond ((not (numberp formatteddisp)) 
     (setq formatteddisp (asmformatexpression disp)))) 
  (setq xreg (operandindexregisterof operand)) 
  (setq breg (operandbaseregisterof operand)) 
  (cond ((eqcar breg 'implicitreg) 
    (progn (setq formatteddisp 
              (cond ((idp operand) (asmformatexpression operand)) 
                     (t (asmformatexpression disp)))) 
        (setq breg nil)))) 
  (return 
   (cond ((null xreg) 
        (cond ((null breg) formatteddisp) (t 
        (bldmsg "%w(%w)" formatteddisp (asmformatregister breg))))) 
   ((null breg) 
        (bldmsg "%w(%w)" formatteddisp (asmformatregister xreg))) (t 
   (bldmsg "%w(%w,%w)" formatteddisp (asmformatregister xreg) 
                             (asmformatregister breg)))))))
 


(de asmformatdisplengthbase (operand limit) 
 (prog (disp formatteddisp l breg expressioncount*) 
  (setq expressioncount* -1)            % suppress top level parens
  (cond ((stringp operand) (return operand))) 
  (setq disp (operanddisplacementof operand)) 
  (setq formatteddisp (asmformatnumber disp "displacement" 4095)) 
  (cond ((not (numberp formatteddisp)) 
     (setq formatteddisp (asmformatexpression disp)))) 
  (setq l (operandbytelengthof operand limit)) 
  (setq breg (operandbaseregisterof operand)) 
  (cond ((eqcar breg 'implicitreg) 
    (progn (setq formatteddisp (asmformatexpression disp)) 
        (setq breg nil)))) 
  (return (cond ((null breg) (bldmsg "%w(%w)" formatteddisp l)) 
                (t (bldmsg "%w(%w,%w)" formatteddisp l 
                                   (asmformatregister breg)))))))
 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  Format Entry                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de asmformatentry (operand) 
 (cond ((or (stringp operand) (idp operand)) operand) 
       (t (stderror (bldmsg "AsmFormatEntry: %w" operand)))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  Format DefineStorage, DefineConstant                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de asmformatdefinestorage (operand) (asmformatdefineconstant operand))

(de asmformatdefineconstant (operand) 
 (cond ((stringp operand) operand) 
 ((eqcar operand 'repeat) 
      (bldmsg "%w%w" (asmformatnumber (cadr operand) "repeat" 999999) 
                      (asmformatdefineconstant (caddr operand)))) 
 ((numberp (car operand)) 
      (bldmsg "%w%w" (asmformatnumber (car operand) "repeat" 999999) 
                      (asmformatdefineconstant (cdr operand)))) 
 ((memq (car operand) '(doubleword d)) 
      (bldmsg "D'%w'" (asmformatexpression (cadr operand)))) 
 ((memq (car operand) '(fullword f)) 
      (bldmsg "F'%w'" (asmformatexpression (cadr operand)))) 
 ((memq (car operand) '(halfword h)) 
      (bldmsg "H'%w'" (asmformatexpression (cadr operand)))) 
 ((eqcar operand 'cl) 
      (bldmsg "CL%w'%w'" 
              (asmformatnumber (cadr operand) "CL length" 999999) 
              (asmformatexpression (caddr operand)))) 
 ((eqcar operand 'al) 
      (bldmsg "AL%w(%w)" 
              (asmformatnumber (cadr operand) "AL length" 999999) 
              (asmformatexpression (caddr operand)))) (t 
 (stderror (bldmsg "AsmFormatDefineConstant: %w" operand)))))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  Format Problems                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de asmformatproblem (operands) 
 (prog (str) 
  (cond ((null operands) (return "<unspecified>"))) 
  (setq str (bldmsg "%w" (car operands))) 
  (foreach operand in (cdr operands) do (setq str (bldmsg "%w,%w" str operand))) 
  (return str)))
 


(fluid '(expressioncount*))

(setq expressioncount* 0)

(de asmformatnumber (x type limit) 
 (prog (y expressioncount*) 
  (setq expressioncount* -1)            % suppress top level parens
  (setq y (resolvewconstexpression x)) 
  (cond ((not (numberp y)) (return nil))) 
  (cond ((or (lessp y 0) (lessp limit y)) 
     (stderror (bldmsg "Out of range %w: %w" type x)))) 
  (return y)))
 

(de asmformatexpression (x) 
((lambda (expressioncount*) 
  (prog (y hd tl fn) 
    (setq y (resolvewconstexpression x)) 
    (cond ((or (numberp y) (stringp y)) (return y))) 
    (cond ((idp y) (return (findlabel y)))) 
    (cond ((atom y) 
    (progn (errorprintf "***** Oddity in expression %r" x) 
        (return y)))) 
    (setq hd (car y)) 
    (setq tl (cdr y)) 
    (cond ((setq fn (get hd 'binaryasmop)) 
       (return (cond ((greaterp expressioncount* 0) 
                    (bldmsg "(%w%w%w)" (asmformatexpression (car tl)) fn 
                                        (asmformatexpression (cadr tl)))) (t 
               (bldmsg "%w%w%w" (asmformatexpression (car tl)) fn 
                                      (asmformatexpression (cadr tl)))))))) 

    (cond ((setq fn (get hd 'unaryasmop)) 
       (return (bldmsg "%w%w" fn (asmformatexpression (car tl)))))) 

    (cond ((setq fn (get hd 'asmexpressionformat)) 
       (return (apply 'bldmsg (cons fn (foreach y in tl collect 
                                   (list 'asmformatexpression (mkquote y)))))))) 

    (cond ((and (setq fn (getd hd)) (equal (car fn) 'macro)) 
       (return (asmformatexpression (apply (cdr fn) (list y)))))) 

    (cond ((setq fn (get hd 'asmexpressionfunction)) 
       (return (apply fn (list y))))) 

    (errorprintf "***** Unknown expression %r" x))) 

  (plus expressioncount* 1)))



(deflist '((plus2 !+) (wplus2 
           !+) 
           (difference !-) (wdifference 
           !-) 
           (times2 *) (wtimes2 
           *) 
           (quotient /) (wquotient 
           /)) 
          'binaryasmop)

(deflist '((minus !-) (wminus 
           !-)) 
          'unaryasmop)


(deflist '((wconst "%w") 
           (mkitem "AL1(%w),AL3(%w)")) 
         'asmexpressionformat)

(off syslisp)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%The following functions MUST overlay their countreparts in 370lap
%and therefore are flagged 'lose AFTER they are compiled.
%
%But they must be compiled whenever this file is compiled.
%and therefore the 'lose flag is removed at compile time.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on syslisp)

   (de computeonelaplength (u) 
    (or (computeonelaplength1 u) 
     (progn (errorprintf "ComputeOneLapLength using 0 for %p%n" u) 0)))
         



   (de computeonelaplength1 (u) 
    (cond ((not (memq (car u) '(ds dc))) nil) 
    ((stringequal (cadr u) "0H") 0) 
    ((atom (cadr u)) nil) 
    ((numberp (car (cadr u))) (wtimes2 
         (car (cadr u)) (computeonelaplength2 (cdr (cadr u))))) (t 
    (computeonelaplength2 (cadr u)))))

   (de computeonelaplength2 (u) 
         (cond ((memq (car u) '(d doubleword)) 8) 
    ((memq (car u) '(f fullword)) 4) 
    ((memq (car u) '(h halfword)) 2) 
    ((memq (car u) '(x byte)) 1) 
    ((memq (car u) '(al cl xl)) (cadr u)) (t 
    (progn (errorprintf "ComputeOneLapLength using 0 for %p%n" u) 0))))
             

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Routines to calculate the length of specific (pseudo-)opcodes      %

(deflist '((*entry lapentrylength)      % usually produces code
          (entry 0)                     % only produces declaration
          (string lapstringlength) 
          (ebcdic lapstringlength) 
          (ascii lapstringlength) 
          (fullword lapwordlength) 
          (halfword laphalfwordlength) 
          (byte lapbytelength) 
          (float 8) 
          (comment 0) 
          (using 0) 
          (samelinecomment 0)) 
          'laplength)

   (de lapentrylength (u) 
    (cond ((weq (cadr u) 'pslmain) 0)   % special case
    ((numberp codeblockbase*)           % second entry in fn (rare)
     (progn (setq codeblockoffsets* (cons (wdifference 
                               (cons (cadr u) lapdepositptr*) codeblockbase*) codeblockoffsets*)) 14)) (t 
         % extra 4 for SH to adjust base reg
    % normal case
     (progn (setq codeblockoffsets* (cons (cons (cadr u) 0) codeblockoffsets*)) 10))))
         

% Now ensure that these functions cannot be redefined if 370lap is
% loaded after 370alap

%flag( '(ComputeOneLapLength
%        ComputeOneLapLength1
%        ComputeOneLapLength2
%        LapEntryLength), 'Lose);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Code to make Internal entries work in cms cross compiler
%

% ResolveOperand will not examine TerminalOperands
% !&ComLis1 will not examine !&ConstP objects

% These apparently wonb't cross-compile because of the way
% expanditem treats "internalcodeptr" by producing a wierd
% item--  0Fnnnnnnin expanditem. The init file ends up with
% an undefined symbol 0Fnnnnnn.  We work around this by
% using  intern


  (flag (list 'internalcodeptr) 'terminaloperand)
  (flag (list 'internalcodeptr) 'const)
  (put 'internalcodeptr 'pass1pseudoop 'saveconstant)


% These functions must overlay prior versions, butbecause
% the cross compile treats 'Lose as "do not compile" we remove
% the lose flag and reinstate below

%Compiletime << RemFlag('(!&MkFunc ExpandItem), 'Lose); >>;


% First, a modified &mkfunc. Overlays version in complr.red

(de &mkfunc (x) 
     (cond ((eqcar x 'internalcodeptr) 
      (progn (put (cadr x) 'asmsymbol (cadr x)) x)) (t 
     (mkquote x))))

%
% Second, Modifed Exanditem to overlay version in CNLAP SL
%

 (de expanditem (expression) 
  (prog (labelofcontents) 
   (return 
    (cond ((inump expression) expression) 
    ((idp expression) 
          (makemkitem (tagnumber expression) (list 'idloc expression))) 
    ((codep expression) 
          (makemkitem (tagnumber expression) expression)) 
    ((eqcar expression 'internalcodeptr) 
          (makemkitem 15 (cadr expression))) (t 
    (progn (setq labelofcontents (savecontents expression)) 
           (makemkitem (tagnumber expression) labelofcontents)))))))
  

% Flag('(!&MkFunc ExpandItem), 'Lose);      % these defs required for
                                           % cross compilation

(off syslisp)
% of file

