%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:IBMXA-CMAC.sl
% Title:          Patterns and predicates for IBM/370 XA PSL cmacro expansion
% Author:         Winfried Neun
% Created:        February 15, 1991
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Revisions:  

% 17-Sep-1991 (Winfried Neun)
% Removed (?) problem in tempregister allocation of Extraregs.

(on fast-integers)

(compiletime
  (progn (on eolinstringok)
     (dm * (u)  nil)
     nil))

(fluid '(AddressingUnitsPerItem
     CharactersPerWord
     StackDirection
     *ImmediateQuote
     AddressingUnitsPerFunctionCell))

(loadtime (off ImmediateQuote))

(flag '(Immediate UnImmediate Indirect displacement predecrement label
	fluid global $fluid $global
	extrareg postincrement doubleindexed Indexed absolute deferred reglist)
    'TerminalOperand)

% MkItem may be used when evaluating WConst expressions.

(de MkItem (TagPart InfPart) % ZIB
  (lor (lsh TagPart 27) (land InfPart 16#7ffffff)))

(de asmp () *writingasmfile)

(de freep (asize) (and (eq 0 asize) (not &links-present&)))

(de FiveP (x)  (eq x 5))

(de TwentySevenP (x)  (eq x 27))

(de evenp (x) (and (fixp x) (eq 0 (land x 1))))

(de indexedp (x) (or (eqcar x 'displacement) 
		     (eqcar x 'doubleindexed ) (eqcar x 'indexed)))

(de singleindexedp (x) (or (eqcar x 'displacement)  (eqcar x 'indexed)))

(de immediatep (x) (and (posintp x) (wlessp x 16#1000)))

(de DISPINUMP  (x) (immediatep x))
(de onebyte-immediatep (x) (and (posintp x) (wlessp x 16#100)))

(de DeferrableP (Expression)
  (and (PairP Expression)
	 (not (memq (first Expression)
	        '(Deferred $Fluid $Global Fluid Global)))))

(de RegisterDeferredP (Expression)
  (and (EqCar Expression 'Deferred) (EqCar (second Expression) 'Reg)))


(de FakeRegisterNumberP (Expression)
  (and (IntP Expression) (GreaterP Expression 5)))

(de negp (value) (and (numberp value) (lessp value 0)))

(de displacementp (x) (and (pairp x) (eq (car x) 'displacement)))

(de NegativeImmediateP (Expression)
  (and (negintp Expression) (wGreaterP Expression -16#1000)))

% expansions of addressing modes.

(DefAnyreg CAR  ZIBanyregCAR)  % ZIB

(de ZIBAnyregCAR (reg source)
   (prog (areg)
	(setq source (resolveoperand reg source))
	(if (and (displacementp source)     % recursive case, e.g. caar
	         (member (cadr source) *nextreglist*))
	       (setq areg (cadr source))
	       (setq areg (nexttempreg)))
	(if (regp source)  % its a reg or a frame reg
	    (push `(*field ,areg ,source infstartingbit infbitlength)
			  Resultingcode*)

	    (push `(*move ,source , areg) Resultingcode*)
	    (push `(*field ,areg ,areg infstartingbit infbitlength)
	                    Resultingcode*))
	(return `(displacement ,areg 0))))

(DefAnyreg CDR  ZIBanyregCDR)  % ZIB

(de ZIBAnyregCDR (reg source)
   (prog (areg)
	(setq source (resolveoperand reg source))
	(if (and (displacementp source)     % recursive case, e.g. cddar
	         (member (cadr source) *nextreglist*))
	       (setq areg (cadr source))
	       (setq areg (nexttempreg)))
	(if (regp source)  % its a reg or a frame reg
	    (push `(*field ,areg ,source infstartingbit infbitlength)
	                    Resultingcode*)
	    (push `(*move ,source , areg) Resultingcode*)
	    (push `(*field ,areg ,areg infstartingbit infbitlength)
	                    Resultingcode*))
	(return `(displacement ,areg 4))))

(fluid '(*nextreg* *nextreglist*))

(setq *nextreg* 0        % Four registers should be sufficient
	*nextreglist* '( (reg t1) (reg t2) (reg r1) ))

(de nexttempreg (x)   % ZIB  :This saves Regs , non static assignment
     (setq *nextreg* (wplus2 *nextreg* 1))
     (when (eq *nextreg* 4) (setq *nextreg* 1))
     (nth *nextreglist* *nextreg*))

(DefAnyreg Memory  ZIBanyregMemory)  % ZIB

(de ZIBAnyregMemory (reg source offset)
  (prog (aReg cadrsource)
   (setq source (resolveoperand reg source))
   (setq offset (resolveoperand reg offset))
   (cond ((and (registerp source) (registerp offset))
	       (return `(doubleindexed ,source ,offset 0)))
					 % 370 has double index
	   ((registerp source)
	(return (if (eq 0 offset) `(Displacement ,source 0)
	               (if (immediatep offset)
	                      `(displacement ,source ,offset)
	                      (progn
	                        (setq aReg (nexttempreg))
	                                       % take a new TempReg
	                        (push `(*move ,offset ,aReg) Resultingcode*)
	                    `(doubleindexed ,source ,aReg 0) )))))
    % what shall that ?
	((and (AddressConstantP source)(ZeroP offset))
	    (return  `(unimmediate ,SOURCE ))))
    (setq aReg (if (and (displacementp source)
	              (member (setq cadrsource (cadr source))
	                           *nextreglist*))
	         cadrsource      (nexttempreg))) % take a new TempReg
    (push `(*move ,source ,aReg) Resultingcode*)
    (if (immediatep offset) (return `(displacement ,aReg  ,offset))
	(push `(*wplus2 ,areg ,offset)  Resultingcode*)
	(return `(displacement ,aReg 0)))))

(DefAnyreg QUOTE
	 AnyregQUOTE
	 ((Null) (REG NIL))
	 ((EqTP) (FLUID T))
	 ((InumP) SOURCE)
	 (       (QUOTE SOURCE)))

(DefAnyreg FRAME ZIBAnyregFrame)  % ZIB

(de ZIBAnyregFRAME (dum number)
    `(displacement (reg r13) ,(plus2 76 (times2 4 number))))

(de ZIBanyregReg (reg source)
   (cond ((Fakeregisternumberp source) 
                (setq reg (nexttempreg))  % '(reg r1))
		(when (equal reg '(reg t1))
			(setq reg (nexttempreg))) %(reg t1) not allowed 
						  % (lambind, progbind)
                (push `(dummy ,reg
			  (lit (fullword (idoffset argumentblock))))
			 resultingcode*) 
		(push `(L ,reg
			 (lit (fullword (Valuecell ($fluid argumentblock)))))
			 resultingcode*)
		(push `(L ,reg (displacement ,reg 0)) resultingcode*)
		`(displacement ,reg ,(wtimes2 4 (wplus2 source -5) )))
         (t   `(REG ,SOURCE))))

(DefAnyreg REG ZIBanyregReg)
%-----------------------------------------------------------------------------

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (if (FlagP FunctionName 'ForeignFunction)
	(list (list '*ForeignLink FunctionName FunctionType NumberOfArguments)
	  '(*Exit 0))
      (if (get functionname 'exitopencode)
              `((L (REG T2) (LIT (FULLWORD (idoffset SYMFNC))))
		(L (REG T2) (LIT (FULLWORD (VALUECELL ($FLUID SYMFNC)))))
		(L (REG T2) (DISPLACEMENT (REG T2) 0))
		(*DeAlloc ,DeAllocCount)
                (*JCall ,FunctionName))
      (if (get functionname 'opencode)
              `((*Call ,FunctionName)
                (*exit ,DeAllocCount))
      (if (InternallyCallableP FunctionName)
	`((l (reg r15) (lit (fullword (InternalEntry ,FunctionName))))
	  (*DeAlloc ,DeAllocCount)
	  (br (reg r15)))
       (if (asmp)
	`((L   (reg t1) (lit "    DC   V(SYMFNC)"))
	  (*DeAlloc ,DeAllocCount)
	  (*JCall ,FunctionName))
        `((L (reg t2) (lit (fullword (idoffset ,FunctionName))))
          (L   (reg r15) (lit (fullword (entry ,FunctionName))))
          (*DeAlloc ,DeAllocCount)
          (L   (reg r15) (indexed (reg r15) 0))
          (BR  (reg r15) ))))))))

(DefCMacro *Call
	 ((InternallyCallableP)  (l (reg r15) 
				     (lit (fullword (InternalEntry ArgOne))))
				 (balr (reg r14) (reg r15)))
	 ((asmp)	(L (reg r14) (lit "    DC   V(SYMFNC)"))
			(La (reg t2) (xxdisplacement (reg r0) argone))
		        (L   (reg r15) (xdisplacement (reg r14) argone))
			(BALR (reg r14) (reg r15)))
	 (		(L (reg t2) (lit (fullword (idoffset argone))))
			(L   (reg r15) (lit (fullword (entry argone))))
		        (L   (reg r15) (indexed (reg r15) 0))
			(Balr (reg r14) (reg r15))))

(DefCMacro *JCall
	 ((InternallyCallableP)  (l (reg r15)
                                     (lit (fullword (InternalEntry ArgOne))))
                                 (br (reg r15)))
	 ((asmp)  	     (La (reg t2) (xxdisplacement (reg r0) argone))
	                     (L   (reg r15) (xdisplacement (reg t1) argone))
			     (BR  (reg r15) ))
	 ( 		     (L (reg t2) (lit (fullword (idoffset argone))))
			     (L   (reg r15) (lit (fullword (entry argone))))
			     (L   (reg r15) (indexed (reg r15) 0))
			     (BR  (reg r15) )))

(de quote-fix-p (x)
    (and (eqcar x 'quote) (fixp (cadr x))))

(remprop '*move 'optfn)

(de quotep (x) (eqcar x 'quote))

(de asm-fluid-arg-p (x) (and *writingasmfile (fluid-arg-p x)))

(de reg0p (x) (equal x '(reg r0)))

(de conflictp (x) (and (singleindexedp x) (equal (cadr x) '(reg t2))))

(DefCMacro *Move
	(Equal)
	((regp regp)		   (lr ArgTwo ArgOne))
	((zerop regp)		   (sr argtwo argtwo))
	((indexedp regp)	   (l  ArgTwo ArgOne))
	((displacementp regp)      (l  ArgTwo Argone))
	((regp indexedp) 	   (st ArgOne ArgTwo))
	((regp displacementp)      (st ArgOne ArgTwo))
	((singleindexedp singleindexedp)
				   (mvc (lengthbased 4 argtwo) argone))
	((regp asm-fluid-arg-p)    (L (reg t2) (lit "    DC   V(SYMVAL) *"))
				   (st argone (xdisplacement (reg t2) argtwo)))
	((asm-fluid-arg-p reg0p)   (L (reg t2) (lit "     DC   V(SYMVAL) *"))
				   (L argtwo (xdisplacement (reg t2) argone)))
	((asm-fluid-arg-p regp)    (L  argtwo (lit "     DC   V(SYMVAL) *"))
				   (l argtwo (xdisplacement argtwo argone)))
	((regp fluid-arg-p)	   (dummy (reg t2)
					(lit (fullword (idoffsetc argtwo))))
				   (L (reg t2)
					 (lit (fullword (Valuecell argtwo))))
				   (ST argone (displacement (reg t2) 0)))
	((fluid-arg-p conflictp)   (dummy (reg t1)
					(lit (fullword (idoffsetc argone))))
				   (L (reg t1) 
					(lit (fullword (Valuecell argone))))
				   (mvc (lengthbased 4 argtwo)
					(displacement (reg t1) 0)))
	((fluid-arg-p singleindexedp)
				   (dummy (reg t2)
					(lit (fullword (idoffsetc argone))))
			           (L (reg t2)
                                        (lit (fullword (Valuecell argone))))
                                   (mvc (lengthbased 4 argtwo)
                                        (displacement (reg t2) 0)))
	((conflictp fluid-arg-p)   (dummy (reg t1)
				   	(lit (fullword (idoffsetc argtwo))))
				   (L (reg t1)
                                        (lit (fullword (Valuecell ArgTwo))))
                                   (mvc (lengthbased 4 
						(displacement (reg t1) 0))
						argone))
	((singleindexedp fluid-arg-p)
				   (dummy (reg t2)
					(lit (fullword (idoffsetc argtwo))))
				   (L (reg t2)
                                        (lit (fullword (Valuecell ArgTwo))))
				   (mvc (lengthbased 4 
						(displacement (reg t2) 0))
						argone))
        ((fluid-arg-p reg0p)       (dummy (reg t2)
					(lit (fullword (idoffsetc argone))))
				   (L (reg t2)
                                         (lit (fullword (Valuecell argone))))
                                   (L argtwo (displacement (reg t2) 0)))
	((fluid-arg-p regp)	   (dummy argtwo
					(lit (fullword (idoffsetc argone))))
				   (L argtwo
				         (lit (fullword (Valuecell argone))))
				   (L argtwo (displacement argtwo 0)))
	((immediatep regp)         (la ArgTwo (displacement (reg r0) ArgOne)))
	((negativeimmediatep regp) (*move (minus ArgOne) ArgTwo)
				   (LCR ArgTwo ArgTwo))
	((numberp regp)            (l Argtwo (lit (fullword argone))))
	((quotep regp)		   (l argtwo argone))
	(                (*move ArgOne (reg r0))
			 (*move (reg r0) ArgTwo)))

(de *Alloc (N)
  (setq NAlloc* N)
  (Expand1OperandCMacro
   (times N (compiler-constant 'AddressingUnitsPerItem))
   '*Alloc))

(defcmacro *ALLOC
	((Freep) (using "*" (reg r15)))
	(	(STM (Reg r14) (reg r12) (displacement (reg r13) 12))
		(*move (plus2 80 ArgOne) (reg r0))
		(L  (reg r2)  (displacement (reg r13) 76))
		(ALR (reg r0) (reg r2))
		(L   (reg r15)  (displacement (reg r13) 72))
		(STM (reg r15) (reg r0)  (displacement (reg r2) 72))
		(MVI  (displacement (reg r2) 0) 16)
		(st  (reg r13)  (displacement (reg r2) 4))
		(Lr  (reg r13) (reg r2))
		(Balr (reg r3) (reg r0))
		(Using "*" (reg r3) (reg r2))
% should be optimize some day
		(la (reg r2) (displacement (reg r3) 2048))
		(la (reg r2) (displacement (reg r2) 2048))))

(DefCMacro *DeAlloc
	 ((FreeP))
	 (     (L  (reg r13) (displacement (reg r13) 4))
		(L (reg r14) (displacement (reg r13) 12))
		(LM (reg r2) (reg r3) (displacement (reg r13) 28))))

(DefCMacro *Exit
     % leaf routine first
	 ((Freep) (Br (reg r14)))
	 (	  (*dealloc ArgOne)
	  	  (Br (reg r14))))

(de deferredp (x) (and (pairp x) (eq (car x) 'deferred)))

(DefCMacro *Jump
	 ((deferredp)        (*move (undeferred ArgOne) (reg t2))
	                     (BR (reg t5)))
	 ((displacementp)    (B  ArgOne))
	 (                   (B  ArgOne)))

(DefCMacro *Lbl           (ArgOne))

(de regp (x) (eqcar x 'Reg))

(de *3op (a1 a2 instruction i2)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	     (list (ResolveOperand '(Reg t1) A1)
	           (ResolveOperand '(Reg t2) A2)
	            Instruction i2)
	           (get '*3op 'CMacroPatternTable)))))

(DefCMacro *3op
	 ((regp regp) (ARGTHREE ArgOne ArgTwo))
	 ((regp fixp) (ARGFOUR ArgOne (lit (fullword ArgTwo))))
	 ((regp AnyP) (*move ArgTwo (reg T1))
	               (*3op ArgOne (reg t1) argthree ArgFour))
	 ((Anyp regp) (*move ArgOne (reg T2))
	               (*3op (reg t2) ArgTwo ArgThree ArgFour)
	               (*move (reg t2) ArgOne))
	 ((Anyp fixp) (*move ArgOne (reg T2))
	               (argfour (reg t2) (lit (fullword ArgTwo)))
	               (*move (reg t2) ArgOne))
	 (            (*move ArgTwo (reg T1))
	               (*move ArgOne (reg T2))
	               (argthree (reg t2) (reg t1))
	               (*move (reg t2) ArgOne)))

(de *WPlus2 (Arg1 Arg2)
  (Expand2OperandCMacro Arg1 Arg2 '*WPlus2))

(de *WDifference (Arg1 Arg2)
  (Expand2OperandCMacro Arg1 Arg2 '*WDifference))

(DefCMacro *WPlus2
	 ((*3op ArgOne ArgTwo AR A)))

(DefCMacro *WDifference
	 ((*3op ArgOne ArgTwo SR S)))

(defcmacro *wtimes2)

(de *wtimes2 (arg1 arg2)
    (expand2operandcmacro arg1 arg2 '*wtimes2))

(DefCMacro *wtimes2
	 ((regp regp) (Mr ArgOne ArgTwo))
% ((regp immediatep) (M ArgOne (displacement (reg r0) ArgTwo)))
	 ((regp AnyP) (*move ArgTwo (reg T1))
	              (*wtimes2 ArgOne (reg t1)))
	 ((Anyp regp) (*move ArgOne (reg T2))
	               (*wtimes2 (reg t2) ArgTwo)
	               (*move (reg t2) ArgOne))
	 (            (*move ArgTwo (reg T1))
	               (*move ArgOne (reg T2))
	               (mr (reg t2) (reg t1))
	               (*move (reg t2) ArgOne)))

(defcmacro *wquotient)

(de *wquotient (arg1 arg2 arg3)
    (expand2operandcmacro arg1 arg2 '*wquotient))

% format ArgOne = ArgTwo * argthree

(DefCMacro *wquotient
	((anyp onep)  (*move ArgTwo ArgOne))
%          ((regp powerof2p)
%                   (*wquotientshift ArgOne ArgTwo))
	((regp regp) (DR ArgOne ArgTwo))
	((regp immediatep)
	              (D ArgOne (displacement (reg 0) ArgTwo)))
	((regp AnyP) (*move ArgTwo (reg T1))
	               (*wquotient ArgOne (reg t1) ))
	((Anyp regp) (*move ArgOne (reg T2))
	               (*wquotient (reg t2) ArgTwo)
	               (*move (reg t2) ArgOne))
	((Anyp immediatep)
	              (*move ArgOne (reg T2))
	               (D (reg t2) (displacement (reg 0) ArgTwo ))
	               (*move (reg t2) ArgOne))
	(           (*move ArgTwo (reg T1))
	               (*move ArgOne (reg T2))
	               (DR (reg t2) (reg t1))
	               (*move (reg t2) ArgOne)))

(DefCMacro *WAnd
	 ((*3op ArgOne ArgTwo NR N)))

(DefCMacro *WOr
	 ((*3op ArgOne ArgTwo OR O)))

(DefCMacro *WXOr
	 ((*3op ArgOne ArgTwo XR X)))

(de fixplusp (x) (and (fixp x) (not (minusp x))))

(DefCMacro *AShift
    ((AnyP ZeroP))
    ((regp negp)     (sra ArgOne (minus ArgTwo)))
    ((regp fixplusp) (sll ArgOne ArgTwo ))
    ((regp regp)    (LCR (reg R1) ArgTwo)
	             (BH templabel)
	             (sla ArgOne (displacement ArgTwo 0))
	             (b templabel2)
	           templabel
	             (sra ArgOne (displacement (reg r1) 0))
	           templabel2)
    ((regp anyp)   (*move ArgTwo (reg t2))
	            (*ashift ArgOne (reg t2)))
    ((Anyp fixp)   (*move ArgOne (reg t2))
	            (*ashift (reg t2) ArgTwo)
		    (*move (reg t2) ArgOne))
    (              (*move ArgOne (reg t1))
	            (*ashift (reg t1) ArgTwo)
	            (*move (reg t1) ArgOne)))


(DefCMacro *WShift
    ((AnyP ZeroP))
    ((regp negp)     (srl ArgOne (minus ArgTwo)))
    ((regp fixplusp) (sll ArgOne ArgTwo ))
    ((regp regp)    (LCR (reg r1) ArgTwo)
                     (BH templabel)
                     (sll ArgOne (displacement ArgTwo 0))
                     (b templabel2)
                   templabel
                     (srl ArgOne (displacement (reg r1) 0))
                   templabel2)
    ((regp anyp)   (*move ArgTwo (reg t2))
                    (*wshift ArgOne (reg t2)))
    ((Anyp fixp)   (*move ArgOne (reg t2))
                    (*wshift (reg t2) ArgTwo)
                    (*move (reg t2) ArgOne))
    (              (*move ArgOne (reg t1))
                    (*wshift (reg t1) ArgTwo)
                    (*move (reg t1) ArgOne)))


(DefCMacro *WNot
	 ((regp regp) (*move -1 (reg t1))
		      (XR ArgOne (reg t1)))
	 ((regp Anyp) (*move ArgTwo ArgOne)
	               (*Wnot ArgOne ArgOne))
	 (            (*move ArgTwo (reg t2))
	               (*wnot (reg t2) (reg t2))
	               (*move (reg t2) ArgOne)))

(DefCMacro *WMinus
	     ((regp regp) (LCR ArgOne ArgTwo))
	     ((regp Anyp) (*move ArgTwo ArgOne)
	                   (LCR ArgOne ArgOne))
	     (            (*move ArgTwo (reg t2))
	                   (*wminus (reg t2) (reg t2))
	                   (*move (reg t2) ArgOne)))

(DefCMacro *MkItem
	 ((regp fixp)      (*wshift ArgOne tagbitlength)
	                    (*wshift ArgOne (minus tagbitlength))
	                    (*move ArgTwo (reg t1))
			    (*wshift (reg t1) infbitlength)
			    (or ArgOne (reg t1)))
	 ((regp regp)     (*wshift ArgOne tagbitlength)
                            (*wshift ArgOne (minus tagbitlength))
                            (*move ArgTwo (reg t1))
                            (*wshift (reg t1) infbitlength)
                            (or ArgOne (reg t1)))
	 ((regp anyp)      (*move ArgTwo (reg t1))
	                    (*Mkitem ArgOne (reg t1)))
	 (                  (*move ArgOne (reg t2))
	                    (*mkitem (reg t2) ArgTwo)
	                    (*move (reg t2) ArgOne)))

(de smallinump (x) (immediatep x))

(de immediate-p (x)    (fixp x))

(DefCmacro *tag
	((regp regp) (*move ArgTwo ArgOne)
		      (srl ArgOne 27))
	((Anyp  regp)(*move ArgTwo (reg t1))
	              (srl (reg t1) 27 )
	              (*move (reg t1) ArgOne))
	((regp anyp) (*move ArgTwo ArgOne)
	              (srl ArgOne 27))
	(            (*move ArgTwo (reg t1))
	              (*tag ArgOne (reg t1))))

(de *tag (ar1 ar2)
     (expand2operandCmacro ar1 ar2 '*tag))

(DefCMacro *JumpType
	(             (*Tag (Reg t1) ArgOne)
	              (*JumpEQ ArgThree (Reg t1) ArgTwo)))

(DefCMacro *JumpNotType
	(             (*Tag (Reg t1) ArgOne)
	              (*JumpNOTEQ ArgThree (Reg t1) ArgTwo)))

(DefCMacro *JumpInType
	 (	       (*tag (reg t1) ArgOne)
			(*move ArgTwo (reg t2))
	                (CR (reg t1) (reg t2))
	                (bnh ARGTHREE)
	                (C (reg t1) (lit (fullword 31))) % negint-tag
	                (be ARGTHREE)))

(DefCMacro *JumpNotInType
         (             (*tag (reg t1) ArgOne)
                        (*move ArgTwo (reg t2))
                        (CR (reg t1) (reg t2))
                        (bnh templabel)
                        (C (reg t1) (lit (fullword 31))) % negint-tag
                        (bne ARGTHREE)
		       templabel ))

%---------------------------------------------------------------------
%   General Purpose Jumper.
%   FORMAT: Operand1 Operand2 Address JumpOpCode

(DefCMacro *JumpIf
     ((regp regp)      (CR ArgOne ArgTwo )
	                (argfour argthree))
     ((regp anyp)       (*move ArgTwo (reg t2))
	                 (CR ArgOne (reg t2))
	                 (argfour argthree))
     ((anyp  regp)      (*move ArgOne (reg t2))
	                 (CR (reg t2) ArgTwo)
	                 (argfour argthree))
     ((anyp conflictp)  (*move ArgTwo (reg r0))
			 (*move ArgOne (reg t2))
			 (CR (reg t2) (reg r0))
			 (argfour argthree))
     (		        (*move ArgOne (reg r0))
	                 (*move ArgTwo (reg t2))
	                 (CR (reg r0) (reg t2))
	                 (argfour argthree)))

(de *JumpIF (ArgOne ArgTwo Label Instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	     (list (ResolveOperand '(Reg t1) ArgOne)
	           (ResolveOperand '(Reg t2) ArgTwo)
	               (ResolveOperand '(Reg Error) Label)
	            Instruction)
	           (get '*JumpIF 'CMacroPatternTable)))))


(defcmacro *jumpeq)

(de *jumpeq (Lbl ArgOne ArgTwo)
	 (*jumpif ArgOne ArgTwo lbl 'BE))

(defcmacro *jumpnoteq)

(de *jumpnoteq(Lbl ArgOne ArgTwo)
	  (*jumpif ArgOne ArgTwo Lbl 'BNE))

(defcmacro *jumpwlessp)

(de *jumpwlessp (Lbl ArgOne ArgTwo)
	  (*jumpif ArgOne ArgTwo lbl 'BL))

(defcmacro *jumpwgreaterp)

(de *jumpwgreaterp (Lbl ArgOne ArgTwo)
	  (*jumpif ArgOne ArgTwo lbl 'BH))

(defcmacro *jumpwleq)

(de  *jumpwleq(Lbl ArgOne ArgTwo)
	  (*jumpif ArgOne ArgTwo lbl 'BNH))

(defcmacro *jumpwgeq)

(de *jumpwgeq (Lbl ArgOne ArgTwo)
	  (*jumpif ArgOne ArgTwo lbl 'BNL))


(de fluid-arg-p (x)
    (and (pairp x)
     (member (car x) '(!$fluid !$global fluid global))))

(DefCMacro *Push
	 ( (*cerror "NO push ")))

(DefCMacro *Pop
	 ( (*cerror "NO pop ")))

(de *loc (ArgOne ArgTwo)
   (let ((resolvedarg (resolveoperand '(reg t2) ArgTwo)))
	(cond ((deferred-p resolvedarg)
	   `((*move (undeferred ,resolvedarg) ,ArgOne))) % get the address.
	   ((displacement-p resolvedarg)
	      `((*move ,(caddr resolvedarg) ,ArgOne)
		(ar,ArgOne ,(cadr resolvedarg) )))
	   ((eqcar resolvedarg 'cdr)
	      `((*move 4 , ArgOne)
		(ar ,ArgOne ,(cadr resolvedarg))))
	   (t `((*move ,resolvedarg ,ArgOne))))))

(DefCMacro *loc)

%t6 is used, because it is not used in lap code, but it destroyed in cmacros.
%look at test file mask.s in ~burdorf/sparc for example.

(de X5p (u) (eq u 5))

(de X27p (u) (eq u 27))

(DefCMacro *Field

  ((regp regp fixp fixp)  (*move argtwo argone)
			   (sll ArgOne ArgThree)
	                   (srl ArgOne (difference 32 ARGFOUR) ))
  ((regp anyp fixp fixp)  (*move ArgTwo (reg t2))
	                   (sll (reg t2) argthree)
	                   (srl (reg t2) (difference 32 ARGFOUR))
			   (*move (reg t2) ArgOne))
  ((anyp regp fixp fixp)  (*field (reg t2) ArgTwo argthree argfour)
	                   (*move (reg t2) ArgOne))
  (                       (*move ArgTwo (reg t2))
	                   (*field (reg t2) (reg t2) argthree argfour)
	                   (*move (reg t2) ArgOne)))

(DefCMacro *SignedField

  ((regp regp fixp fixp)  (*move argtwo argone)
                           (sll ArgOne ArgThree)
                           (sra ArgOne (difference 32 ARGFOUR) ))
  ((regp anyp fixp fixp)  (*move ArgTwo (reg t2))
                           (sll (reg t2) argthree)
                           (sra (reg t2) (difference 32 ARGFOUR))
                           (*move (reg t2) ArgOne))
  ((anyp regp fixp fixp)  (*signedfield (reg t2) ArgTwo argthree argfour)
                           (*move (reg t2) ArgOne))
  (                       (*move ArgTwo (reg t2))
                           (*signedfield (reg t2) (reg t2) argthree argfour)
                           (*move (reg t2) ArgOne)))

(DefCMacro *PutField

     (    (*move (reg r14) (displacement (reg st) 8))
          (*move -1 (reg r14))
	  (*move   ArgOne (reg t2))
	  (sll     (reg r14) (difference 32 argfour))
	  (*wshift (reg r14) (minus argthree)) % for 0 opt
	  (sll   (reg t2) (difference 32 (plus argthree argfour)))
	  (nr   (reg t2) (reg r14))
	  (*move -1 (reg t1))
	  (XR  (reg r14) (reg t1))
	  (*move ArgTwo (reg t1))
	  (nr   (reg t1) (reg r14))
	  (or   (reg t2) (reg t1))
	  (*move (reg t2) ArgTwo)
          (*move (displacement (reg st) 8) (reg r14))))


(deflist '((Byte        ((*move (reg 1) (reg t2))
			 (SR (reg 1) (reg 1))
			 (ic (reg 1) (doubleindexed (reg t2) (reg 2) 0))))
	   (PutByte     ((stc (reg 3) (doubleindexed (reg 1) (reg 2) 0))))
	   (HalfWord    ((sll (reg 2) 1 )
	                 (lh (reg 1) (doubleindexed (reg 1) (reg 2) 0) )))
	   (PutHalfWord ((sll (reg 2) 1 )
	                 (sth (reg 3) (doubleindexed (reg 1) (reg 2) 0)))))
  'OpenCode)

(loadtime (remprop 'HalfWord 'DESTROYS))
(loadtime (remprop 'putHalfWord 'DESTROYS))

(on fast-integers)

(for (in function '(idapply0 idapply1 idapply2 idapply3 idapply4))
     (from register 1)
     (do
	(put function 'opencode
	  `((*move ($fluid symfnc) (reg t1))
	    (sll (reg ,register) 5)
	    (lr (reg t2) (reg ,register))
	    (srl (reg ,register) 3)
	    (l (reg r15) (doubleindexed (reg t1) (reg ,register) 0))
	    (srl (reg t2) 5)
	    (balr (reg r14) (reg r15))))
	(put function 'exitopencode
	  `((sll (reg ,register) 5)
            (srl (reg ,register) 3)
            (l (reg r15) (doubleindexed (reg t2) (reg ,register) 0))
	    (lr (reg t2) (reg ,register))
	    (srl (reg t2) 2)
            (balr (reg r0) (reg r15))))))

(for (in function '(addressapply0 addressapply1 addressapply2
	          addressapply3 addressapply4))
     (from register 1)
     (do
	(put function 'opencode
	   `((*move (reg ,register) (reg r15))
	     (BALR (reg r14) (reg r15)) ))
	(put function 'exitopencode
	   `((*move (reg ,register) (reg r15))
	     (BALR (reg r0) (reg r15))))))


(defcmacro *fast-apply-load
   ( (*move ArgOne (Reg t1))))

(put 'fast-idapply    'opencode
     '((*move ($fluid symfnc) (reg t2)) 
	 (sll (reg t1) 5 )
	 (srl (reg t1) 3 ) %multiply by 4 to get offset into symfnc table
	 (l (reg r15) (doubleindexed (reg t1) (reg t2) 0))
	 (lr (reg t2) (reg t1))
	 (srl (reg t2) 2) % for interpreted fns
	 (Balr (reg r14) (reg r15))))

(put 'fast-codeapply    'opencode
     '( (sll (reg t1) 5)
	(srl (reg t1) 5) %strip off tag
	(*move (reg t1) (reg r15))
	(Balr (reg r14) (reg r15))))

(put 'fast-idapply    'exitopencode
     '(  (sll (reg t1) 5 )
         (srl (reg t1) 3 ) %multiply by 4 to get offset into symfnc table
         (l (reg r15) (doubleindexed (reg t1) (reg t2) 0))
	 (lr (reg t2) (reg t1))
         (srl (reg t2) 2) % for interpreted fns
         (Br (reg r15))))

(put 'fast-codeapply    'exitopencode
     '( (sll (reg t1) 5)
        (srl (reg t1) 5) %strip off tag
        (*move (reg t1) (reg r15))
        (Br (reg r15))))

(De *LamBind (Regs Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cregs cfluids cadrcfluids lab initload freeregs)
	(setq lab (gensym))
	(setq n 0)
	(setq Regs (rest Regs))   % Remove "Registers" from the front
	(setq Fluids (rest fluids)) % Remove NONLOCALVARS
	(setq lng (wtimes2 (length Regs) 8)) % two words per BndStk entry *4

   %looking for free register to preload values of fluid s

	(setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
	             (function (lambda (x)
	                          (cond ((member x regs) nil)
	                                 (t (cons x nil))
	)              ) )         )  )
	(setq freeregs (nconc (list '(reg r1)) freeregs))
	(setq cfluids fluids) % copy of fluids

preload  (setq initload
	 (progn (setq cadrcfluids
	          (nconc cadrcfluids (cons (car freeregs) nil)))
	  (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
	 )   )
	 (setq freeregs (cdr freeregs))
	 (setq cfluids (cdr cfluids))

	 (cond ((and freeregs cfluids) (go preload)))

	 (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

	 % freeregs contains the list of preloaded regs
	 % and not preloaded fluids if those exist

	(setq list `((*move ($fluid BndStkPtr) (Reg t1))
		 (*move ($fluid BndstkUpperBound) (Reg r0))
		 (*move ,lng (reg t2))
	         (*wplus2 (Reg t2) (Reg t1))
	         (cr    (reg t2)(reg r0))
	         (bnh   ,lab)
		 (*call bstackoverflow)
		,lab 
		 (*move (reg t2) (reg r0))
	         (*move (Reg r0) ($fluid BndstkPtr))  )) %start of code

	(setq list (append initload list))

  loop
	(setq cregs (car Regs))
	(setq cfluids (car Fluids))
	(setq cadrcfluids (cadr cfluids))
	(when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	     (stderror "T and NIL cannot be rebound"))
	(setq n (wplus2 n 8))
        (if (FakeRegisterNumberP (cadr cregs))
	(Setq list (append list
	     `((*move ,(car freeregs) (indexed (Reg t1) ,n))
         	(*move (quote ,Cadrcfluids)
                     (indexed (reg t1) ,(wplus2 n -4)))
         	(*move ,cregs (reg r0))
		(*move (reg r0) ,cfluids)
        	)          ))

	(Setq list (append list
	     `((*move ,(car freeregs) (indexed (Reg t1) ,n))
		 (*move (quote ,Cadrcfluids)
	             (indexed (reg t1) ,(wplus2 n -4)))
	 	(*move ,cregs ,cfluids)
		)          )))
	(setq fluids (cdr Fluids))
	(setq freeregs (cdr freeregs))
	(cond ((setq regs (cdr Regs)) (go loop)))
	(return list)
)    )

(defcmacro !*lambind)

(De *ProgBind (Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cfluids cadrcfluids lab initload freeregs)
	(setq lab (gensym))
	(setq n 0)
	(setq Fluids (rest fluids)) % Remove NONLOCALVARS
	(setq lng (wtimes2 (length Fluids) 8)) % two words per BndStk entry
	                                   % * 4 addressingunits
	(setq freeregs '((reg r1)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
	(setq cfluids fluids) % copy of fluids

preload  (setq initload
	 (progn (setq cadrcfluids
	          (nconc cadrcfluids (cons (car freeregs) nil)))
	  (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
	 )   )
	 (setq freeregs (cdr freeregs))
	 (setq cfluids (cdr cfluids))

	 (cond ((and freeregs cfluids) (go preload)))

	 (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

	 % freeregs contains the list of preloaded regs
	 % and not preloaded fluids if those exist


        (setq list `((*move ($fluid BndStkPtr) (Reg t1))
		 (*move ($fluid BndstkUpperBound) (Reg r0))
                 (*move ,lng (reg t2))
		 (*wplus2 (Reg t2) (Reg t1))
                 (cr    (reg t2)(reg r0))
                 (bnh   ,lab)
                 (*call bstackoverflow)
                ,lab
                 (*move (reg t2) (reg r0))
                 (*move (Reg r0) ($fluid BndstkPtr))  )) %start of code

	(setq list (append initload list))

 loop
	(setq cfluids (car Fluids))
	(setq cadrcfluids (cadr cfluids))
	(when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	     (stderror "T and NIL cannot be rebound"))
	(setq n (wplus2 n 8))
	(Setq list (append list
	       `((*move ,(car freeregs) (displacement  (Reg t1) ,n))
	         (*move (quote ,Cadrcfluids)
	                     (displacement (reg t1)  ,(wplus2 n -4)))
	         (*move (quote nil) ,cfluids)
	)          ))
	(setq freeregs (cdr freeregs))
	(cond ((setq Fluids (cdr Fluids)) (go loop)))
	(return list)
)    )
(defcmacro *progbind)

(De *FreeRstr (Fluids)

 (prog (n lng list cfluids listfluids lab initload freeregs)
	(setq lab (gensym))
	(setq n 0)
	(setq Fluids (rest fluids)) % Remove NONLOCALVARS
	(setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
	(setq freeregs '((reg r1)(reg 2)(reg 3)(reg 4)(reg 5)))
	(setq cfluids fluids) % copy of fluids
	(setq n (wtimes2 4 (wdifference 2 lng)))
	(setq lng (wtimes2 lng 4)) % * addressingunitperitem
	(if (eq n 0) 
		(setq initload (list '(*move ($fluid Bndstkptr) (reg t1))))
		(setq initload (list
			 '(*move ($fluid Bndstkptr) (reg t1))
			 `(*move ,(minus n) (reg t2))
			 '(*wdifference (reg t1) (reg t2)))))
	(setq n 0)

preload  (setq initload
	 (progn (setq listfluids
	         (if freeregs
	          (nconc listfluids (cons (car freeregs) nil))
	          (nconc listfluids (cons nil nil))) )
	  (nconc initload
	   (if freeregs
                 `((*move (displacement (reg t1) ,n) ,(car freeregs))) nil)
	 )   ))
	 (setq n (wplus2 n 8))
	 (when freeregs (setq freeregs (cdr freeregs)))
	 (setq cfluids (cdr cfluids))

	 (cond (cfluids (go preload)))

	 (setq freeregs listfluids ) %end of preloadcode

	 % freeregs contains the list of preloaded regs
	 % and nil if not enough regs available

        (setq list `((*move ($fluid BndstklowerBound) (Reg r0))
		 (*move ($fluid Bndstkptr) (reg t1))
                 (*move ,lng (reg t2))
                 (*wdifference (Reg t1) (Reg t2))
                 (cr    (reg t1)(reg r0))
                 (bnl   ,lab)
                 (*call bstackunderflow)
                ,lab
                 (*move (Reg t1) ($fluid BndstkPtr))  )) %start of code

	(setq list (append initload list))
	(setq n 0)

 loop
	(setq cfluids (car Fluids))
	(setq n (wplus2 n 8))

  % insert reloaded register or memory reference

	(setq list (append list
	     (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
	          `((*move (displacement (Reg t1) ,n) ,cfluids )))

	)          )
	(setq freeregs (cdr freeregs))
	(cond ((setq Fluids (cdr Fluids)) (go loop)))
	(return list)
)    )
(defcmacro *freerstr)

(setq *unsafebinder t)   % has to save Registers across calls

(global '(*writingasmfile))

(de !*xjumpon (register lowerbound upperbound labellist)
    (for (from x lowerbound upperbound)
	   (join
	    `((*move ,x (reg t1))
	      (CR ,register (reg t1))
	      (be ,(pop labellist))))))

(de !*jumpon (register lowerbound upperbound labellist)
   (if *writingasmfile
	 (*xjumpon register lowerbound upperbound labellist)
     (PROG (X LL LL2)
	 (setq ll  (gensym))
	 (setq ll2 (stringgensym))
	 (SETQ X
	  (if (and (weq lowerbound 0) (weq upperbound 31) *syslisp)
	            % jumpon on tags (most probably)
	`((*move ,register (reg t2))
	  (sll (reg t2) 2)  % 4 bytes per jumptable entry
	  (LA (reg t1) ,ll2)
	  (*wplus2 (reg t1) (reg t2))
	  (br (reg t1))
	 ,ll2)
	`((*move ,register (reg t2))
	  (*move ,upperbound (reg r0))
	  (*jumpeq ,(lastcar labellist) (reg t2) (reg r0) )
	  (BH (label ,ll))
	  (*move ,Lowerbound (reg r0))
	  (*jumpwlessp (label ,ll) (reg t2) (reg r0) )
	  (*wdifference  (reg t2) (reg r0) )
	  (sll (reg t2) 2 )  % 4 bytes per jumptable entry
          (la (reg t1) ,ll2)
	  (*wplus2 (reg t1) (reg t2))
	  (br (reg t1))
	 ,ll2) ) )
	Loop  (Setq x (nconc X `((B ,(car Labellist)))))
	      (setq Labellist (cdr Labellist))
	      (cond (Labellist (go loop)))

	      (setq x (nconc x `((*lbl (Label ,ll))) ))
	      (return x)
)  )  )

(defcmacro !*jumpon)

% Call foreign fns (loaded by oload) by putting SymFunc cell addr in a reg,
% then doing "calls" instr indirect via the target address of the jmp
% instruction stored there.  Use old direct link for asm (kernel) compile.

(de *foreignlink (functionname functiontype numberofarguments)

   (let ((sequence nil))
     (setq sequence
	         (cond ((equal numberofarguments 0) nil)
	               (t (pnth '((*move (reg 5)  (displacement (reg r13) 116))
	                          (*move (reg 4)  (displacement (reg r13) 112))
				(*move (reg 3)  (displacement (reg r13) 108))
				(*move (reg 2)  (displacement (reg r13) 104))
				(*move (reg 1)  (displacement (reg r13) 100)))
	                        (difference 6  numberofarguments)))))
     (setq sequence
	 (append sequence
	    `((la (reg r1) (displacement (reg r13) 100))
	      (L (reg t1) (displacement (reg r13) 76))
              (*wplus2 (reg t1) 200)
	      (*wshift (reg t1) -3)
	      (*wshift (reg t1)  3)  % double word alignment of st
              (st (reg t1) (displacement (reg r13) 76))
	      (L (reg r15) (lit ,(bldmsg "    DC   V(%w)" functionname)))
	      (BALR (reg r14) (reg r15)))))
     (setq sequence
	 (append sequence
	            '((*move (reg r15) (reg 1))   %store return value in r1
	             )))))

(defcmacro !*foreignlink)

