%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:VAX-ASM.SL
% Title:          VAX VMS specific information for LAP-TO-ASM
% Author:         Eric Benson
% Created:        5 January 1982
% Modified:       10-Jul-84 15:28:20
% Package:        Kernel
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 7-July 1991 (Winfried Neun)
%  Copied info from the PSL 3.2 VMS version. 
%  Needed because VAX/VMS and VAX/Ultrix have different assembler.
% 27-May-87 (Leigh Stoller & Harold Carr)
%  Added global declaration of $eol$.
% 10-Jul-84 15:27:52 (Brian Beach)
%  Added missing argument to CODEPRINTF in MCPRINT and INSTRUCTIONPRINT.
% 06-Dec-83 10:00 (Brian Beach)
%   Translated from Lisp to RLisp.  Included build file stuff in this
%   file.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn (load if!-system)
        (off usermode)))

(compiletime
  (if_system vax nil 
             (progn (dskin "vax-data-machine.sl")
                    (dskin "pc:data-machine.sl")
                    )))

(global '($eol$))

(fluid '(codefilenameformat!* datafilenameformat!* inputsymfile!* 
         outputsymfile!* commentformat!* labelformat!* 
         externaldeclarationformat!* exporteddeclarationformat!* 
         fullwordformat!* doublefloatformat!* reservezeroblockformat!* 
         reservedatablockformat!* definedfunctioncellformat!* 
         undefinedfunctioncellinstructions!* mainentrypointname!* 
         codeout!* dataout!* !*lower asmopenparen!* asmcloseparen!* 
         numericregisternames!* ))

(setq codefilenameformat!* "%w.mar")

(setq datafilenameformat!* "d%w.mar")

(setq inputsymfile!* "vax.sym")

(setq outputsymfile!* "vax.sym")

(setq mainentrypointname!* (intern "pslmain"))

% this has to be lowercase                                                 
(setq numericregisternames!* 
      '["r0" "r1" "r2" "r3" "r4" "r5" "r6" "r7" "r8" "r9" "r10" "r11" 
        "r12" "r13" "r14" "r15"])

(setq commentformat!* "; %p %n")

(setq labelformat!* "%w:")

(setq externaldeclarationformat!* "	.EXTERNAL %w%n")

(setq exporteddeclarationformat!* "	.GLOBL %w%n")

(setq fullwordformat!* "	.LONG %e%n")

% FullWord expects %e for parameter                                        
(setq doublefloatformat!* "	.DOUBLE 0d%w%n")

(setq reservezeroblockformat!* "%w:	.BLKB  %e%n")

(setq reservedatablockformat!* "	.BLKB %e %n")

(setq definedfunctioncellformat!* "	jmp @#%w%n")

(setq undefinedfunctioncellinstructions!* 
      '((jsb (absolute (plus2 symfnc (times2 6 (idloc undefinedfunction)))))))

(setq asmopenparen!* '!<)

(setq asmcloseparen!* '!>)

(deflist '((remainder !%) (land !&) (lor !|) (lxor !^) (lsh !<!<)) 
         'binaryasmop)

(put 'lnot 'unaryasmop '!!)

(deflist '((t1 "r6") (t2 "r7") (t3 "r8") (t4 "r9") (t5 "r10") (nil "r11") 
           (st "sp"))
         'registername)

(put 'mkitem 'asmexpressionformat "<%e@27>+%e")

(de codefileheader ()
  (codeprintf "	.PSECT  CODE,WORD%n"))

(de datafileheader ()
  (dataprintf "	.PSECT  DATA,WORD%n"))

(de codefiletrailer ()
  (dataprintf "	.END"))

(de datafiletrailer ()
  (dataprintf "	.END"))

(de codeblockheader ()
  (codeprintf "	.ALIGN WORD%n"))

(de codeblocktrailer ()
  nil)

(de dataalignfullword ()
  (dataprintf "	.ALIGN WORD%n"))

(de printstring (s)
  (prog (n)
        (setq n (size s))
        (for (from i 0 n 1) (do (printbyte (indx s i))))
        (printbyte 0)
        (printf "	.ALIGN WORD%n")))

(de printbytelist (l)
  (foreach x in l do (printbyte x)))

(de printbyte (x)
  (printf "	.BYTE ")
  (printexpression x)
  (prin2 !$eol!$))

(de printhalfwordlist (l)
  (foreach x in l do (printhalfword x)))

(de printhalfword (x)
  (printf "	.word ")
  (printexpression x)
  (prin2 !$eol!$))

(de PrintOpcode (x)
        (prin2 x))

(de specialactionformainentrypoint ()
  (codeprintf "	.globl _main%n_main:	.word 0xff%n"))

(de asmsymbolp (x)
  (unixsymbolp (if (idp x)
                 (id2string x)
                 x)))

(de unixsymbolp (x)
  (prog (n c i)
        (setq n (size x))
        (setq c (indx x 0))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i 1)
   loop
        (when (greaterp i n)
          (return t))
        (setq c (indx x i))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !0)) (leq c (char !9))) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i (plus i 1))
        (go loop)))

(de printnumericoperand (x)
  (prin2 '!#)
  (prin2 x))

(de operandprintdeferred (x)
  (setq x (cadr x))
  (if (registerp x)
    (progn (prin2 '!()
           (printoperand x)
           (prin2 '!)))
    (progn (prin2 '!@)
           (printoperand x))))

(put 'deferred 'operandprintfunction 'operandprintdeferred)

(de operandprintdisplacement (x)
  (setq x (cdr x))
  (prin2 'L!^)
  (printexpression (cadr x))
  (prin2 '!()
  (printoperand (car x))
  (prin2 '!)))

(put 'displacement 'operandprintfunction 'operandprintdisplacement)

(de operandprintindexed (x)
  (setq x (cdr x))
  (printoperand (cadr x))
  (prin2 '![)
  (printoperand (car x))
  (prin2 '!]))

(put 'indexed 'operandprintfunction 'operandprintindexed)

(de operandprintimmediate (x)
  (prin2 '!#)
  (printexpression (cadr x)))

(put 'immediate 'operandprintfunction 'operandprintimmediate)

(de operandprintautoincrement (x)
  (prin2 '!()
  (printoperand (cadr x))
  (prin2 ")+"))

(put 'autoincrement 'operandprintfunction 'operandprintautoincrement)

(de operandprintautodecrement (x)
  (prin2 "-(")
  (printoperand (cadr x))
  (prin2 '!)))

(put 'autodecrement 'operandprintfunction 'operandprintautodecrement)

(de operandprintabsolute (x)
  (prin2 "@#")
  (printexpression (cadr x)))

(put 'absolute 'operandprintfunction 'operandprintabsolute)

(de operandprintforeignentry (x)
        (prin2 (cadr x)))

(put 'foreignentry 'operandprintfunction 'operandprintforeignentry)

(de mcprint (x)
  % Echo of MC's                                                           
  (codeprintf "; %p %n" x)
  )

(de instructionprint (x)
  (codeprintf ";         %p %n" x)
  )

(de !*cerror (x)
  (prog (i)
        (setq i (wrs nil))
        (printf "%n *** CERROR: %r %n " x)
        (wrs i)
        (return (list (list 'cerror x)))))

(put 'cerror 'asmpseudoop 'printcomment)

(defcmacro !*cerror)

(de initializesymboltable ()
  (let ((maxsymbol (compiler-constant 'maxsymbols)))
    (when (lessp maxsymbol nextidnumber*)
      (errorprintf "*** MaxSymbols %r is too small; at least %r are needed"
                   maxsymbol nextidnumber*)
      (setq maxsymbol (plus nextidnumber* 100)))
    (flag '(nil) 'nilinitialvalue)
    (put 't 'initialvalue 't)
    (setq nilnumber* (compileconstant nil))
    (dataalignfullword)
    (initializesymval)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymprp)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymnam)
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymfnc)
    (datareservefunctioncellblock
     (plus (difference maxsymbol nextidnumber*) 1))
    (initializesymget)   % SYMGET feature
    (datareserveblock (plus (difference maxsymbol nextidnumber*) 1))
    ))

(de initializesymget ()
   (dataprintgloballabel (findgloballabel 'symget))
   (for (from i 0 128 1) (do (dataprintfullword nilnumber*)))
   (foreach x in (car orderedidlist*) do
                               (dataprintfullword nilnumber*))
   )

