%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PK:FASLIN.SL 
% Title:        Loading of binary format files. 
% Author:       E. Benson 
% Created:      ??? 
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  PL:FASL-DECLS.B 
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 4-Feb-1993 (Winfried Neun)
%  Initial version fro DEC Alpha
% 10-Jan-84 (Brian Beach)
%  Commented-in checking of fasl-magic-number.
% 13-Nov-84 14:13:11 (Brian Beach)
%  Moved FASLIN-INTERN to INTERN.SL.
% 10-May-84 14:35:05 (Brian Beach)
%  Changed fasl-magic-number to faslin-magic (which is defined now in SYS-CONSTS.)
% 01-Dec-83 14:44:33 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load fasl-decls))

(compiletime (flag '(do-relocation relocate-word relocate-inf 
		     relocate-right-half compute-relocation 
		     local-to-global-id read-id-table faslin-bad-file 
		     relocate-movex) 'internalfunction))

(on fast-integers)

(fluid '(code-base-hack
	 symfnc
	 symval
	 tokenbuffer
	 argumentblock))

(de faslin (file)
  (prog (fid                  % file pointer 
	 local-id-count       % number of ids in the file
	 local-id-table       % table for mapping local ID numbers to global ID numbers.
	 code-size            % number of words of code
	 code-base            % location of the start of the code
	 init-function-address% Offset into the code of the init function
	 bit-table-size       % number of words in the bit table
	 bit-table            % the bit table
	 Btop
	 )

    % Open the file
    (setf fid (binaryopenread file))

    % Check that the first word is the correct magic number.
    (let ((first-word (binaryread fid)))
      (unless (weq first-word faslin-magic)
	(binaryclose fid)
	(faslin-bad-file file)
	(return nil)
	))

    % Read in the ID table.
    (setf local-id-table (read-id-table fid))

    % Read the code.
    (setf code-size (binaryread fid)) % Size of code segment in words

 % align 'real' code on double word boundary
    (setq nextbps (wand (wplus2 nextbps 7) (wnot 7)))

    (setf code-base (gtbps code-size)) % Allocate space in BPS
    (setq Btop (GtBPS 0))              % pointer to top of alloc. BPS
    (setf init-function-address (wplus2 code-base (binaryread fid)))
    (binaryreadblock fid (loc (wgetv code-base 0)) code-size)

    % Read the bit table
    (setf bit-table-size (binaryread fid))
    (setq bit-table (gtwrds bit-table-size))
    (setq bit-table (mkwrds bit-table))
    (binaryreadblock fid (loc (words-fetch bit-table 0)) bit-table-size)

    % Close the file
    (binaryclose fid)

    % Twiddle the bits.
    (do-relocation code-base code-size bit-table local-id-table)

    % Call the init code
    (let ((temp code-base-hack))  % avoid use of fluid binding
      (setf code-base-hack code-base)
      (flushcache code-base (wdifference btop code-base) 0 )
      (addressapply0 init-function-address)
      (setf code-base-hack temp)
      (DelBPS (wplus2 init-function-address 8) Btop)
      )
    ))


(de do-relocation (code-base code-size bit-table id-table)
  % CODE-AU-SIZE is the size of the code measured in addressing
  % units, rather than words.

  (let ((code-au-size (wtimes2 code-size addressingunitsperitem)))
    (for (from i 0 (wdifference code-au-size 1) 4)
     (do 
      (let ((bit-table-entry  (bittable (loc (words-fetch bit-table 0)) i)) 
	    (code-location    (wplus2 code-base i)))
	(case bit-table-entry
	((reloc-word) (relocate-word code-location code-base id-table))
	((reloc-inf) (relocate-inf code-location code-base id-table))
	((reloc-right-half)
	       (relocate-right-half code-location code-base id-table))
))))))
		    
		     
(de relocate-word (code-location code-base id-table)
  (let ((reloc-tag (reloc-word-tag (getmem code-location)))
	(reloc-inf (reloc-word-inf (getmem code-location))))
    (cond ((posintp (getmem code-location)) % a naked pointer from movex
	     (setf (getmem code-location)
		 (wplus2 code-base (getmem code-location)) ))
	  (t (setf (getmem code-location) % an item
		(compute-relocation reloc-tag reloc-inf code-base id-table))))
      ))

(de relocate-inf  (code-location code-base id-table)
  (let ((reloc-tag (reloc-inf-tag (getmem code-location)))
	(reloc-inf (reloc-inf-inf (getmem code-location))))
    (setf (getmem code-location)
     (mkitem (tag (getmem code-location))
      (compute-relocation reloc-tag reloc-inf code-base id-table)
      )))) % avoiding PUTFIELD
%
%
(compiletime (put 'halfword_getmem 'opencode 
		'((ldl (reg 1) (displacement (reg 1) 0))))
	     (put 'halfword_putmem 'opencode
                '((stl (reg 2) (displacement (reg 1) 0)))))

(commentoutcode
(de relocate-right-half (code-location code-base id-table)
  (let ((temp1 (halfword_getmem code-location))
	(temp2 nil))
       (setq temp2 (wand temp1 16#ffff))
       (when (local-id-number? temp2) 
		(setq temp2 (local-to-global-id temp2 id-table))
		(halfword_putmem code-location 
			(wor (wand temp1 (wnot 16#ffff)) temp2)))
       16))
)

(de relocate-right-half (code-location code-base id-table)
  (let ((temp1 (halfword_getmem code-location))
        (temp2 nil))
       (setq temp2 (wand temp1 16#7fff))
       (when (local-id-number? temp2)
                (setq temp2 (local-to-global-id temp2 id-table)))
       (when (not (weq 0 (wand temp1 16#8000)))
                % bit1 = on: xidloc
                (setq temp2 (wtimes2 temp2 addressingunitsperitem))
                (when (wgreaterp temp2 16#ffff)
                  (stderror "symfnc offset too big for xidloc relocation")))
       (halfword_putmem code-location
                        (wor (wand temp1 (wnot 16#ffff)) temp2))
       16))

  
(de compute-relocation (reloc-tag reloc-inf code-base id-table)
  (case reloc-tag
    ((reloc-code-offset)  (wplus2 code-base reloc-inf ))
    ((reloc-value-cell) 
     (cond ((local-id-number? reloc-inf) 
	      (loc (symval (local-to-global-id reloc-inf id-table))))
	   (t (loc (symval reloc-inf)))))
    ((reloc-function-cell)
     (progn
      (when (local-id-number? reloc-inf)
	(setq reloc-inf (local-to-global-id reloc-inf id-table)))
      (wplus2 symfnc                      %%% Should be (LOC (SYMFNC xxx)) ???
	      (wtimes2 addressingunitsperfunctioncell reloc-inf))))
    ((reloc-id-number)
     (if (local-id-number? reloc-inf)
       (local-to-global-id reloc-inf id-table)
       reloc-inf
       ))
    ))

(de local-to-global-id (local-id-number id-table)
  (words-fetch id-table (wdifference local-id-number 2048))
  )

(de read-id-table (fid)
  % Read in the table of local IDs at the front of the FASL file.
  % Each ID is stored as one word which holds the length, followed 
  % by the appropriate number of words holding the string.

  (let* ((local-id-count  (binaryread fid))
	 (id-table        (mkwrds (gtwrds (wplus2 local-id-count 1)))))

    (for 
     (from i 0 local-id-count)
     (do (setf (wgetv tokenbuffer 0) (binaryread fid)) 
			% word is length of ID name
	 (binaryreadblock fid  (loc (wgetv tokenbuffer 1)) 
		               (strpack (wgetv tokenbuffer 0)))
	 (setf (words-fetch id-table i) 
	   (idinf (faslin-intern (mkstr (loc (wgetv tokenbuffer 0))))))
	 ))
    id-table
    ))

(de putentry (name type offset)
  % Called by the initcode.
  % CODE-BASE-HACK is set by FASLIN before the initcode is called.
  (putd name type (mkcode (wplus2 code-base-hack offset) )))

(de faslin-bad-file (name)    % To be redefined
  (console-print-string  "File is not FASL format")
  (console-newline) )

(de delbps (bottom top)      % returns space to Bps
       (when (weq nextbps top)
	    (flushcache bottom (wdifference nextbps bottom) 1 )
            (setq  nextbps bottom)))

(de flushcache (addr len inv) (cflush addr len inv))

(off fast-integers)
