%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         YFASLIN.SL
% Description:  Fasl as modified for the CRAY Y-MP.
% Author:       Winfried Neun, ZIB Berlin , based on X-MP version
% Created:      29-Feb-88
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, University of Utah, all rights reserved
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime(flag '(bittable crayfaslputparcel crayfaslgetparcel
                    faslput-2-parcel-reverse faslget-2-parcel-reverse
                    faslput-2-parcel faslget-2-parcel)
                  'internalfunction))
 
% Do the definitions.
 
(compiletime (progn
 (define-constant  reloc_word  1)
 (define-constant  reloc_id_number 2)
 (define-constant  escape_to_second 3)
 (define-constant  parcel_label 0)
 (define-constant  word_label 1)
 (define-constant  reloc_value_cell 2)
 (define-constant  reloc_function_cell 3)
 (define-constant  faslmagic 199)
 (define-constant  faslmagic1 99)))
 
(fluid '(code-base-hack symfnc symval tokenbuffer optionaddresses*
        argumentblock))
 
(CompileTime
 (progn
  (ds LocalIDNumberP (U) (wgeq U 2048))
  (ds LocalToGlobalID (U) (words-fetch IDTable (wdifference U 2048)))
  (ds ExtraArgumentP (U) (wgeq U 8150))  % Something enough less than 8192
  (ds MakeExtraArgument (U)
    (wdifference U (wplus2 8150 (wplus2 MaxRealRegs 1)))) nil))
 
(de FaslIN (File)
  (prog (F N M IDTable CodeBase CodeSize OldCodeBase
         E BT BitTable R RT RI BI Btop actbps)
    (setq actbps nextbps)
    (setq F (BinaryOpenRead File))
    (setq N (BinaryRead F))           % First word is magic number
 
    (when (and (not (weq N FASLMAGIC)) (not (eq N Faslmagic1)))
           (console-print-string  "*** File is not FASL format")
           (return NIL))
 
    (setq M (BinaryRead F))           % Number of local IDs
    (setq IDTable (mkwrds(gtwrds(wplus2 m 1))))% Allocate space for table
    (Ifor (from I 0 M 1)
       (do (progn
            (setf (wgetv tokenbuffer 0)  (BinaryRead F))     % length
            (BinaryReadBlock F
                         (loc (wgetv tokenbuffer 1))
                         (StrPack (wgetv tokenbuffer 0)))
            (setf (words-fetch  IDTable I)
             (IDInf (faslin-intern(mkstr(loc(wgetv tokenbuffer 0)))))
    )  )   ))
 
    (setq CodeSize (BinaryRead F))    % Size of code segment in words
    (setq CodeBase (GtBPS CodeSize))    % Allocate space in BPS
    (setq Btop (GtBPS 0))               % pointer to top of alloc. BPS
    (setq E (wplus2(wshift CodeBase 2)(BinaryRead F)))     % offset
                                        % Will be called after code is read
                                        % Put the next N words there
    (BinaryReadBlock F (loc (wgetv CodeBase 0)) CodeSize)              %  h
    (setq N (BinaryRead F))           % Next word is size of bit table in word
    (setq BitTable (mkwrds(gtwrds n)))  % Allocate space for bit t ab
    (setq BT BitTable)                  % BitTable alias
    (BinaryReadBlock F                  % read bit table
           (loc (words-fetch BitTable 0)) N)
    (BinaryClose F)                     % close the file
    (setq CodeSize
      % Change for cray to be parcel address.
      (wdifference (wshift CodeSize 2) 1 ))
 
    (Ifor (from I 0 CodeSize 1)
      (do (progn
            % Need to build bittable function, which will return the
            % bittable entry based upon a starting word address, and
            % an index.
         (setq R (BitTable BT I))
            % R has the 2 bit entry in the bittable for the current
            % parcel.
         (setq BI (wplus2 (wshift CodeBase 2) I))
            % BI has the parcel address of the current parcel in the
            % code space.
            % (quotient BI 4) is the word address of the current parcel.
         (case R
           ((RELOC_WORD)
             (progn
               (setf (wgetv (wshift BI -2) 0)
                     (wplus2 CodeBase (wgetv (wshift BI -2) 0))  )
               (setq i (wplus2 i 3))))
           ((RELOC_ID_NUMBER)
              (progn
               (setq RT (BitTable BT (wplus2 I 1)))
               (case RT
                 ((0) (progn
                    (setf RI (crayfaslgetparcel(wplus2 BI 1)))
                    (cond ((LocalIDNumberP RI)
                            (setq RI (LocalToGlobalID RI))))
                    (crayfaslputparcel (wplus2 BI 1) RI)))
                 ((1) (progn  % 3 parcel instruction on Y-MP
                    (setf RI (faslget-2-Parcel-reverse BI))
                    (setf N (wand RI 15))
                    (setq RI (wshift (wand RI 16#FFFF0) -4))
                    (cond ((localidnumberp RI)
                           (setq RI (localtoglobalid RI))))
                    (faslput-2-Parcel-reverse BI
                      (Wor (Wand (faslget-2-Parcel-reverse BI)
                                 (wshift 16#FFF 20))
                           (Wor (Wshift RI 4) N))))))
               (Setf i (Wplus2 i 1))))
           ((ESCAPE_TO_SECOND)
            (progn        % RT now has the second bittable entry.
             (setq RT (BitTable BT (wplus2 I 1)))
             (case RT
              ((PARCEL_LABEL)
                (progn
                 (setq RI (crayfaslgetparcel BI ))
                 (if (eq 8#020 (wrshift RI 9)) %Ploc
                  (faslput-2-Parcel-reverse (wplus2 BI 1)
                           (wplus2 (wlshift CodeBase 2)
                              (faslget-2-parcel-reverse (wplus2 BI 1))))
                  (faslput-2-Parcel BI (wplus2 (wshift CodeBase 2)
                                           (faslget-2-parcel BI))))))
              ((WORD_LABEL)
                    % We want to add in the code base to the 2
                    % parcels and then turn into a word address.
                  (faslput-2-Parcel-reverse BI
                    (wplus2 CodeBase (faslget-2-Parcel-reverse BI))))
              ((RELOC_VALUE_CELL)   % 3 parcel instruction on Y-MP
                    (progn
                     (setf RI (faslget-2-parcel-reverse BI))
                     (cond
                      ((ExtraArgumentP RI)
                       (setq RI (wplus2 (loc(wgetv ArgumentBlock 0))
                                        (MakeExtraArgument RI))))
                      ((LocalIDNumberP RI)
                       (setq RI (wplus2 (loc(wgetv symval 0))
                                        (LocalToGlobalID RI))))
                      (t (setq RI (wplus2 (loc(wgetv symval 0))
                                          RI))))
                     (faslput-2-Parcel-reverse BI RI)))
              ((RELOC_FUNCTION_CELL)
                   (progn
                    (setf RI (crayfaslgetparcel (wplus2 BI 1)))
                    (when (LocalIDNumberP RI)
                          (setq RI (LocalToGlobalID RI)))
                    (faslput-2-Parcel BI
                         (wplus2 (wshift (crayfaslgetparcel BI) 16)
                             (wplus2 % (calcparcel 'SymFnc)
                                 (wshift (loc(wgetv symfnc 0)) 2)
                                 (wshift RI 2 )))))))
               (setq i (wplus2 1 i))))))))
 
    (setq OldCodeBase code-base-hack)
    (setq IDTable (setq BitTable (setq BT NIL))) % give up aux space
    (setq code-base-hack CodeBase)
    (Bittable Bittable 0 BT IDTable)    % strange things to fool compile
    (AddressApply E nil)                % Call the init routine
    (setq code-base-hack OldCodeBase)   % restore previous value for CodeBase
    (DelBPS (wplus2 (wquotient E 4) 1) BTop)% dealloc space of initcode
    (setq optionaddresses*
          (cons (cons file (cons actbps (wrshift e 2)) )
                 optionaddresses* ))
))
                                        %
 
% Access the bittable at the word-address + offset number of bittable
%  entries, and return the bits.
 
(de bittable (Word-Address Offset)
  (prog (mem)
    (setq Word-Address (wplus2 Word-Address (wrshift Offset  5)))
    (setq mem (wgetv Word-Address 1))   % memory contents
    (setq Offset (wand Offset 31))      % bit number
    (setq Offset (wdifference 62 (wplus2 Offset Offset)))  % shiftsize
    (return (wrshift (wand (wlshift 16#3 Offset) mem) Offset))
) )
 
% FaslIn Versions of Parcel-Routines (internal callable)
 
% Get 2 parcel elements from the argument address and return as a
% single parcel.
 
(de faslget-2-Parcel (Parcel-Address)
      (wor (wshift (crayfaslgetparcel Parcel-Address) 16)
              (crayfaslgetparcel (wplus2 Parcel-Address 1))))
 
(de faslget-2-Parcel-Reverse (Parcel-Address)
      (wor (crayfaslgetparcel Parcel-Address)
         (wshift (crayfaslgetparcel (wplus2 Parcel-Address 1)) 16 )))
 
% Store the 2 parcel argument into memory at the parcel address
% provided.
 
(de faslput-2-Parcel (Parcel-Address Two-Parcels)
  (crayfaslputparcel Parcel-Address (wshift Two-Parcels -16))
  (crayfaslputparcel (wplus2 1 Parcel-Address)
             (wand 16#FFFF Two-Parcels)))
 
(de faslput-2-Parcel-Reverse (Parcel-Address Two-Parcels)
 (crayfaslputparcel (wplus2 Parcel-Address 1) (wrshift Two-Parcels 16))
  (crayfaslputparcel Parcel-Address
                                    (wand 16#FFFF Two-Parcels)))
 
(fluid '(*writingfaslfile ))
 
(de CrayfaslputParcel (paddress Parcel)
  (prog (waddress pnumber shiftsize mem)
        (setq waddress (wrshift paddress 2))
        (setq mem (getmem waddress))
        (setq pnumber (wand paddress 3))
        (setq shiftsize (wdifference 48(wshift pnumber 4)))
        (putmem waddress (wor (wlshift parcel shiftsize)
                              (wand (wnot (wlshift 16#FFFF shiftsize))
                                        mem          )))))
 
 
(de CrayfaslgetParcel (paddress)
  (prog (waddress pnumber shiftsize mem)
        (setq waddress (wshift paddress -2))
        (setq mem (getmem waddress))
        (setq pnumber (wand paddress 3))
        (setq shiftsize (wdifference 48(wshift pnumber 4)))
        (return (wand (wrshift mem shiftsize)
                      16#FFFF))   ))
 
 
% General Purpose Versions of Parcel-Routines
 
% Get 2 parcel elements from the argument address and return as a
% single parcel.
 
(de Get-2-Parcel (Parcel-Address)
  (prog (Return-Parcels)
    (setf Return-Parcels
      (wplus2 (wshift (craygetparcel Parcel-Address) 16)
              (craygetparcel (wplus2 Parcel-Address 1))))
    (return Return-Parcels)))
 
% Store the 2 parcel argument into memory at the parcel address
% provided.
 
(de Put-2-Parcel (Parcel-Address Two-Parcels)
  (crayputparcel Parcel-Address (wshift Two-Parcels -16))
  (crayputparcel (wplus2 1 Parcel-Address)
             (wand 16#FFFF Two-Parcels)))
 
(fluid '(*writingfaslfile ))
 
(de CrayPutParcel (paddress Parcel)
  (prog (waddress pnumber shiftsize mem)
        (cond (*writingfaslfile (auto-update-bittable)))
        (setq waddress (wrshift paddress 2))
        (setq mem (getmem waddress))
        (setq pnumber (wand paddress 3))
        (setq shiftsize (wdifference 48(wshift pnumber 4)))
        (putmem waddress (wor (wlshift parcel shiftsize)
                              (wand (wnot (wlshift 16#FFFF shiftsize))
                                    mem )))))
 
 
(de CrayGetParcel (paddress)
  (prog (waddress pnumber shiftsize mem)
        (setq waddress (wrshift paddress 2))
        (setq mem (getmem waddress))
        (setq pnumber (wand paddress 3))
        (setq shiftsize (wdifference 48(wshift pnumber 4)))
        (return (wand (wrshift mem shiftsize) 16#FFFF))
                                  ))
 
(fluid '(show-new-ids))
 
(de putentry (f e g)
   (cond (show-new-ids (progn
                         (putoct (wplus2 Code-Base-Hack g))
                         (terminalwritechar 32 32)
                         (console-print-string (symnam (idinf f)))
   )     )             )
 
   (putd f e (mkcode (wplus2 Code-Base-Hack g))))
%(faslend) (quit)
