/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXCC:UNIX-IO.C
% Description:  Unix PSL FileDescriptors are implemented as stdio streams
%                 ("FILE *".)
% Author:       Russell D. Fish
% Created:      Thu Feb 16 1984
% Modified:     17-Jul-84 22:49:12 (RAM)
% Mode:         Text
% Package:
% Status:       Experimental (Do Not Distribute)
%
% Copyright (c) 1984 Hewlett-Packard Company, all rights reserved.
% Copyright (c) 1984 University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 15-Sep-88 (T. Yamamoto and C. Burdorf)
%  Moved collect out of expand_file_name so it won't be overwritten as a
%  local before it is referenced.
% 20-Sep-86 (Leigh Stoller)
%  Removed assembler alias statements because they are not portable. Instead,
%  a sed script will be used to convert the _variables of C to VARIABLES of
%  PSL.
% 17-Jul-84 22:48:32 (RAM)
%  Added unixcd, a routine that calls expand_file_name before calling chdir.
% 3-Jul-84 10:45:57 (RAM)
%  Added expand_file_name (called from unixopen) to expand shell variable
%  references and ~ home directory syntax.
% 29-Jun-84 14:21:21 (RAM)
%  Added unixputs and unixopen.
% 21-May-84 17:41:41 (Vicki O'Day)
%  Added unixcleario.  It is called by syscleario.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/
 
#include <stdio.h>
#include <setjmp.h>

#define INCL_DOSFILEMGR
#include <os2.h>


#define TOGGLE 26   
#define ACKCHAR 5 

/* Initialize some PSL external variables with FileDescriptors for SysClearIO.
 */
extern FILE * unixstdin, * unixstdout, * unixstderr, * unixtty;
extern FILE * scriptin, * scriptout; 
extern int scriptmode;
extern jmp_buf mainenv;
extern int win_mode;

/* Import NULL and EOF constants for error returns from stdio fns.
 */
extern int unixnull, unixeof;
 
FILE * select_file_in(FILE* f)
   { 
     if(f == unixstdin && scriptin != NULL)
     return(scriptin);
       else
     return(f);
   }

FILE * select_file_out(FILE* f)
   { 
     if(f == unixstdout && scriptout != NULL) 
     return(scriptout);
       else
     return(f);
   }


/* Tag( unixinitio )
 */
unixinitio()
{
    unixstdin = stdin;
    unixstdout = stdout;
    unixstderr = stderr;
    unixnull = NULL;
    unixeof = EOF;
    unixtty = stdin; /* fopen("/dev/tty", "r"); */
}
 

wputs(char * str,long l)
  /* window: write string and sync */
  {
    char c;
    
    while(l-- > 0)
    { fputc(c = *str++ , stdout);
      if(c == '\n')
      {
	  fputc(ACKCHAR, stdout);
	  while(ACKCHAR != fgetc(stdin));
      }
    }
   }

unixputc(c)
char c;
{
    if(win_mode) wputs(&c,1L);
    else fputc(c, stdout);
}
 

unixputs(str)
char *str;
{
    if(win_mode) wputs(str,(long)strlen(str));
    else fputs(str, stdout);
}
 

unixputn(n)
int n;
{
    fprintf(stdout, "%x", n);
}
 
unixfread(char*buf, int size,int count, FILE*fp)
   { 
     if(fp == unixstdin && win_mode) unixputc(TOGGLE);
     return(fread(buf,size,count,select_file_in(fp)));
   }

unixfputc(char ch,FILE*fp)
   {
     if(fp == unixstdout && scriptmode==1)
       { 
	 fputc(ch,unixstdout);
	 return(fputc(ch,scriptout));
       }
       else
     if(win_mode && fp==unixstdout) wputs(&ch,1L);
     else fputc(ch,select_file_out(fp));
   }

unixfgetc(FILE*fp)
   {
     if(fp == unixstdin && win_mode) unixputc(TOGGLE);
     return(fgetc(select_file_in(fp)));
   }

int unixfgets(char*buf,size_t count,FILE*fp)
   {
     if(fp == unixstdin && win_mode) unixputc(TOGGLE);
     return(fgets(buf,count,select_file_in(fp)));
   }

unixfwrite(char*buf,int len,int count,FILE*fp)
   {
     if(fp == unixstdout && scriptmode==1)
       { 
	 fwrite(buf,len,count,unixstdout);
	 return(fwrite(buf,len,count,scriptout));
       }
       else
     if(win_mode && fp == unixstdout)
	 wputs(buf,(long)(len*count));
     else fwrite(buf,len,count,select_file_out(fp));
   }

unixfflush(FILE*fp)
   {
     if(fp == unixstdout && scriptmode==1)
       { 
	 fflush(unixstdout);
	 return(fflush(scriptout));
       }
       else
     return(fflush(select_file_out(fp)));
   }


unixcleario()
{
    unixinitio();
}
 
/* The function expand_file_name is used to expand shell variable references
   and ~ notation for directories in file names before calling fopen.
   This eliminates the need for the HPUX-PATH and VAX-PATH kludges that were
   once required, however, such a mechanism may still be used to override the
   values supplied by the shell environment.  The file name is first copied
   to a temporary copy because the parsing algorithm must write into the
   string (this may change).  For now, the maximum string length supported
   is 255 characters, but no checking is done to see if this is exceeded.
   $ and ~ variables are expanded in one pass by calling the functions getenv,
   getuid, getpwuid, and getpwnam.  If any expansion fails, the original
   string is returned.
*/
 
#include <stdlib.h>

char collect[255], copy[255];  /* Made global so it won't be overwritten 
				  Used to be local to expand_file_name */

/* Tag( expand_file_name )
 */
char *expand_file_name(fname)
char *fname;
{
  char *c, *t, *e, *s, save;
  c = copy;
  s = fname;
  while (*c++ = *s++);
  s = copy;
  c = collect;
  *c = '\0';
  while (*s)
    {
      if (*s == '$')
	{
	  for (e = ++s; (*e != '/' && *e != '\\'  && *e != '.'
			  && *e != '\0' && *e != '$'); e++)
	    ;
	  t = 0;                        /* default initialization */
	    {
	      save = *e;
	      *e = '\0';
	      t = getenv(s);
	      *e = save;
	      s = e;
	    }
	  if (t)
	    while (*c++ = *t++)
	      ;
	  else
	    return(fname);   /* name not found, just return original fname */
	  c--;
	}
    for (; (*s != '\0' && *s != '$'); *c++ = *s++)
      ;
      *c = '\0';
  }
  return (collect);
}

/* Tag( unixopen )
 */
unixopen(filename, type)
     char *filename, *type;
{
  int fptr;
 
  /*  printf("open %s %s ",filename,type);    */
  fptr = (int) fopen(expand_file_name(filename), type);
  if(fptr==NULL)
  { /* try file name in dos syntax */
    char c,nfname[255];
    int i,j,k,kmax;
    /*  printf("open failed %s %s ",filename,type); /*  */
    k=0;kmax=8;j=0;
    for(i=0;filename[i];i++)
    { c=filename[i]; nfname[j++]=c; k++; 
      if(c == '\\')     {k=0; kmax=8;}
      else if(c == '.') {k=0; kmax=3;}
      else if(k > kmax) j--;
    };
    nfname[j]='\0';
    /*   printf(" reformatted  %s  ",nfname); /*  */
    fptr = (int) fopen(expand_file_name(nfname), type);
    /*   printf(" --> %x\n",fptr);  /* */
  };
  return(fptr);
}
 

unixcd(filename)
     char *filename;
{
   ULONG d; char c;
   filename = expand_file_name(filename);
   if(filename[1] == ':') 
   { c=filename[0];
     if(c>= 'a') d=c-'a'; else d=c-'A';
     d=DosSetDefaultDisk(d+1);
   }
   chdir(filename);
}
 

external_system(command)
     char *command;
{
  int value;
  value = system(command);
  return(value);
}
 

external_exit(status)
     int status;
{
  if(status>=0) status=1+status; else status=1;
  longjmp(mainenv,status);
  exit(status);
}
 
char *static_argv[20];  /* static place to hold argv so it doesn't get gc'd */
 
copy_argv(argc,argv)    /* copy argv into static space. */
int argc;
char *argv[];
{
  int i;
 
  for (i=0; i < argc; i++)
     static_argv[i]=argv[i];
 
  return(static_argv);
}

long getw(FILE * f)
   { 
     long pu;
     fread(&pu,4,1,f);
     return(pu);
   }

putw(long w,FILE * f)
   { 
     fwrite(&w,4,1,f);
     return(1);
   }

