%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXK:MAIN-START.SL
% Title:          First routine called on startup
% Author:         Brian Beach, Hewlett-Packard CRC
% Created:        16-Feb-84
% Modified:       19-Feb-85 09:36:17
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 20-Jun-89 (Winfried Neun)
%  added reference to symget vector as internal structure
% 31-Aug-88 (Julian Padget)
%  Changed fluids to globals, since latter needed in CPSL.
% 28-Oct-87 (Harold Carr)
%  Made panic exit call exit-with-status, rather than redundantely doing
%  what panic exit already did then calling external_exit.
% 13-Apr-87 (Harold Carr & Leigh Stoller)
%  Added external fluid _infbitlength_ to be used in bpsheap.c.
% 26-Sep-86 (Leigh Stoller)
%  Changed exit to unix to be a call to exit library function. The Sun does
%  accept the exit status in (reg 1) like the Bobcat. Called it external_exit
%  so it does not interfere with the cmacro of the same name.
% 22-Sep-86 (Leigh Stoller)
%  Commented out calls 68020_advise which is not defined on the sun.
% 18-Jul-86 (Leigh Stoller)
%  Changed the allocation of argumentblock to be the number of maxregs since
%   it is indexed by base + reg number.
% 15-Feb-85 13:27:53 ()
%  Made it so that PSL allocates the stack it will use.
%  Removed TST.B for stack expansion.
%  Included declaration of STACKSIZE (should go in SYS-CONSTS eventually).
% 9-Jul-84 09:11:55 (RAM)
%  Added fluids (and initialization) unixstdin, unixstdout, unixstderr,
%  unixnull, unixeof, unixtty.
% 3-Jul-84 10:51:27 (RAM)
%  Changed name of initial faslin file to $pnkl/load-psl.b since
%  binaryopenread can now expand shell variable references.
% 29-Jun-84 12:33:38 (RAM)
%  Chagged exit_to_unix to exit-to-unix to match call in kernel-fatal-error.
%  Removed save and restore of A6.
% 31-May-84 08:15:32 (Brian Beach)
%  Changed SYSCLEARIO call to HP_CLEARIO.  SYSCLEARIO is now called by the 
%  loader as part of its startup.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(commentoutcode
(compiletime    %% Move this hack somewhere else later.
 (defmacro exported-fluid (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'exportedsymbol)
         (put id 'symbol id)
         ))
   `(fluid ,list)
   ))
)

(compiletime    %% Move this hack somewhere else later.
 (defmacro exported-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'exportedsymbol)
         (put id 'symbol id)
         (put id 'vartype 'global)
         ))
   ))

(commentoutcode
(compiletime    %% Move this hack somewhere else later.
 (defmacro internal-fluid (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'internalsymbol)
         (put id 'symbol id)
         ))
   `(fluid ,list)
   ))
)

(compiletime    %% Move this hack somewhere else later.
 (defmacro internal-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'internalsymbol)
         (put id 'symbol id)
         (put id 'vartype 'global)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro system-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (put id 'vartype 'global)
         ))
   ))

(on fast-integers)

(compiletime (load sys-consts))

%%(dskin "my-consts.sl")

(declare-kernel-structure 
 (stack             #.(compiler-constant 'stacksize))
 (argumentblock     #.(compiler-constant 'maxargs))
%(theirregs         #.(compiler-constant 'savesize))
%(temptheirregs     #.(compiler-constant 'savesize))
 (tokenbuffer       #.(compiler-constant 'maxtokensize))
 (bndstk            #.(compiler-constant 'bndstksize))
 (catchstack        #.(times2 (compiler-constant 'catchstacksize) 4))
 (hashtable         #.(quotient (compiler-constant 'hash-table-size) 2))
 (onewordbuffer     1)
 (saveargc          1)
 (saveargv          1)
 (datebuffer        1)
 )

(exported-global
 '(bndstkptr
   bndstkupperbound
   gcarraylowerbound
   gcarrayupperbound
   heaplast
   oldheaplast
   heaplowerbound
   heaptrapbound
   oldheaptrapbound
   heapupperbound
   _infbitlength_
   lastbps
   bpslowerbound
   mainstartinitialize
   nextbps
   oldheapupperbound
   oldheaplowerbound
   stacklowerbound
   stackupperbound
   saveda7
   unixstdin
   unixstdout
   unixstderr
   unixnull
   unixeof
   unixtty
   ))

(internal-global
 '(symget
   symnam
   symval
   symfnc
   symprp
   ))

(system-global
 '(symget
   symnam
   symval
   symprp
   symfnc
%  stack
   catchstackptr
   catchstack
   bndstklowerbound
   bndstkupperbound
   bndstkptr
   bndstk
   heap
   nextsymbol    %% Had SYMBOL as its assembler symbol.  why????
   % Add these so they are always available to compiled code.
   ***MUST-BE-NIL***
   ***MUST-BE-NIL-TOO***
   *fastcar
   ))

(de init-pointers()
% (setq stacklowerbound (wplus2 stack stacksize))
% (setq stackupperbound (wplus2 stack 100))
  (setq catchstackptr catchstack)
  (setq bndstklowerbound bndstk)
  (setq bndstkupperbound (loc (wgetv bndstk (wdifference bndstksize 1))))
  (setq bndstkptr bndstk)
  % most of the heap pointers are set up in the pascal world..
  % i.e. heaplowerbound, heaptrapbound, heaplast, heapupperbound, heaptrapbound.
  (setq heap heaplowerbound)
)

(de init-fluids ()
        % Until problem of labels not appearing is fixed...
  (setq gcarraylowerbound nil)
  (setq gcarrayupperbound nil)
  (setq heaplast nil)
  (setq oldheaplast nil)
  (setq heaptrapbound nil)
  (setq oldheaptrapbound nil)
  (setq heapupperbound nil)
  (setq _infbitlength_ nil)
  (setq lastbps nil)
  (setq bpslowerbound nil)
  (setq mainstartinitialize nil)
  (setq nextbps nil)
  (setq oldheapupperbound nil)
  (setq oldheaplowerbound nil)
  (setq stackupperbound nil)
  (setq unixstdin nil)
  (setq unixstdout nil)
  (setq unixstderr nil)
  (setq unixnull nil)
  (setq unixeof nil)
  (setq unixtty nil)
  (setq ***MUST-BE-NIL*** NIL)
  (setq ***MUST-BE-NIL-TOO*** NIL)
  (setq *fastcar nil)
)

(lap '((*entry !m!a!i!n expr 0)

       (*move (displacement (reg st) 4) (fluid argc))
       (*move (displacement (reg st) 8) (fluid argv))

       (*move   (fluid stack) (reg st))
       (*move   (reg st)      (fluid stackupperbound))
       (*wplus2 (reg st)      (wconst (times (sub1 stacksize) 
					     addressingunitsperitem)))
       (*move   (reg st)      (fluid stacklowerbound))

       %  Do OS specific initializations (uses argc and argv)
       (*move (fluid argc) (reg 1))
       (*move (fluid argv) (reg 2))
       (*move infbitlength (fluid _infbitlength_))
       (*link os_startup_hook expr 2)

       (*call init-pointers)

       (*call pre-main)                                 % call PSL

panic-exit                      % need to do UNIX cleanup after
                                % a fatal error, so jump here
       (*move (wconst 0) (reg 1))
       (*linke 0 exit-with-status expr 1)
       (*exit 0)

       (*entry exit-with-status expr 1)
       (*push (reg 1))
       (*link os_cleanup_hook expr 0)
       (*pop (reg 1))
       (*move (fluid saveda7) (reg st))
       (*link external_exit expr 1)
       (*exit 0)
       ))

(de pre-main ()
  (unixcleario)
  (initialize-symbol-table)
  (initcode)
  (console-print-string "About to load LOAD-PSL")
  (console-newline)
  (faslin "load-psl.b")
  (loader-main)
  )

(off fast-integers)
