%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:COPIERS.SL 
% Title:        Functions for copying various data types 
% Author:       Eric Benson 
% Created:      27 August 1981 
% Modified:     19 Oct 1984 0617-PDT (Mark R. Swanson)
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 29-Feb-96 Winfried Neun (for the sake of this special leapyear day).
%  Invented copyfltn in Totalcopy such that alignment of floats is not lost
% 17-Feb-89 Winfried Neun, ZIB Berlin,  Added a fast (lapped) copier for Sun4
% 4-Jul-85 Edited by Connie O'Dell to install fix for uninitialized vectors
%  in totalCopy proposed by Will Galway, to prevent unpleasant consequences
%  of garbage collecting while part of copy structure is uninitialized.
% 19 Oct 1984 0616-PDT (Mark R. Swanson)
%  Modified length calculation in COPYSTRINGTOFROM to take count word out of 
%   consideration
% 18-Jul-84 08:08:34 (Brian Beach)
%  Minor cleanup of TOTALCOPY
% 01-Dec-83 14:26:58 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(compiletime (flag '(fast-copier) 'internalfunction))
(compiletime (off lapopt))

(lap '((*entry fast-copier expr 4) % this is for SUN4 only
   %
   % copying using lots of registers
   %
   % parm 1: source            (will be inf'ed)
   % parm 2: destination       (will be inf'ed)
   % parm 3: number of items to copy
   %  
   % parm4 : is returned, such  allowing linke!

      (*alloc 12)
      (andn (reg 1) (reg i5) (reg 1))
      (*move (reg 4) (reg o5))        % for return
      (andn (reg 2) (reg i5) (reg 2)) % addresses inf'ed now
      (*move 15 (reg 4))
      (and (reg 3) (reg 4) (reg 5))   % the remainder mod 16 
      (*wshift (reg 3) -4)            % length / 16
      (*jumpwleq (label exitlupe) (reg 3) 0)
  lupe 
      (*move (indexed (reg 1) 0)  (frame 1))
      (*move (indexed (reg 1) 4)  (frame 2))
      (*move (indexed (reg 1) 8)  (frame 3))
      (*move (indexed (reg 1) 12) (frame 4))
      (*move (indexed (reg 1) 16) (frame 5))
      (*move (indexed (reg 1) 20) (frame 6))
      (*move (indexed (reg 1) 24) (frame 7))
      (*move (indexed (reg 1) 28) (frame 8))
      (*move (indexed (reg 1) 32) (frame 9))
      (*move (indexed (reg 1) 36) (frame 10))
      (*move (indexed (reg 1) 40) (frame 11))
      (*move (indexed (reg 1) 44) (frame 12))
      (*move (indexed (reg 1) 48) (reg o3))
      (*move (indexed (reg 1) 52) (reg o0))
      (*move (indexed (reg 1) 56) (reg o1))
      (*move (indexed (reg 1) 60) (reg o2))
      (*move (frame 1)  (indexed (reg 2) 0))
      (*move (frame 2)  (indexed (reg 2) 4))
      (*move (frame 3)  (indexed (reg 2) 8))
      (*move (frame 4)  (indexed (reg 2) 12))
      (*move (frame 5)  (indexed (reg 2) 16))
      (*move (frame 6)  (indexed (reg 2) 20))
      (*move (frame 7)  (indexed (reg 2) 24))
      (*move (frame 8)  (indexed (reg 2) 28))
      (*move (frame 9)  (indexed (reg 2) 32))
      (*move (frame 10) (indexed (reg 2) 36))
      (*move (frame 11) (indexed (reg 2) 40))
      (*move (frame 12) (indexed (reg 2) 44))
      (*move (reg o3) (indexed (reg 2) 48))
      (*move (reg o0) (indexed (reg 2) 52))
      (*move (reg o1) (indexed (reg 2) 56))
      (*move (reg o2) (indexed (reg 2) 60))
      (*wplus2 (reg 3) -1)
      (*wplus2 (reg 1) 64)
      (*wplus2 (reg 2) 64) 
      (*jumpwgreaterp (label lupe)(reg 3) 0) 
   exitlupe
      (*wdifference (reg 4) (reg 5))  % 15 -rem (lng,16)
      (*wshift (reg 4) 3)             % * 3 (8 bytes)
      (move-relocatable-object "Hugo" (reg 5))
      (jmpl (displacement (reg 5) (reg 4)) (reg 0))
      (or (reg 0) 0 (reg 0))
    "Hugo"   
      (*move (indexed (reg 1) 56) (reg o1))
      (*move (reg o1) (indexed (reg 2) 56))
      (*move (indexed (reg 1) 52) (reg o0))
      (*move (reg o0) (indexed (reg 2) 52))
      (*move (indexed (reg 1) 48) (reg o3))  
      (*move (reg o3) (indexed (reg 2) 48))
      (*move (indexed (reg 1) 44) (frame 12))
      (*move (frame 12) (indexed (reg 2) 44))
      (*move (indexed (reg 1) 40) (frame 11))
      (*move (frame 11) (indexed (reg 2) 40))
      (*move (indexed (reg 1) 36) (frame 10))
      (*move (frame 10) (indexed (reg 2) 36))
      (*move (indexed (reg 1) 32) (frame 9))
      (*move (frame 9)  (indexed (reg 2) 32))
      (*move (indexed (reg 1) 28) (frame 8))
      (*move (frame 8)  (indexed (reg 2) 28))
      (*move (indexed (reg 1) 24) (frame 7))
      (*move (frame 7)  (indexed (reg 2) 24))
      (*move (indexed (reg 1) 20) (frame 6))
      (*move (frame 6)  (indexed (reg 2) 20))
      (*move (indexed (reg 1) 16) (frame 5))
      (*move (frame 5)  (indexed (reg 2) 16))
      (*move (indexed (reg 1) 12) (frame 4))
      (*move (frame 4)  (indexed (reg 2) 12))
      (*move (indexed (reg 1) 8)  (frame 3))
      (*move (frame 3)  (indexed (reg 2) 8))
      (*move (indexed (reg 1) 4)  (frame 2))
      (*move (frame 2)  (indexed (reg 2) 4))
      (*move (indexed (reg 1) 0)  (frame 1))
      (*move (frame 1)  (indexed (reg 2) 0))  
      (*move (reg o5) (reg 1))               % return destination
      (*exit 13)))

(compiletime (on lapopt))
(de copystringtofrom (new old)
  % Copy all chars in Old to New
  (prog (slen stripnew stripold)
	(setq stripnew (strinf new))
	(setq stripold (strinf old))
	(setq slen (strlen stripold))
	(when (wlessp (strlen stripnew) slen)
	  (setq slen (strlen stripnew)))
        (setq slen (- (strpack slen) 1))
%       (for (from i 0 slen 1)
%              (do (setf (vecitm stripnew i) (vecitm stripold i))))
%       (return new)))
	(return (fast-copier (+ stripold 4) (+ stripnew 4) (+ slen 1) new))))  

(de copystring (s)
  % copy to new heap string
  (prog (s1)
	(setq s1 (gtstr (strlen (strinf s))))
	(copystringtofrom s1 (strinf s))
	(return (mkstr s1))))

(de copybytes (s)
  % copy to new heap string
  (prog (s1)
	(setq s1 (gtstr (strlen (strinf s))))
	(copystringtofrom s1 (strinf s))
	(return (mkbytes s1))))

(de copywarray (new old uplim)
  % copy UpLim + 1 words
% (progn (for (from i 0 uplim 1) (do (setf (wgetv new i) (wgetv old i))))
%         new))
  (fast-copier old new (+ uplim 1) new))  

(de copyvectortofrom (new old)
  % Move elements, don't recurse
  (prog (slen stripnew stripold)
	(setq stripnew (vecinf new))
	(setq stripold (vecinf old))
	(setq slen (veclen stripold))
%       (for % assumes VecLen New has been set
%
%             (from i 0 slen 1)
%             (do (setf (vecitm stripnew i) (vecitm stripold i))))
%       (return new)))
  (return (fast-copier (+ stripold 4) (+ stripnew 4) (+ slen 1) new))))  

(de copyvector (s)
  % Copy to new vector in heap
  (prog (s1)
	(setq s1 (gtvect (veclen (vecinf s))))
	(copyvectortofrom s1 (vecinf s))
	(return (mkvec s1))))

(de copyevectortofrom (new old)
  % Move elements, don't recurse
  (prog (slen stripnew stripold)
	(setq stripnew (evecinf new))
	(setq stripold (evecinf old))
	(setq slen (eveclen stripold))
%       (for % assumes EVecLen New has been set
%
%             (from i 0 slen 1)
%             (do (setf (evecitm stripnew i) (evecitm stripold i))))
%       (return new)))
  (return (fast-copier (+ stripold 4) (+ stripnew 4) (+ slen 1) new))))  

(de copyevector (s)
  % Copy to new EVector in heap
  (prog (s1)
	(setq s1 (gtevect (eveclen (evecinf s))))
	(copyevectortofrom s1 (evecinf s))
	(return (mkevect s1))))

(de copywrdstofrom (new old)
  % Like CopyWArray in heap
  (prog (slen stripnew stripold)
	(setq stripnew (wrdinf new))
	(setq stripold (wrdinf old))
	(setq slen (wrdlen stripold))
%       (for % assumes WrdLen New has been set
%
%             (from i 0 slen 1)
%             (do (setf (wrditm stripnew i) (wrditm stripold i))))
%       (return new)))
  (return (fast-copier (+ stripold 4) (+ stripnew 4) (+ slen 1) new))))  

(de copyhalfwordstofrom (new old)
  % Like CopyWArray in heap
  (prog (slen stripnew stripold)
	(setq stripnew (halfwordinf new))
	(setq stripold (halfwordinf old))
	(setq slen (halfwordlen stripold))
	(for % assumes WrdLen New has been set
 
	      (from i 0 slen 1)
	      (do (setf (halfworditm stripnew i) (halfworditm stripold i))))
	(return new)))

(de copywrds (s)
  % Allocate new WRDS array in heap
  (prog (s1)
	(setq s1 (gtwrds (wrdlen (wrdinf s))))
	(copywrdstofrom s1 (wrdinf s))
	(return (mkwrds s1))))

(de copyhalfwords (s)
  % Allocate new HALFWORDS array in heap
  (prog (s1)
	(setq s1 (gthalfwords (halfwordlen (halfwordinf s))))
	(copyhalfwordstofrom s1 (halfwordinf s))
	(return (mkhalfwords s1))))

% CopyPairToFrom is RplacW, found in EASY-NON-SL.RED
% CopyPair is: car S . cdr S;
% Usual Lisp definition of Copy only copies pairs, is found in EASY-NON-SL.RED

(de totalcopy (s)
  % Unique copy of entire structure blows up on circular structures
  (case (tag s)
    ((pair-tag)      (cons (totalcopy (car s)) (totalcopy (cdr s))))
    ((string-tag)    (copystring s))
    ((vector-tag)    (let* ((strips (vecinf s))
	                    (len (veclen strips))
	                    (ptr (mkvect len)))
	               (for (from i 0 len 1)
	                    (do
	                     (setf (vecitm (vecinf ptr) i)
	                       (totalcopy (vecitm (vecinf s) i)))))
	               ptr))
    ((words-tag)     (copywrds s))
    ((bytes-tag)     (copybytes s))
    ((halfwords-tag) (copyhalfwords s))
    ((fixnum-tag)    (mkfixn (inf (copywrds s))))
    ((floatnum-tag)  (copyfltn s))
    (nil             s)
    ))

(de copyfltn (s)
  % Allocate new Floatnum in heap
  % May have special alignment needs
  (prog (s1)
        (setq s1 (gtfltn (wrdlen (wrdinf s))))
        (copywrdstofrom s1 (wrdinf s))
        (return (mkfltn s1))))

(off fast-integers)
