%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxu/disassemble.sl
% Title:          Disassemble of instructions for HP-UX 9000/700
% Author:         Iain Beckingham ZIB,Winfried Neun ZIB.
% Created:        December 10 1991
% Status:         Experimental
%
% Copyright (c) 1991,1992 Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Program gets the name of the function first from the ext code. However, there
% are some special-cases which have no ext code but they have unique opcodes.
% The location of the operands/exts/suffixes are taken out of the format 
% definitions at the end of the program. The last 26 bits of the word are in
% rest-of-word and the operands/exts/suffixes are taken out of it.
% All the instructions are held in the vector *instrsvect*. Further vectors
% (e.g *svector*) are used to hold the suffixes. The immediates and tgt's are
% handled specially as they require manipulation(e.g disassemble_21 takes a 21
% bit immediate and returns(catx{20},x{9..19},x{5..6},x{0..4},x{7..8}))! The
% instructions are held in hp-instrs, and the suffixes in hp-suff.
% There are a couple or print routines to print things like x(s,b) d(s,b),t etc
% Also registers are aliased to their psl-friendly names.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(load compiler)
(on fast-integers)

(compiletime (flag '(dispatch-disassemble safe-cdr special-dispatch-disassemble
        cop1-dispatch-disassemble bcond-dispatch-disassemble
        normal-dispatch-disassemble print-instruction prinbl prinb prinbo
        prinbn prindig) 'iinternalfunction))

(fluid '(*svector* *ovector* *evector* *instrsvect* rest-of-word name hp-instrs hp-suff imm21 prim fvector svector ovector ops last-name instrs *suffixvect* *formatvector*
*operandsvect* klingelbeutel))

(fluid '(csic csic1 aic aic1 lic lic1 uic uic1 ilsc ilsc1 sbsc sbsc1 sdlsc))

(compiletime (setq *formatvector* (mkvect 80))
	     (setq *rest*	  (mkvect 512))
	     (setq *operandsvect* (mkvect 80))
	     (setq *suffixvect*   (mkvect 40))
)

(setq   *instrsvect* (mkvect 60)
	*formatvector* (mkvect 80)
        *operandsvect* (mkvect 80)
	*suffixvect* (mkvect 40))

(de read-info (x) %collect information from Compiler , store in instrsvect
   (prog (primary)
     (setq primary (get x 'opcode))
     (unless primary (return nil))
     (putv *instrsvect* primary
	 (cons (list x (get x 'extension) (get x 'format) (get x 'operands)
                       (get x 'suffixes)) 
               (getv *instrsvect* primary )))
     (return nil)))

(mapobl (function read-info))

(de disassemble (x)
 (prog (reg special-case stop i ll primary)
  (setq imm21 nil)
  (setq last-name nil)
  (setq special-case '(5 6 8 10 13 14 16 17 18 19 24 25 26 27 32
   33 34 35 36 38 40 41 42 43 48 49 50 51 56 57))	%fns  with no ext codes
  (setq outputbase* 16)
  (setq ll (linelength 120))
  (setq i x)
  (when (idp x) (when (funboundp x) (error 0 (list x "not compiled")))
                (setq i (wgetv symfnc (idinf x))))  
  (prin2l (list "**** Function :" x " at " (inf i))) (terpri) (terpri)
  (while (wgreaterp (setq primary (wshift (getmem i) -26)) 0)
    (setq rest-of-word (wshift (wshift (getmem i) 6) -6))
    (cond ((memq primary special-case)   		%if a special case
	      (setq name (caar (getv *instrsvect* primary)))
	  	(go loop1)))
    (get-format-type primary)
    (get-ext-code prim 0)
    (setq name (get-name 1))     %get the name if it has an ext code.
loop1
    (get-format-type primary)	 %get the operands here
    (get-operands prim 0)
	(collect-operands 1 name)
    (get-suffixes prim 0)	 %get the suffixes here
 	(collect-suffixes 1 name)
    (setq outputbase* 10)
    (prinbn i 8)				%print the address
    (tab 11)
    (prinbn (getmem i) 8)			%print the contents
    (tab 22)
    (prin1 name)				%print the name
    (cond ((null hp-suff) nil)
          (t (prin2 ",")			%if suffix then print it
	     (prin1 hp-suff)))
    (tab 31)
    (cond ((or (equal 'd (car klingelbeutel))	%if d(s,b),t or d(b),t
		(equal 'x (car klingelbeutel))	%if x(s,b),t etc
		(equal 'wd (car klingelbeutel)))	%if wd(s,b)
		(setq reg (register-alias (car klingelbeutel) (car hp-instrs)))
	   (prin2l reg)
	   (print-format-2 (cdr klingelbeutel) (cdr hp-instrs) name))
    	   (t (print-format-1 klingelbeutel hp-instrs i)))
						%print the operands and comments
    (setq last-name name)			%to see if ldo follows ldil etc
    (setq i (wplus2 i 4)))
  (linelength ll)
))

% Routine to print out the operands

(de print-format-1 (actual-ops given-ops address)
	(prog (actual given a b)
	   (setq actual actual-ops)
	   (setq given given-ops)
		(cond ((equal 'tgt (car actual))
			(setq b (plus2 (cadar given) 8))
			(prinbn (wplus2 (true-val b) address) 8) 
			(terpri)
			(return nil)))
		(cond ((equal 'd (car actual)) 
			(setq a (register-alias (car actual) (car given)))
			(prin2l a)
			(prin2 "(")
			(setq a (register-alias (cadr actual) (cadr given)))
			(prin2l a)
			(prin2 ",")
			(setq a (register-alias (caddr actual) (caddr given)))
			(prin2l a) (prin2 ")")
			(return (comments name hp-instrs))))
%We don't wish to print the names of the constants like im21,p,len etc
		(cond ((or (equal 'p (car actual)) 
			   (equal 'im5 (car actual))
			   (equal 'target (car actual))
			   (equal 's (car actual))
			   (equal 'sr (car actual))
			   (equal 'im11 (car actual))
			   (equal 'im14 (car actual))
			   (equal 'im21 (car actual))
			   (equal 'len (car actual)))
			(prin1 (cadar  given))
			(prin2 ",")
			(return
			   (print-format-1 (cdr actual) (cdr given) address))))
		(setq a (register-alias (car actual) (car given)))
		(cond ((null (cdr actual))
		        (prin2l a)
			(return (comments name hp-instrs)))
			(t (prin2l a)
			   (prin2 ",")))
	(print-format-1 (cdr actual) (cdr given) address)))


% Routine to print out in form d(s,b),t or x(b),t etc

(de print-format-2 (actual-ops given-ops name)
	(prog (actual given a)
	     	 (setq actual actual-ops)
		 (setq given given-ops)
	         (cond ((equal 'bv name)			%if x(b) case
			(setq a (register-alias (car actual) (car given)))
			(prin2 "(")
			(prin2l a)
			(prin2 ")") (terpri)
			(return nil)))
		 (cond ((not (null (cddr given)))	%if (s,b),t case
			(prin2 "(")
			(setq a (register-alias (car actual) (car given)))
			(prin2l a)
			(prin2 ",")
			(setq a (register-alias (cadr actual) (cadr given)))
			(prin2l a)
			(prin2 "),")
			(setq a (register-alias (caddr actual) (caddr given)))
			(prin2l a)
			(return (comments name hp-instrs))))
		 (cond ((memq name '(ldwas ldo ldwax))
	   	        (prin2 "(")			%if (b),t case
		        (setq a (register-alias (car actual) (car given)))
			(prin2l a)
			(prin2 "),")
			(setq a (register-alias (cadr actual) (cadr given)))
			(prin2l a)
			(return (comments name hp-instrs))))
%otherwise we have the wd(s,b) and x(s,b) cases
		 (prin2 "(")
		 (setq a (register-alias (car actual) (car given)))
		 (prin2l a)
		 (prin2 ",")
		 (setq a (register-alias (cadr actual) (cadr given)))
		 (prin2l a)
		 (prin2 ")") (terpri) 
		 (return nil)))

%Some aliasing is done here to give registers their psl friendly name.

(de register-alias (actual given)
	(prog (register a actual-ops given-ops)
		(setq actual-ops actual)	%get first actual op
		(setq given-ops given)		%get first given op
		(setq register '((0 (reg zero)) (2 (reg returnaddr))
				 (3 (reg bndstkptr)) (4 (reg bndstklowerbound))
				 (5 (reg bndstkupperbound)) (6 (reg ff1))
				 (7 (reg ff2)) (8 (reg ff3)) (9 (reg ff4))
				 (10 (reg heaplast)) (11 (reg heaptrapbound))
				 (12 (reg symfncstart)) (13 (reg symvalstart))
				 (15 (reg t)) (16 (reg t6)) (17 (reg t5))
				 (18 (reg t4)) (19 (reg t3)) (20 (reg t2))
				 (21 (reg t1)) (22 (reg argfive))
				 (23 (reg argfour)) (24 (reg argthree))
				 (25 (reg argtwo)) (26 (reg argone))
				 (28 (reg nil)) (30 (reg sp))))
		(cond ((and (memq actual-ops '(r r1 r2 t b x)) %if a register
			    (assoc (cadr given-ops) register)) %if op is above
			(setq a (cadr (assoc (cadr given-ops) register)))
			(return a)))
%We don't want to print the d,wd,sr,s or r1
		(cond ((or (equal 'd actual-ops) 
			   (equal 'sr actual-ops)
 			   (equal 's actual-ops)
			   (equal 'r1 actual-ops)
			   (equal 'wd actual-ops))
			(setq a (list (cadr given-ops))))	
			(t (setq a (list actual-ops (cadr given-ops)))))
				(return a)))

% Some print routines...

(de prinbl (l)                  % binary (octal) printing of a list
     (mapc l (function prinb)))

(de prinb (it)                  % binary (octal) printing of an item
     (cond ((numberp it)(prinbo it))
           ((eq it 't1) (tab 35))
           ((eq it 't2) (tab 60))
           (t       (prin2 it))))

(de prinbo (it)
      (cond ((wlessp it 0) (prin2 "-") (prinbo (wminus it)))
            ((wgeq it 10)  (prin2 "x'") (prinbn it 1))
    (t            (prinbn it 1))))

(de prinbn (it n)                % print an octal number
     (cond ((and (weq it 0) (wleq n 0)) nil)
            (t (progn
                  (prinbn (wshift it -4) (wplus2 n -1))
                  (prindig (wand it 15))
)    )     )  )

(de prindig (dig)        % print a numeric digit
     (if (wlessp dig 10) (writeChar  (wplus2 dig 48))
           (writechar (wplus2 65 (wdifference dig 10)))))

%extensions in *formatvector*,suffixes in *suffixvect* and ops in *operandsvect*

(dm define-format (u)
    (prog (sum number )
      (setq sum 0)
      (setq u (cdr u))
      (setq number (pop u))
      (while u 
       (cond ((and (idp (car u)) (memq (car u) '(ext3 ext4 e ext8 ext7)))
            (putv *formatvector* number (cons sum (cadr u))))
	     ((member (car u) '(c n u a m f))
		(putv *suffixvect* number
				(append (getv *suffixvect* number)
					(list (list (car u) sum (cadr u))))))	 
	        ((member (car u) '(im5 im11 im14 im21 w1 w2 w x r im26 im13 b r1 sr clen cp p r2 t r2!/p s t!/b r!/im5 r1!/im5 t!/r r!/x!/im5))
			(putv *operandsvect* number 
				(append (getv *operandsvect* number)
					(list (list (car u) sum (cadr u)))))))
         (setq sum (plus2 sum (cadr u)))
	 (setq u (cddr u)))
      (return nil)))


%Program to get the length and offset of the extension field.
%This, along with the fn name and it's extension, are stored in *evector*.
%i.e ((fn ext) offset.length)

(de get-ext-code (x sum)
	(prog (fn position)
	   (cond ((null x) (return nil))   %while there is still a list to work on
	      (t (setq fn (car x))	   %get the first fn in the opcode list
		   (setq position (getv fvector (caddr fn))) %get format code
	 	   (putv *evector* sum (cons (list (caar x) (cadar x)) position))
		   (setq sum (plus2 sum 1))
		   (return (get-ext-code (cdr x) sum))))))


%Program to get opcode no.(primary) out of the vector *instrsvect*.

(de get-format-type (x)
   (prog (primary) 			   		%local variables
   	(setq prim (getv *instrsvect* x)))   		%get op-code
   	(setq *evector* (mkvect (- (length prim 1) 1))) %ext. vector
   	(setq *ovector* (mkvect (- (length prim 1) 1))) %operands vector
	(setq *svector* (mkvect (- (length prim 1) 1))) %suffix vector
   (return nil))


%Routine to get the actual name of the function that we are looking for.
%It checks if the ext corresponds to a function that is held in *evector*.
%However, there are a number of op-codes that have functions in them that
%do not contain an ext. code, and therefore must be treated differently.

(de get-name (x)
	(prog (function name ext-field xtension offset len)
	loop
	   (cond ((> x (+ (size *evector*) 1)) (return nil))) %otherwise  
							      %return nil
		(setq function (getv *evector* (- x 1)))
		(setq name (caar function))
		(setq xtension (cadar function))
		(setq offset (cadr function))
		(setq len (cddr function))
		(setq ext-field (offset-length offset len rest-of-word))
			(cond ((equal ext-field xtension) (return name))
		  		(t (setq x (+ x 1))
		   			(go loop)))))

%routine that gets the correct number of bits from the word
%The offset, length, and the last 26 bits of the word are passed.
(de offset-length (x y z)
	(prog (word offset length shift result)
		(setq word z)
		(setq offset x)
		(setq length y)
		(setq shift (- 32 (+ offset length)))
	(setq result (land (lshift word (- shift)) (- (expt 2 length) 1)))
		(return result)))
		
%Puts operand types in *ovector*, (for operands).
(de get-operands (x sum)
	(prog (fn position)
		(cond ((null x) nil)
			(t (setq fn (car x))
			   (setq position (getv ovector (caddar x)))
			   (putv *ovector* sum (append (list (caar x)
							   (fourth (car x)))
							   position))
			   (setq sum (plus2 sum 1))
			   (get-operands (cdr x) sum)))))

%There now follows a variety of programs used to get operands that
%use w1 w2 w immediates and other wondrous things.

% Removes the rightmost bit of x and extends the field to the left with that
% bit to form a 32-bit quantity
(de dis_low_sign_ext_11 (x)
	(prog (a b c)
		(setq a (wand 1 x))
		(setq b (wshift x -1))
		(setq c (wor (wshift a 10) b))
	(return (signedfield c 21 11))))

(de dis_low_sign_ext_14 (x)
	(prog (a b c)
		(setq a (wand 1 x))
		(setq b (wshift x -1))
		(setq c (wor (wshift a 13) b))
	(return (signedfield c 18 14))))

(de dis_low_sign_ext_5 (x)
	(prog (a b c)
		(setq a (wand 1 x))
		(setq b (wshift x -1))
		(setq c (wor (wshift a 4) b))
	(return (signedfield c 27 5))))

% returns(cat(y,x{10},x{0..9}))
(de disassemble_12 (x y)
	(prog (a b c d e)
		(setq a y)
		(setq b (wand 1 x))
		(setq c (wand (wshift x -1) (- (expt 2 10) 1)))
		(setq d (wor (wshift b 10) c))
		(setq e (wor (wshift a 11) d))
	(return e))) 

% returns cat(z,x,y{10},y{0..9})
(de disassemble_17 (x y z)
	(prog (a b c d e f g)
		(setq a z)
		(setq b x)
		(setq c (wand 1 y))
		(setq d (wand (wshift y -1) (- (expt 2 10) 1)))
		(setq e (wor (wshift c 10) d))
		(setq f (wor (wshift b 11) e))
		(setq g (wor (wshift a 16) f))
	(return g)))

% returns(cat(x{20},x{9..19},x{5..6},x{0..4},x{7..8}))
(de disassemble_21 (x)
	(prog (a b c d e f g h i)
		(setq a (wand 1 x))
		(setq b (wand (wshift x -1) (- (expt 2 11) 1)))
		(setq c (wand (wshift x -14) (- (expt 2 2) 1)))
		(setq d (wand (wshift x -16) (- (expt 2 5) 1)))
		(setq e (wand (wshift x -12) (- (expt 2 2) 1)))
		(setq f (wor (wshift d 2) e))
		(setq g (wor (wshift c 7) f))
		(setq h (wor (wshift b 9) g))
		(setq i (wor (wshift a 20) h))
	(return i )))

% returns(cat(x{2},x{0..1}))
(de disassemble_3 (x)
	(prog (a b c)
		(setq a (wand 1 x))
		(setq b (wand (wshift x -1) 3))
		(setq c (wor (wshift a 2) b))
	(return c)))

% Extends x to the left with sign bits to form a 32-bit quantity,taking the 
% leftmost bit for the field as the sign bit.(size 12 in this case)
(de sign_ext_12 (x)
	(prog (a b c)
		(setq a (wand (wshift x 1) (- (expt 2 12) 1)))
		(setq b (wand (wshift a -1) (- (expt 2 11) 1)))
		(setq c (ashift (wshift b 21) -21))
(return c)))

% Comment as above but with size 17
(de sign_ext_17 (x)
	(prog (a b c)
		(setq a (wand (wshift x 1) (- (expt 2 17) 1)))
		(setq b (wand (wshift a -1) (- (expt 2 16) 1)))
		(setq c (ashift (wshift b 16) -16))
(return c)))

%end of section

%The following two routines get the operands and store them in hp-instrs

%Main routine that collects the operands on the hp machine.
(de collect-operands (x name)
	(prog (function results fname offset len)
	   (cond ((> x (+ (size *ovector*) 1)) (return nil)))
	(setq function (getv *ovector* (- x 1)))
	(setq fname name)
	(cond ((equal fname (car function)) 	%found the instrs we need
	    (progn 
		(go loop1)
		(return results)))
		 (t (setq x (+ x 1))
		   (return (collect-operands x name))))
	loop1
		(setq ops (cadr function))
		(setq klingelbeutel (squash ops))	%global operands
		(return (get-values ops (cddr function) nil))))

%routine that get's the actual bits...and also handles the aliasing
(de get-values (a b c)
	(prog (offset len imm value)
		(setq imm nil)
		(cond ((null a) (setq hp-instrs (reverse c)) (return hp-instrs))
		   (t (setq a (squash a))
		      (setq instrs (car a))
%do something here if instrs is an im5 or an i
	(cond ((or (equal 'im5 instrs) (equal 'i instrs))
		(setq imm instrs)))
		      (setq instrs (aliasing instrs b))
			(cond ((equal 'im26 instrs)
				(setq c (special-operands 'im26 b c))
				(return (get-values (cdr a) b c))))
			(cond ((not (null imm))	%an im5 but aliased to an r/im5
				 (setq c (special-operands instrs b c))
				 (return (get-values (cdr a) b c))))
			(cond 
			  ((member instrs '(target tgt wd sr im5 cp clen im11 im14 im21)) 
				(setq c (special-operands instrs b c))
				(return (get-values (cdr a) b c))))
		      (setq offset (cadr (assoc instrs b)))
		      (setq len (caddr (assoc instrs b)))
		      (setq value (offset-length offset len rest-of-word))
		      (setq c (cons (list instrs value) c))
		      (get-values (cdr a) b c)))))

%routine that will alias t's and r's to t!/r e.t.c
(de aliasing (l m)
		(prog (alias)
		(if (null (assoc l m))
		    (cond ((equal l 'b)   	%if instrs "b" alias it to t/b
		  		(setq l 't!/b))
			  ((equal l 'x)		%if "x" alias it to r/x/im5
				(setq l 'r!/x!/im5))
			  ((equal l 'i1)	%alias i1 to im13 
				(setq l 'im13))
 			  ((equal l 'i2)        %alias i2 to im5
                                (setq l 'im5))
			  ((equal l 'len)	%alias len to clen
				(setq l 'clen))
			  ((equal l 'r1)	%alias r1 to r1/im5
				(setq l 'r1!/im5))
			  ((equal l 'r2)	%alias r2 to r2/p
				(setq l 'r2!/p))
%alias im5 to r/im5 or r1/im5
			  ((equal l 'im5)	
			  (cond ((null (assoc 'r!/im5 m))
				 (setq l 'r1!/im5))
				 (t (setq l 'r!/im5))))

%alias p to cp or r2/p
			  ((equal l 'p)
			  (cond ((null (assoc 'cp m))
				(setq l 'r2!/p))
				(t (setq l 'cp))))

%alias i to r/x/im5 or im26 
			  ((equal l 'i)
			  (cond ((null (assoc 'im26 m))
				(setq l 'r!/x!/im5))
				(t (setq l 'im26))))
%alias d to im14 or im5
			  ((equal l 'd)		
			  (cond ((null (assoc 'im14 m))
				(setq l 'im5))
				(t (setq l 'im14))))
%alias r to t/r or r/im5
			  ((equal l 'r)		
			  (cond ((null (assoc 'r!/im5 m))
				(setq l 't!/r))
				(t (setq l 'r!/im5))))

		          ((equal l 't)		%alias t to t/r or t/b
			  (cond ((null (assoc 't!/r m))
				(setq l 't!/b ))
				(t (setq l 't!/r))))))
		(return l)))


%routine to take out the parentheses in a list
(de squash (s)
	(cond ((null s) nil)
	      ((atom s) (list s))
	      (t (append (squash (car s)) (squash (cdr s))))))

%end of operand main routine. The following routine handles the immediates etc.

(de special-operands (instrs b iain)
	(prog (offset len hugo temp x y z)
		    (setq temp b)	%let temp be fn parameters
		(cond ((or (equal 'wd instrs) (equal 'target instrs))	%for bl and gate instrs,and also for ble and be
		    (setq offset (cadr (assoc 'w1 temp)))
		    (setq len (caddr (assoc 'w1 temp)))
		    (setq x (offset-length offset len rest-of-word))
		    (setq offset (cadr (assoc 'w2 temp)))
		    (setq len (caddr (assoc 'w2 temp)))
		    (setq y (offset-length offset len rest-of-word))
		    (setq z (offset-length 31 1 rest-of-word)) %as w is always
							       %length 1
		(setq hugo (lshift (sign_ext_17 (disassemble_17 x y z)) 2))	%pass to disasseble routine
		(setq iain (cons (list 'target hugo) iain))
	(return iain)))

%now for the space registers

		(cond ((equal 'sr instrs)
			(setq offset (cadr (assoc 'sr temp)))
			(setq len (caddr (assoc 'sr temp)))
			(setq x (offset-length offset len rest-of-word))
			(setq hugo (disassemble_3 x))
			(setq iain (cons (list 'sr hugo) iain))
		(return iain)))

%now for the immediate 21 (ldil and addil instrs)

		(cond ((equal 'im21 instrs)
			(setq offset (cadr (assoc 'im21 temp)))
			(setq len (caddr (assoc 'im21 temp)))
			(setq x (offset-length offset len rest-of-word))
			(setq hugo (lshift (disassemble_21 x) 11))
			(setq iain (cons (list 'im21 hugo) iain))
		(return iain)))

%now for the immediate 14 (format-code 1)

                (cond ((equal 'im14 instrs)
                        (setq offset (cadr (assoc 'im14 temp)))
                        (setq len (caddr (assoc 'im14 temp)))
                        (setq x (offset-length offset len rest-of-word))
                        (setq hugo (dis_low_sign_ext_14 x))
                        (setq iain (cons (list 'im14 hugo) iain))
                (return iain)))
		
%now for the immediate 11 (format-code 6)

                (cond ((equal 'im11 instrs)
                        (setq offset (cadr (assoc 'im11 temp)))
                        (setq len (caddr (assoc 'im11 temp)))
                        (setq x (offset-length offset len rest-of-word))
                        (setq hugo (dis_low_sign_ext_11 x))
                        (setq iain (cons (list 'im11 hugo) iain))
                (return iain)))

%now for the tgt case (for comb,branch ,and move functions)
%special case for the BB function

		(cond ((equal 'tgt instrs)
			(setq offset (cadr (assoc 'w1 temp)))
			(setq len (caddr (assoc 'w1 temp)))
			(setq x (offset-length offset len rest-of-word))
			(setq y (offset-length 31 1 rest-of-word))
			(setq hugo (lshift (sign_ext_12 (disassemble_12 x y)) 2))
			(setq iain (cons (list 'tgt hugo) iain))
		(return iain)))

%now for the im26 case which is put in for ease here
%Only one im26 thus always length 26 offset 6.

		(cond ((equal 'im26 instrs)
			(setq x (offset-length 6 26 rest-of-word))
			(setq iain (cons (list 'im26 x) iain))
		(return iain)))

%now for the clen case (for format codes 8 and 9)

		(cond ((equal 'clen instrs)
			(setq offset (cadr (assoc 'clen temp)))
			(setq len (caddr (assoc 'clen temp)))
			(setq x (offset-length offset len rest-of-word))
			(setq hugo (plus2 (minus x) 32))
			(setq iain (cons (list 'len hugo) iain))
		(return iain)))

%now for the cp case (for format codes 8 and 9)

		(cond ((equal 'cp instrs)
			(setq offset (cadr (assoc 'cp temp)))
			(setq len (caddr (assoc 'cp temp)))
			(setq x (offset-length offset len rest-of-word))
			(setq hugo (plus2 (minus x) 31))
			(setq iain (cons (list 'p hugo) iain))
		(return iain)))

%now for the im5 case (for format codes 3,4,9,11,15,16 and possibly 23)

                (cond ((equal 'im5 instrs)
                        (setq offset (cadr (assoc 'im5 temp)))
                        (setq len (caddr (assoc 'im5 temp)))
                        (setq x (offset-length offset len rest-of-word))
                        (setq hugo (dis_low_sign_ext_5 x))
                        (setq iain (cons (list 'im5 hugo) iain))
                (return iain)))

%now for the r/im5 and the r1/im5 cases
%It's easier to do it this way as both are always in the same place.
%i.e offset 11 and length 5.

                (cond ((or (equal 'r!/x!/im5 instrs)
			   (equal 'r!/im5 instrs) 
			   (equal 'r1!/im5 instrs))
                        (setq x (offset-length 11 5 rest-of-word))
                        (setq hugo (dis_low_sign_ext_5 x))
                        (setq iain (cons (list 'im5 hugo) iain))
                (return iain)))))

% Routine that handles the comments

% Routine to put a few comments in for things like ldil,ldo etc

(de comments (a b)
	(prog (x y z)
		(cond ((equal 'ldil a)
		       (setq imm21 (cadar b))
		       (terpri)
		       (return nil)))
%ldil followed by an ldo instr
	        (cond ((and (not (null imm21)) (equal 'ldil last-name) (equal 'ldo a))
		       (setq x (plus2 imm21 (cadar b)))
		       (setq imm21 nil)	%reset imm21
		       (tab 67)
		       (setq x (true-val x))
		       (cond ((eq (tag x) 8) (terpri) (return nil))
			     (t (print (true-val x))))
		       (return nil)))
%ldil followed by an ldwm or ldw  instr
		(cond ((and (not (null imm21)) (equal 'ldil last-name)
				(or (equal 'ldwm a) (equal 'ldw a)))
		       (setq x (plus2 imm21 (cadar b)))
		       (setq y (quotient (wdifference (fixval x) symval) 4))
		       (setq imm21 nil) 	%reset imm21
		       (tab 67)
		       (print (int2id y))
		       (return nil)))
%ldo followed by a ldwx etc
		(cond ((and (equal 'ldo a) (wgreaterp (cadar b) 0))
		       (setq x (cadar b))
		       (tab 67)
		       (print (int2id x))
		       (return nil)))
		(terpri)
		(return nil))
	(return nil))

%Get suffixes from svector and put them in another vector along with
%the name of the function

(de get-suffixes (x sum)
	(prog (fn position)
		(cond ((null x) (return nil))
		       (t (setq fn (car x))
		       (setq position (getv svector (caddar x)))
		       (putv *svector* sum
					(append (list (caar x) 
						(car (last (car x)))) position))
		       (setq sum (plus2 sum 1))
		       (get-suffixes (cdr x) sum)))))

% Routine similar to 'collect-operands' but to collect the suffixes 

(de collect-suffixes (x name)
	(prog (function actual-suffix results fname offset len)
		 (cond ((> x (+ (size *svector*) 1)) (return nil)))
		 (setq function (getv *svector* (- x 1)))
		 (setq fname name)
		 (cond ((equal fname (car function))
				(go loop1))
			(t (setq x (plus2 x 1))
			   (return (collect-suffixes x name))))

%We have found the function that we are looking for
	loop1
		(setq actual-suffix (cadr function))
		(return (suffix-values actual-suffix (cddr function) nil))))

%Routine to get the values and meanings of the suffixes
%Assoc lists are used as there are 9 condition tables
% cd5 and cp4 are handled separately as they only have the one specifier

% csic-Compare/Subtract Instruction Conditions,aic-Add Instruction Conditions
% lic-Logical Instruction Conditions,sdlsc-Short Displaced Load/Store Completers
% uic-Unit Instruction Conditions,ilsc- Indexed Load/Store Completers
% sbsc-Short Bytes Short Completers

(de suffix-values (x y z)
	(prog (offset len value value1 suffix sedic)
		(cond ((null x) (setq hp-suff nil) (return hp-suff)))
		(setq suffix x)
		(cond ((equal '(cp4) suffix)
		(setq offset (cadr (first y)))
		(setq len (caddr (first y)))
	        (setq value
			(offset-length offset len rest-of-word))
		(cond ((zerop value) (setq hp-suff nil)
		       (return hp-suff))
		       (t (setq hp-suff 'm)
		          (return hp-suff))))
	              ((equal '(n) suffix)
		      (setq offset (cadr (first y)))
                             (setq len (caddr (first y)))
                             (setq value
                                (offset-length offset len rest-of-word))
		      (cond ((zerop value) (setq hp-suff nil)
			    (return hp-suff))
			    (t (setq hp-suff 1)
			       (return hp-suff))))

		      ((equal '(cd5) suffix)
			     (setq offset (cadr (first y)))
                             (setq len (caddr (first y)))
                             (setq value
                                (offset-length offset len rest-of-word))
%Shift/Extraction/Deposit Instruction Conditions
			     (setq sedic '((0 nil) (1 =) (2 <) (3 OD)
				          (4 TR) (5 <>) (6 >=) (7 EV)))
					(setq hp-suff (cadr 
						       (assoc value sedic)))
					(return hp-suff)))
%otherwise cd1,2,3,4 and cp1,2,3 cases
			            
			      (setq offset (cadr (first y)))
			      (setq len (caddr (first y)))
			      (setq value 
				(offset-length offset len rest-of-word))
			      (setq offset (cadar (last y)))
			      (setq len (caddar (last y)))
			      (setq value1
				(offset-length offset len rest-of-word))
			      (cond ((and (or (equal '(cd1) suffix)
					      (equal '(cd1 n) suffix))
				     (zerop value1))
			             (setq hp-suff (cadr
						  (assoc value csic)))
				     (return hp-suff)))
			      (cond ((and (or (equal '(cd1) suffix)
                                              (equal '(cd1 n) suffix))
                                     (zerop value))
				      (setq hp-suff (cadr
						  (assoc value csic1)))
				      (return hp-suff)))
			      (cond ((and (equal '(cd2) suffix)
				     (zerop value1))
				     (setq hp-suff (cadr
						  (assoc value aic)))
				     (return hp-suff)))
			      (cond ((and (equal '(cd2) suffix)
                                     (zerop value))
				     (setq hp-suff (cadr
						    (assoc value aic1)))
				      (return hp-suff)))
			      (cond ((and (equal '(cd3) suffix)
                                     (zerop value1))
                                     (setq hp-suff (cadr
                                                    (assoc value lic)))
                                     (return hp-suff)))
			      (cond ((and (equal '(cd3) suffix)
                                     (zerop value))
                                     (setq hp-suff (cadr
                                                       (assoc value lic1)))
                                     (return hp-suff)))
			      (cond ((and (equal '(cd4) suffix)
                                     (zerop value1))
                                     (setq hp-suff (cadr
                                                    (assoc value uic)))
                                     (return hp-suff)))
			      (cond ((and (equal '(cd4) suffix)
                                     (zerop value))
                                     (setq hp-suff (cadr
                                                       (assoc value uic1)))
                                        (return hp-suff)))
			      (cond ((and (equal '(cp1) suffix)
				     (zerop value1))
				     (setq hp-suff (cadr
						 (assoc value ilsc)))
				     (return hp-suff)))
			      (cond ((and (equal '(cp1) suffix)
                                     (zerop value))
				     (setq hp-suff (cadr
						   (assoc value ilsc1)))
				     (return hp-suff)))
			      (cond ((and (equal '(cp3) suffix)
                                     (zerop value1))
                                     (setq hp-suff (cadr
                                                    (assoc value sbsc)))
                                     (return hp-suff)))
			      (cond ((and (equal '(cp3) suffix)
                                     (zerop value))
                                     (setq hp-suff (cadr
                                                       (assoc value sbsc1)))
                                        (return hp-suff)))
			      (cond ((and (equal '(cp2) suffix)
				     (zerop value1))
				     (setq hp-suff nil)
				     (return hp-suff)))
			      (cond ((and (equal '(cp2) suffix)
                                     (zerop value))
				     (setq hp-suff (cadr
						  (assoc value sdlsc)))
				     (return hp-suff)))))
% Format Definitions
% Format 1: Load and Stores

(define-format 1 op 6 b 5 t!/r 5 s 2 im14 14)

% Format 2: Indexed Loads

(define-format 2 op 6 b 5 x 5 s 2 u 1 0 1 0 2 ext4 4 m 1 t 5)

% Format 3: Short displacement loads

(define-format 3 op 6 b 5 im5 5 s 2 a 1 1 1 0 2 ext4 4 m 1 t 5)

% Format 4: Short displacement stores, Store bytes Short

(define-format 4 op 6 b 5 r 5 s 2 a 1 1 1 0 2 ext4 4 m 1 im5 5)

% Format 5: Long immediates

(define-format 5 op 6 t!/r 5 im21 21)

% Format 6: Arithmetic/Logical

(define-format 6 op 6 r2 5 r1 5 c 3 f 1 ext7 7 t 5)

% Format 7: Arithmetic immediate

(define-format 7 op 6 r 5 t 5 c 3 f 1 e 1 im11 11)

% Format 8: Extract

(define-format 8 op 6 r 5 t 5 c 3 ext3 3 p 5 clen 5)

% Format 9: Deposit

(define-format 9 op 6 t 5 r!/im5 5 c 3 ext3 3 cp 5 clen 5)

% Format 10: Shift

(define-format 10 op 6 r2 5 r1 5 c 3 ext3 3 cp 5 t 5)

% Format 11: Conditional Branch

(define-format 11 op 6 r2!/p 5 r1!/im5 5 c 3 w1 11 n 1 w 1)

% Format 12: Branch External Branch and Link External

(define-format 12 op 6 b 5 w1 5 sr 3 w2 11 n 1 w 1)

% Format 13: Branch and Link, Gateway

(define-format 13 op 6 t 5 w1 5 ext3 3 w2 11 n 1 w 1)

% Format 14: Branch and Link Register, Branch vectored

(define-format 14 op 6 t!/b 5 x 5 ext3 3 0 11 n 1 0 1)

% Format 15: Data memory Management, Probe

(define-format 15 op 6 b 5 r!/x!/im5 5 s 2 ext8 8 m 1 t 5)

% Format 16: Instruction Memory Management

(define-format 16 op 6 b 5 r!/x!/im5 5 sr 3 ext7 7 m 1 0 5)

% Format 17: Break

(define-format 17 op 6 im13 13 ext8 8 im5 5)

% Format 18: Diagnose

(define-format 18 op 6 im26 26)

% Format 19: Move to/from Space Register

(define-format 19 op 6 0 5 r!/0 5 sr 3 ext8 8 0!/t 5)

% Format 20: Load Space ID

(define-format 20 op 6 b 5 0 5 s 2 0 1 ext8 8 t 5)

% Format 21: Move to Control register

(define-format 21 op 6 t 5 r 5 0 3 ext8 8 0 5)

% Format 22: Move from Control Register

(define-format 22 op 6 r 5 0 5 0 3 ext8 8 t 5)

% Format 23: System Control

(define-format 23 op 6 0!/b 5 0!/r!/im5 5 0 3 ext8 8 0!/t 5)

% Format 24: Special Operation Zero

(define-format 24 op 6 sop1 15 0 2 sfu 3 n 1 sop2 5)

% Format 25: Special Operation One

(define-format 25 op 6 sop1 15 0 2 sfu 3 n 1 t 5)

% Format 26: Special Operation Two

(define-format 26 op 6 r 5 sop1 10 2 2 sfu 3 n 1 sop2 5)

% Format 27: Special Operation Three

(define-format 27 op 6 r2 5 r1 5 sop1 5 3 2 sfu 3 n 1 sop2 5)

% Format 28: Coprocessor Operation

(define-format 28 op 6 sop1 17 uid 3 n 1 sop2 5)

% Format 29: Coprocessor Indexed Loads

(define-format 29 op 6 b 5 x 5 s 2 u 1 0 1 0 2 0 1 uid 3 m 1 t 5)

% Format 30: Coprocessor Indexed Stores

(define-format 30 op 6 b 5 x 5 s 2 u 1 0 1 0 2 1 1 uid 3 m 1 r 5)

% Format 31: Coprocessor Short Displacement Loads

(define-format 31 op 6 b 5 im5 5 s 2 a 1 1 1 0 2 0 1 uid 3 m 1 t 5)

% Format 32: Coprocessor Short Displacement Stores

(define-format 32 op 6 b 5 im5 5 s 2 a 1 1 1 0 2 1 1 uid 3 m 1 r 5)

% Format 33: Floating Point Operation Zero

(define-format 33 op 6 r 5 0 5 sop 3 fmt 2 0 2 0 3 0 1 t 5)

% Format 34: Floating Point Operation One

(define-format 34 op 6 r 5 0 4 sop 2 df 2 sf 2 1 2 0 3 0 1 t 5)

% Format 35: Floating Point Operation Two

(define-format 35 op 6 r1 5 r2 5 sop 3 fmt 2 2 2 0 3 n 1 c 5)

% Format 36: Floating Point Operation Three

(define-format 36 op 6 r1 5 r2 5 sop 3 fmt 2 3 2 0 3 0 1 t 5)

% Format 37: "Linked triad" operations

(define-format 37 op 6 rm1 5 rm2 5 ta 5 ra 5 f 1 tm 5)

% Something happens funny at compile time and hence these macros
(dm xxx (u)
 `(setq fvector ',*formatvector*))

(xxx)

(dm yyy (u)
`(setq svector ',*suffixvect*))

(yyy)

(dm zzz (u)
`(setq ovector ',*operandsvect*))

(zzz)

(dm true-val (x)
	(cond ((fixnp x) (fixval x))
		(t x)))


%These assoc lists relate to the suffixes

(setq csic '((0 nil) (1 =) (2 <) (3 <=) (4 <<) (5 <<=) (6 sv) (7 od)))
(setq csic1 '((0 tr) (1 <>) (2 >=) (3 >) (4 >>=) (5 >>) (6 nsv) (7 ev)))
(setq aic '((0 nil) (1 =) (2 <) (3 <=) (4 nuv) (5 znv) (6 sv) (7 od)))
(setq aic1 '((0 tr) (1 <>) (2 >=) (3 >) (4 uv) (5 vnz) (6 nsv) (7 ev)))
(setq lic '((0 nil) (1 =) (2 <) (3 <=) (7 od)))
(setq lic1 '((0 tr) (1 <>) (2 >=) (3 >) (7 ev)))
(setq uic '((0 nil) (2 sbz) (3 shz) (4 sdc) (6 sbc) (7 shc)))
(setq uic1 '((0 tr) (2 nbz) (3 nhz) (4 ndc) (6 nbc) (7 nhc)))
(setq ilsc '((0) (1 s)))
(setq ilsc1 '((0 m) (1 sm)))
(setq sdlsc '((0 ma) (1 mb)))
(setq sbsc '((0 b) (1 e)))
(setq sbsc1 '((0 bm) (1 em)))
