%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
% File:         PXU:MULS.SL
% Description:  Support for BIGNUM package with double INUM operations
%               Lap source for SPARC processor
% Author:       H. Melenk, W. Neun
% Created:      25 January 1989
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
%
% (c) Copyright 1992, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% i860 version by W. Neun, ZIB Berlin
%
% 14 Aug 1992 (Winfried Neun)
% Added SpuerSparc support and removed surplus wdivide from wquotientdouble
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the 68020 processor.
% These are the open codes which embed the extended arithmetic
% operations such that they are accessible from compiled LISP top level.
% They map the 2*30 bit operations needed by BIG to the 2*32 bit operations
% of the processor by shifting etc.

(fluid '(*second-value*))


(fluid '(*second-value*))

(compiletime (progn

 (defcmacro *wrshift3)

 (de *wrshift3 (a1 a2 a3)
   (prog (resultingcode*)
     (setq a1 (resolveoperand '(reg t1) a1))
     (setq a2 (resolveoperand '(reg t1) a2))
     (setq a3 (resolveoperand '(reg t1) a3))
     (return `((*move ,a2 (reg t1))(shr ,a3 (reg t1) ,a1)))))
))

(de WTIMESDOUBLE (x y)
     (prog (resu second carry)
	(setq resu (wtimesdou x y))
	(setq second *second-value*)
	(setq carry (wshift resu -30))
	(setq resu (wshift (wshift resu 2 ) -2))
	(setq second (wor (wshift second 2) carry))
	(setq *second-value* second)
	(return resu)))
	 
(lap '((*entry wquotientdouble expr 3)
       (*alloc 6)
       (*move (reg 1) (reg t1)) % arg1
       (*move (reg 2) (reg t2)) % arg2
       (*move (reg 3) (reg t3)) % arg3 
%      (*move (reg 3) (reg 2))
%      (*link wdivide expr 2)
       (*move (quote 0) (reg t4))  % quot
       (*move (reg 1) (reg t6)) % rem
       (*wshift (reg t2) 2)   
                                   30 times!
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab1) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab1
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab2) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab2
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab3) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab3
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab4) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab4
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab5) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab5
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab6) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab6
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab7) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab7
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab8) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab8
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab9) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab9
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab10) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab10
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab11) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab11
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab12) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab12
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab13) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab13
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab14) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab14
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab15) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab15
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab16) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab16
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab17) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab17
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab18) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab18
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab19) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab19
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab20) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab20
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab21) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab21
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab22) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab22
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab23) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab23
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab24) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab24
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab25) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab25
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab26) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab26
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab27) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab27
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab28) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab28
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab29) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab29
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t6) 1) 
       (*wplus2 (reg t6) (reg 2))
       (*jumpwlessp (label lab30) (reg t6) (reg t3))
       (*wdifference (reg t6) (reg t3))
       (*wplus2 (reg t4) 1)
    lab30                           %done
       (*move (reg t6) ($fluid *second-value*))
       (*move (reg t4) (reg 1))
       (*exit 6)))

(put 'addcarry 'opencode
       '(
	 (*MOVE ($FLUID carry*) (reg 2))
	 (*WPLUS2 (reg 1)(reg 2))
	 %(*MOVE (reg 1)(reg 3)) (*WSHIFT (reg 3) -30)
	 (shr 30 (reg 1) (reg 3))
	 (*MOVE (reg 3) ($FLUID carry*))
	 (*WSHIFT (reg 1) 2)
	 (*WSHIFT (reg 1) -2)))

(put 'subcarry 'opencode
       '(
	 (*MOVE ($FLUID carry*) (reg 2))
	 (*WDIFFERENCE (reg 1)(reg 2))
	 %(*MOVE (reg 1) (reg 2)) (*WSHIFT (reg 2) -31)
	 (shr 31 (reg 1) (reg 2))
	 (*MOVE (reg 2) ($FLUID carry*))
	 (*WSHIFT (reg 2) 30)
	 (*WPLUS2 (reg 1) (reg 2))
       ))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   funktions supporting word arithmetic with bbase*
%   bbase* is too large to be stored in a fluid cell;
%   so we perform operations with bbase* by shift and mask
%   instructions.

%   get the value for bbase* into a register
(put 'bbase** 'opencode
%     '((*MOVE 1 (reg 1))
%       (*WSHIFT (reg 1) 30)))
      '((orh 16#4000 (reg 0) (reg 1))))

%   get the value for logicalbits* into a register
(put 'logicalbits** 'opencode
      '((*MOVE -1 (reg 1))
	(*WSHIFT (reg 1) -2)))

%   calculate (remainder x bbase*)
(put 'remainder-bbase 'opencode
      '((*WSHIFT (reg 1) 2)
	(*WSHIFT (reg 1) -2)))

%   calculate (quotient x bbase*)
(put 'quotient-bbase 'opencode
      '((*WSHIFT (reg 1) -30)))

(put 'bbase** 'destroys '((reg 1)))
(put 'logicalbits** 'destroys '((reg 1)))
(put 'remainder-bbase 'destroys '((reg 1)))
(put 'quotient-bbase 'destroys '((reg 1)))
